#ifndef lint
static char *convert_inspec_sccsid = "@(#) convert_inspec.c 1.6 97/11/18\n";
#endif  /* lint */



/* (eds).
   Depunct non-author initials
   Remove , after surname?

/* Program to convert INSPEC database into 11-field Okapi exchange format */

/* November 1989 */

/* INSPEC comes in a MARC-like ISO 2709 format. On a unix system the variable
   blocklength tapes can be satisfactorily read using cat /dev/rmt* > whatever.

   Each record consists of a 24-character record label, a directory and some
   data fields. The directory and individual data fields, though not the record
   label, are terminated by a field mark (hex 1E or CTRL-^). The whole record
   ends with a record mark (hex 1D or CTRL-]). Each data field starts with a
   digit (often 1) which is a version number, and each subfield starts with
   a subfield mark (hex 1F or CTRL-_).

   Bytes 0-4 of the record label are the length of the record (inclusive of
   label and all field, subfield and record marks). Byte 5 contains N for new
   (it could maybe contain C for changed). Bytes 12-16 contain the offset of
   the start of the data fields from the beginning of the record label. The
   rest of the record label is unimportant.

   Byte 24 up to a field mark consists of a sequence of directory entries,
   one for each data field. Each directory entry conatins three subfields.
   There is one directory entry for each field in the record, and entries
   are in the same order as the fields to which they refer (although this
   is not important). The directory subfields are (bytes 0-2) a three-digit
   tag which identifies the field, (bytes 3-6) the field's length, (bytes
   7-11) the offset relative to the start of the data fields. Thus a suitable
   procedure for locating the data for a specified tag is: search directory
   the tag, read the offset subfield and add it to the offset of the start
   of the data fields (bytes 12-16 of the record label). This gives the offset
   of the required data from the start of the whole record. If the length is
   required this is given in bytes 3-6 of the directory entry.

*/

/*

  Procedure:

  Input is from stdin and output to stdout.

  get_next_record_inspec(buf) reads the next record into buf, returning its
  length, or -1 if end of input file. It checks for a complete record by
  reading the first 5 bytes as a length and looking for a record mark at the
  end. If this doesn't work it writes an error message, reads to a record mark
  and tries again.

  process_record_inspec(record, obuf) then does most of the work. It attempts
  to assemble an output record in Okapi exchange format field order. If
  successful it returns the length of the output record, otherwise a value
  less than or equal to zero depending on the cause of failure. Processing
  of different fields is individually coded.

  A certain amount of punctuation is put in, and some is removed. Authors'
  initials are depunctuated where possible. Titles have full stop added.
  Subfields in the controlled and free index term fields are separated
  by semicolons, the first term is given an initial capital, and the field
  terminated with a full stop. All names in the author field are given
  a "corporate name" indexing mark ('$'), because it is not possible to
  determine algorithmically whether they are personal or corporate. Titles
  have no indexing mark. Each subfield of the index terms fields is given
  a "title" indexing mark ('^').

  For details of which tags are used for which Okapi fields, see the comments
  with the process_record_inspec() function, together with the INSPEC Tape
  Services Manual.

  Two functions which are used in processing all fields are
  tag_data_address() and writefield_inspec(ip, op).

  tag_data_address() must be given the start of a directory entry. It returns
  the address and length of the first matching field, or NULL if
  there are no more occurrences of the tag (in case a tag may be repeated).

  writefield_inspec(ip, l, op) performs a number of standard substitutions
  on the contents of a field (of length l), returning the new length of the
  field. It:

    elides ~ ^ @ _ (unless (1997) modern is true)
    replaces $ by (dollars) followed by a blank if necessary
    replaces # or \234 by (pounds) followed by a blank if necessary
    replaces ` by ' and \ by / (unless modern is true)
    elides and reports non-ascii and control characters other than subfield
      marks (it is not given field or record marks).

  writefield() is called before the individual processing for specific fields.

  Following the individual processing of each field, # is written.

****************************************************************************/
   

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <values.h>

#define RECORD_MARK 0x1d
#define FIELD_MARK 0x1e
#define SUBFIELD_MARK 0x1f

#define LAB_LEN 24		/* Size of record label */
#define DIR_LEN 12		/* Length of a directory subfield */
#define TAG_LEN 3		/* Length of a 3-digit tag */

#define RECSIZE   65536
#define FIELD_LEN 16384

#define HASH      '#'		/* end of field for Okapi
				   No longer used (1997, use FIELD and
				   RECORD_MARKS) */

#define TRUE      1	
#define FALSE     0


char fcno[20] = "BOF" ;
char start_fcno[20] ;		/* Control number to start at */
char firstnewfcno[20] ;		/* Control number limiting "additions" */
char sections[20] = "CD" ;	/* Acceptable sections, default CD */
int searching = FALSE ;		/* True if not starting at the beginning */
int silent = FALSE ;		/* True if not displaying much */
int noisy = FALSE ;
int modern = FALSE ;		/* Elides fewer characters if true */
int newtape = 0 ;		/* If true removes two chars at end of rec (1997) */

char help[] = 
"Input and output are from stdin/stdout. All errors & diagnostics are to\n\
stderr.\n\
\
Command line arguments: all are optional. They include:\n\
  -start <control number> \n\
  -firstnew <first new control number> \n\
  -newtape (skips two bytes after end of record) \n\
  -modern (elides fewer characters than the old program) \n\
  -sections <ABCD> (default CD) \n\
  -recs <number of records to process>\n\
  -samples <frequency> (e.g. -samples 50 will dump every 50th record)\n\
  -diag (output records contain record type and tag values\n\
  -silent (inhibits most stderr output except samples, dumps and stats)\n\
  -noisy \n\
  -help\n\n" ;

int diag = 0 ;

char rec_buf[RECSIZE] ;	

/* statistical stuff */

int tag_occs[1000] ;		/* No of times tag n occurs */
int rec_types[20] ;

int no_auths = 0 ;
int no_date = 0 ;

main(argc, argv)
     int argc ;
     char *argv[] ;
{
  char out_buf[RECSIZE] ;
  int rec_len ;
  int status = 0 ;
  int record_count = 0 ;
  int good_recs = 0 ;
  int tot_length = 0 ;
  int skipped_wrong_class = 0 ;
  int before_start = 0 ;
  int wrong_rectype = 0 ;
  int no_cn, no_ti = 0 ;
  int i ;

  int argerr = FALSE ;		/* TRUE if argument error */
  char *thing ;
  int recs_to_do = MAXINT ;	/* Overridden by command arg -recs */
  int freq = 0 ;		/* Output diagnostic every freqth record */

  i = 1 ;
  while ( i < argc ) {
    if ( strcmp(argv[i], "-start") == 0 ) {
      if ( strlen(argv[i + 1]) > 19 ) {
	fprintf(stderr, "Start control number too long: %s\n", argv[i + 1]) ;
	exit(1) ;
      }
      strcpy(start_fcno, argv[i + 1]) ;
      searching = TRUE ;
      i++ ;
    }
    else if ( strcmp(argv[i], "-firstnew") == 0 ) {
      if ( strlen(argv[i + 1]) > 19 ) {
	fprintf(stderr, "First new control number too long: %s\n",
		argv[i + 1]) ;
	exit(1) ;
      }
      strcpy(firstnewfcno, argv[i + 1]) ;
      i++ ;
    }
    else if ( strcmp(argv[i], "-sections") == 0 ) 
      strcpy(sections, argv[++i]) ;
    else if ( strcmp(argv[i], "-newtape") == 0 ) newtape = 1 ;
    else if ( strcmp(argv[i], "-modern") == 0 ) modern = 1 ;
    else if ( strcmp(argv[i], "-dump") == 0 ) {	/* Not implemented */
      thing = argv[i + 1] ;
      i++ ;
    }
    else if (strcmp(argv[i], "-recs") == 0 ) {
      recs_to_do = atoi(argv[i + 1]) ;
      if ( recs_to_do <= 0 ) {
	fprintf(stderr, "Bad argument %s after -recs\n", argv[i + 1]) ;
	exit(1) ;
      }
      i++ ;
    }
    else if (strcmp(argv[i], "-diag") == 0 ) {
      diag = TRUE ;
    }
    else if (strcmp(argv[i], "-samples") == 0 ) {
      freq = atoi(argv[i + 1]) ;
      if ( freq <= 0 ) {
	fprintf(stderr, "Bad argument %s after -samples\n\
Must be a number. If the number is n every nth record will be output\n",
		argv[i + 1]) ;
	exit(1) ;
      }
      i++ ;
    }
    else if ( strcmp(argv[i], "-silent") == 0 ) silent = TRUE ;
    else if ( strcmp(argv[i], "-noisy") == 0 ) {
      noisy = TRUE ;
      silent = FALSE ;
    }
    else if (strcmp(argv[i], "-help") == 0) {
      fprintf(stderr, "%s", help) ;
      exit(1) ;
    }
    else {
      argerr = TRUE ;
      fprintf(stderr, "Unknown arg %s\n", argv[i]) ;
      }
    i++ ;
    }
  if ( argerr ) {
    fprintf(stderr, "%s", help) ;
    exit(1) ;
  }

  /* get next record and process */

  while (good_recs < recs_to_do) {
  /* get next record and process */

    *rec_buf = '\0' ;
    /* get next record ... break if at end of file */
    if ( (rec_len = get_next_record_inspec(rec_buf)) == -1 ) break;
    /* process next record */

    status = process_record_inspec(rec_buf, rec_len, out_buf) ;
    if (status >= 0) {
      /* negative indicates bad record or start control number
	 not yet found; will be skipped */
      
      if ( searching ) {
	searching = FALSE ;
	if ( ! silent ) fprintf(stderr, "Found start record %s\n", fcno) ;
      }
      /* write output record */
      printf("%s\n", out_buf) ;
      tot_length += status + 1 ; /* 1 for the \n */
      good_recs++ ;
      if ( freq && good_recs % freq == 0 ) 
	fprintf(stderr, "\nRecord %d. Input:\n%s\nOutput:\n%s\n",
		good_recs, rec_buf, out_buf) ;
      
    }
    else {
      if ( status == -50 ) skipped_wrong_class++ ;
      else if ( status == -100 ) before_start++ ;
      else if ( status == -10 || status == -999 ) wrong_rectype++ ;
      else if ( status == -2 ) no_ti++ ;
      else if ( status == -1 ) no_cn++ ;
    }
    record_count++ ;
  } /* while more records */

  fprintf(stderr, "Total number of records processed = %d\n", record_count) ;
  fprintf(stderr, "Good records: %d total length %d\n", good_recs, tot_length);
  fprintf(stderr, "%d with no date\n", no_date) ;
  if ( ! silent ) fprintf(stderr, "\nRecord types:\n") ;
  for ( i = 0 ; i < 20 ; i++ ) {
    if ( rec_types[i] > 0 ) {
      if ( ! silent ) fprintf(stderr, "%02d:\t%d\n", i, rec_types[i]) ;
    }
  }
  if ( ! silent ) fprintf(stderr, "Tag occurrences:\n") ;
  for ( i = 0 ; i < 1000 ; i++ ) {
    if ( tag_occs[i] > 0 ) {
      if ( ! silent ) fprintf(stderr, "%03d:\t%d\n", i, tag_occs[i]) ;
    }
  }
  if ( skipped_wrong_class ) 
    fprintf(stderr, "Skipped: wrong classification letters: %d\n",
	   skipped_wrong_class) ;
  if ( wrong_rectype ) fprintf(stderr, "Skipped: unwanted record type: %d\n",
			       wrong_rectype) ;
  if ( no_ti ) fprintf(stderr, "Skipped: no title or no author: %d\n",
			       no_ti) ;
  if ( no_cn ) fprintf(stderr, "Skipped: no control number: %d\n",
			       no_cn) ;
  if ( before_start )
    fprintf(stderr, "Skipped: control number less than start number: %d\n",
	    before_start) ;
  exit(0) ;
}



get_next_record_inspec(buf)
     char *buf ;

/* Documentation see head of file
   Now (Oct 97) simply reads through record mark, just issues warning
   if length read not same as record length field in header
*/
{
  register char *op ;
  register int ch ;
  int nch ;
  register int outl ;
  int reclen ;

  int good = FALSE ;		/* Becomes TRUE when good record read */
  int i ;
  static char label[6] = "00000" ; /*Storage for 1st 5 bytes = record length */

  while ( ! good ) {
    op = buf ; outl = 0 ; 
    while ( ( ch = getchar() ) != RECORD_MARK && ch != EOF ) {
      if ( newtape ) {
	if ( ch == 13 ) {
	  if ( (nch = getchar()) == 142 ) continue ;
	  else ungetc(nch, stdin) ;
	}
      }
      if ( ch == '\0' ) {
	if ( ! silent ) fprintf(stderr, "Null char elided\n") ;
	continue ;
      }
      *op++ = ch ;
      outl++ ;
      if ( outl >= RECSIZE ) {
	fprintf(stderr, "Record longer than %d characters, can't continue\n",
		outl) ;
	exit(1) ;
      }
    }
    if ( ch == EOF ) {
      if ( outl )
	fprintf(stderr, "Incomplete record at EOF, length %d\n", outl) ;
      return -1 ;
    }
    /* Then must have read record mark - check for good record */
    *op++ = ch ; outl++ ;	/* This should be the record mark */
    memcpy(label, buf, 5) ;
    reclen = atoi(label) ;
    if ( *(buf + reclen - 1) == RECORD_MARK ) good = TRUE ;
    else {	
      if ( ! silent ) fprintf(stderr,
       "Defective record, got length %d, reclen field says %d\n",
	      outl, reclen) ;
      if ( outl > 0 ) {
	if ( reclen <= 0 || noisy ) {
	  fprintf(stderr, "Dump follows\n") ;
	  for ( i = 0 ; i < outl ; i++ ) fprintf(stderr, "%c", buf[i]) ;
	  fprintf(stderr, "\n") ;
	}
	if ( reclen > 0 ) good = TRUE ;		/* (Oct 97) */
      }
    }
    *op = '\0' ;
  }
  return(outl) ;
}


/***************************************************************************

   find_tag_inspec()

   Specialised routine looks for string 'tag' at dirp and forward
   so long as dirp is below data_start. The calling program works
   out data_start from the contents of bytes 12-16 of the record label.
     It returns NULL if there is no directory entry for the tag is
   not present, else the address of the directory entry. In the latter
   case tag_data_address() can be used to find the start and length of the
   tag's data.
   In case it is called with dirp not pointing to a directory entry it gives
   up as soon as it would have to look beyond data_start.

****************************************************************************/

char *
find_tag_inspec(tag, dirp, data_start)
     register char *tag, *dirp, *data_start ;
{
  int found = FALSE ;

  while (! found && dirp < data_start) {
    if ( strncmp(dirp, tag, TAG_LEN) == 0 ) found = TRUE ;
    else dirp += DIR_LEN ;
  }
  if ( found ) return(dirp) ;
  else return(NULL) ;
}

/*****************************************************************************

  tag_data_address(tag, dirp, data_start, &length)

  Given dirp pointing to a directory entry returns the start and length of
  the data for the tag whose directory entry this is, or NULL if there is
  no further directory entry for the tag.

*****************************************************************************/

char *
tag_data_address(tag, dirp, data_start, length)
     char *tag, *dirp, *data_start ;
     int *length ;
{
  static char len[5] = "0000", start[6] = "00000" ;
  if ( ( dirp = find_tag_inspec(tag, dirp, data_start)) == NULL )
    return(NULL) ;
  (void) memcpy(len, dirp + 3, 4) ;
  (void) memcpy(start, dirp + 7, 5) ;
  *length = atoi(len) ;
  return(data_start + atoi(start)) ;
}
		 

/***************************************************************************

  writefield_inspec(ip, l, op) performs a number of standard substitutions
  on the contents of a field (of length l), returning the new length of the
  field. It:

    replaces $ by (dollars) followed by a blank if necessary
    replaces # or \234 by (pounds) followed by a blank if necessary
    unless "modern"
      elides ~ ^ @ _
      replaces ` by ' and \ by / 
    elides and reports non-ascii and control characters other than subfield
      marks (except field, subfield and record marks are left alone).

  writefield() is called before the individual processing for specific fields.

***************************************************************************/

writefield(ip, l, op)
     register char *ip, *op ;
     register int l ;
{
  register int outl = 0 ;
  register int i ;
  register char ch ;
  char *start = ip ;

  for ( i = 0 ; i < l ; i++ ) {
    ch = *ip++ ;
    if ( ! modern ) {
      if ( ch == '~' || ch == '^' || ch == '@' || ch == '_' ) {
	if ( ! silent ) fprintf(stderr, "%c elided\n", ch) ;
	continue ;
      }
      else if ( ch == '`' ) ch = '\'' ;
      else if ( ch == '\\' ) ch = '/' ;
    }
    if ( ch == '$' ) {
      strcpy(op, "(dollars)") ;
      op += 9 ; outl += 9 ;
      if ( *ip != ' ' ) {
	*op++ = ' ' ; outl++ ;
      }
    }
    else if ( ch == '#' || ch == '\234' ) {
      strcpy(op, "(pounds)") ;
      op += 8 ; outl += 8 ;
      if ( *ip != ' ' ) {
	*op++ = ' ' ; outl++ ;
      }
    }
    else if ( ! isascii(ch) ) {
      if ( ! silent ) fprintf(stderr,"Non-ascii character %c elided\n", ch) ;
      if ( noisy ) fprintf(stderr, "Context is%.*s\n", l, start) ;
      continue ;
    }
    else if ( iscntrl(ch) && ch != FIELD_MARK && ch != RECORD_MARK
	     && ch != SUBFIELD_MARK) {
      if ( ! silent ) fprintf(stderr,"Control character %d elided\n", ch) ;
      if ( noisy ) fprintf(stderr, "Context is %.*s\n", l, start);
      continue ;
    }
    else {
      *op++ = ch ;
      outl++ ;
    }
  }

  return(outl) ;
}


/***************************************************************************

  process_record_inspec()

*****************************************************************************/

/* process_record ... 21 Nov 89

   Assembles output record in obuf. Returns record length if OK.
   Negative return code indicates defective or unwanted record.

   Raw input record is in ibuf.

   An output record is assembled in Okapi field order.

*/

process_record_inspec(ibuf, rec_length, obuf)
     char * ibuf ;
     int rec_length ;
     char * obuf ;
{

  register char * ip, * op, * p ;
  char pbuf[RECSIZE] ;		/* Field processing buf for writefield */
  char *copy_ip ;
  char * fstart ;		/* Start of current input field */
  int fdlen ;			/* Output field length */
  int reclen ;

  register int in_word, word_length ;

  int i, l ;
  int iflen, ilen ;			/* Input field length */
  register char ch ;
  static char data_start_array[6] = "00000" ;
  unsigned limits = 0 ;		/* Dec 93 */
  /* char *limit_place ;	/* Limit will be put in here when we've done
                                   the date (Dec 93) */
  char *data_start ;		/* Address of start of first data field */
  char *dir_start ;		/* Start of directories */
  char lang[6] ;
  static char date_array[5] = "0000" ; /* To store date from issue date */
  int date ;
  int may_be_date = FALSE ;	/* True is there is an issue date field */
  int no_auth = FALSE ;
  int parens = FALSE ;
  int subfield_count ;

  int rec_type ;		/* Contents of 010 field as integer */
  int J = FALSE ;		/* Flag set if rec_type 05J or 06J*/
  int numtags ;
  char tags[10][4] ;		/* To store wanted tags for a field */

  dir_start = ibuf + LAB_LEN ;
  memcpy(data_start_array, ibuf + 12, 5) ;
  data_start = ibuf + atoi(data_start_array) ;

  /* Get control number */

  if ( (fstart = tag_data_address("001", dir_start, data_start, &iflen))
      == NULL || iflen < 4 ) {	/* Note every field begins with version
				   number and subfield mark */
    if ( ! silent ) fprintf(stderr, "No control number - record is:\n%.*s\n",
	    rec_length, ibuf) ;
    return -1 ;
  }
  tag_occs[atoi("001")]++ ;
  (void) memcpy(fcno, fstart + 2, iflen - 3) ; /* Don't want field mark */
  fcno[iflen - 3] = '\0' ;

  if ( searching ) {
    if ( strcmp(fcno, start_fcno) >= 0 ) {
      if ( ! silent ) fprintf(stderr, "Found start record %s\n", fcno) ;
      searching = FALSE ;
    }
    else return -100 ;
  }


  /* Get record type from 010 */

  if ( (fstart = tag_data_address("010", dir_start, data_start, &iflen))
      == NULL || iflen < 4 ) {	/* Note every field begins with version
				   number and subfield mark */
    if ( ! silent ) fprintf(stderr, "%s no record type field\n", fcno) ;
    return -10 ;
  }
  tag_occs[atoi("010")]++ ;
  rec_type = atoi(fstart + 2) ;
  rec_types[rec_type]++ ;
  
  if (rec_type == 5 || rec_type == 6) {
    if ( (fstart = tag_data_address("152", dir_start, data_start, &iflen))
	== NULL ) J = FALSE ; /* If in journal must be journal title */
    else J = TRUE ;
  }

    
  if ( rec_type != 2 && rec_type != 3 && rec_type != 4 && rec_type != 5
      && rec_type != 6 && rec_type != 10 && rec_type != 11 && rec_type != 12)
    return -999 ;

  /* Assemble fields and output them */

  op = obuf ; reclen = 0 ;

  if ( diag) {
    sprintf(op, "[%02d]", rec_type) ; op += 4 ; reclen += 4 ;
    if ( J ) { *op++ = 'J' ; reclen++ ; }
  }

  /* FIELD (1) (was field 6) CONTROL NUMBER */

  fdlen = 0 ;
  strcpy(op, fcno) ;
  l = strlen(fcno) ;
  op += l ; fdlen += l ; 
  *op++ = FIELD_MARK ; fdlen++ ;
  reclen += fdlen ;

  /* FIELD 2 (was field 1) - AUTHORS etc: tags 200, 210, 220 */

  fdlen = 0 ;

  strcpy(tags[0], "200") ; numtags = 1 ;
  if ( rec_type != 4 && rec_type != 6 && rec_type != 12 ) {
    strcpy(tags[1], "210") ;
    strcpy(tags[2], "220") ;
    numtags += 2 ;
  }

  for ( i = 0 ; i < numtags ; i++ ) {

    if ( (fstart = tag_data_address(tags[i], dir_start, data_start, &iflen))
	!= NULL && iflen > 2 ) {  /* Note every field begins with version
				     number and subfield mark */

      tag_occs[atoi(tags[i])]++ ;
      ip = fstart + 2 ;		/* Skip version number & 1st subfield mark */
      ilen = iflen - 3 ;	/* 2 for version number, 1st subfield mark
				   and field mark */
      ilen = writefield(ip, ilen, pbuf) ;
      ip = pbuf ;

      in_word = FALSE ;		/* For removing punct after initials */
      parens = FALSE ;
      word_length = 0 ;
      subfield_count = 1 ;

      *op++ = '$' ; fdlen++ ;

      while ( ilen > 0 ) {
	ch = *ip ;
	ilen-- ;
	if ( ch == SUBFIELD_MARK ) {
	  subfield_count++ ;
	  *op++ = ';' ; *op++ = ' ' ;
	  fdlen += 2 ;
	  ch = '$' ; /* Put in indexing mark */
	}
	else if ( ch == '.' && word_length == 1 ) { /* Depunctuate initials */
	  if ( isalpha(*(ip+1)) )	/* Need to put in a blank */
	    ch = ' ' ;
	  else {			/* Squash the dot */
	    ip++ ;
	    continue ;
	  }
	}
	if ( ch == ' ' || ch == '$') {
	  in_word = FALSE ; word_length = 0 ;
	}
	else {
	  if ( in_word ) word_length++ ;
	  else if ( isalnum(ch) ) {
	    in_word = TRUE ;
	    word_length = 1 ;
	  }
	}
	*op++ = ch ; ip++ ; fdlen++ ;
      }
      if ( fdlen > 0 && strcmp(tags[i], "210") == 0 ) { /* Write (ed) */
	if ( subfield_count == 1 ) {
	  strcpy(op, "(ed)") ;
	  op += 4 ; fdlen += 4 ;
	}
	else {
	  strcpy(op, "(eds)") ;
	  op += 5 ; fdlen += 5 ;
	}
      }
    }
  }
  /* Remove a trailing blank if necessary */
  if ( fdlen > 0 && *(op-1) == ' ' ) {
    op-- ; fdlen-- ;
  }

  if ( fdlen == 0 ) {
    no_auth = TRUE ;
    no_auths++ ;
    /*    if ( ! silent ) fprintf(stderr, "%s no author\n", fcno) ;  */
  }

  *op++ = FIELD_MARK ; fdlen++ ;
  reclen += fdlen ;

  /* FIELD 3 (was field 2)  - TITLE */
  
  fdlen = 0 ;
  strcpy(tags[0], "100") ; numtags = 1 ;
  if ( rec_type == 5 ) {
    strcpy(tags[1], "500") ; strcpy(tags[2], "800") ;
    numtags += 2 ;
  }

  for ( i = 0 ; i < numtags ; i++ ) {

    if ( (fstart = tag_data_address(tags[i], dir_start, data_start, &iflen))
	!= NULL && iflen > 2 ) {  /* Note every field begins with version
				     number and subfield mark */

      tag_occs[atoi(tags[i])]++ ;
      ip = fstart + 2 ;		/* Skip version number & subfield mark */
      ilen = iflen - 3 ;	/* Version number, subfield mark, field mark */
      ilen = writefield(ip, ilen, pbuf) ;
      ip = pbuf ;

      while ( ilen > 0 ) {
	ch = *ip ;
	ilen-- ;
	if ( ch == SUBFIELD_MARK ) {
	  ip++ ; continue ; /* Shouldn't have any */
	}
	*op++ = ch ; ip++ ; fdlen++ ;
      }
      if ( fdlen > 0 && numtags > 1 && ( i == 0 || i == 1 ) ) {
	*op++ = ',' ; *op++ = ' ' ;
	fdlen += 2 ;
      }
    }
  }
  if ( fdlen > 0 ) {
    *op++ = '.' ; fdlen++ ;
  }
  else {
    if ( ! silent ) fprintf(stderr, "%s no title\n", fcno) ;
    if ( no_auth ) {
      if ( ! silent ) fprintf(stderr,
	      "%s No author or titles, skipping\nRecord was:\n%.*s\n\n",
	      fcno, rec_length, ibuf) ;
      return -2 ;		/* Denotes no author or title */
    }
  }
  *op++ = FIELD_MARK ; fdlen++ ;
  reclen += fdlen ;

  /* FIELD 4 (was field 3)  ... SOURCE. 
     150 or 152 + 400 + 810 + 620 if rec type 02
     740 + 810 if rec type 03 (book) or 05 (conf proc)
     In: 150 + 740 + 810 + 210 + 620 if 04 (book chapter) or 06 (conf paper)
     or 12 (report section)
     150 + 400 + 810 if 05J (conf proc in journal)
     150 + 400 + 810 + 620 + "(Paper given at: " 170 + 500 + 800) if 06J
     740 + 810 + "(Dissertation. <if 820> Submitted " + 820)
     "(Report. " + 700.) + 740 + 810 if rec type 11 (report)
     
     640 is media
  */

  fdlen = 0 ;
  parens = FALSE ;

  if ( rec_type == 2 ) {
    strcpy(tags[0], "150") ; strcpy(tags[1], "400") ;
    strcpy(tags[2], "810") ; strcpy(tags[3], "620") ;
    numtags = 4 ;
  }
  else if ( rec_type == 3 ) {
    strcpy(tags[0], "740") ; strcpy(tags[1], "810") ;
    numtags = 2 ;
  }
  else if ( rec_type == 4 || (rec_type == 6 && ! J ) || rec_type == 12 ) {
    strcpy(tags[0], "150") ; strcpy(tags[1], "740") ;
    strcpy(tags[2], "810") ; strcpy(tags[3], "210") ;
    strcpy(tags[4], "620") ;
    numtags = 5 ;
  }
  else if ( rec_type == 5 ) {
    if ( J ) {
      strcpy(tags[0], "150") ; strcpy(tags[1], "400") ;
      strcpy(tags[2], "810") ;
      numtags = 3 ;
    }
    else {
      strcpy(tags[0], "740") ; strcpy(tags[1], "810") ;
      numtags = 2 ;
    }
  }
  else if ( rec_type == 6 && J ) {
    strcpy(tags[0], "150") ; strcpy(tags[1], "400") ;
    strcpy(tags[2], "810") ; strcpy(tags[3], "620") ;
    strcpy(tags[4], "170") ; strcpy(tags[5], "500") ;
    strcpy(tags[6], "800") ;
    numtags = 7 ;
  }
  else if ( rec_type == 10 ) {
    strcpy(tags[0], "740") ; strcpy(tags[1], "810") ;
    strcpy(tags[2], "820") ;
    numtags = 3 ;
  }
  else if ( rec_type == 11 ) {
    strcpy(tags[0], "700") ; strcpy(tags[1], "740") ;
    strcpy(tags[2], "810") ;
    numtags = 3 ;
  }
    
  else numtags = 0 ; /* Only other case is patents, which we don't bother
			about */

  for ( i = 0 ; i < numtags ; i++ ) {

    if ( rec_type == 10 && strcmp(tags[i], "810") == 0 ) {
      strcpy(op, "(Dissertation") ;
      op += 13 ; fdlen += 13 ;
      parens = TRUE ;
    }
    else if ( rec_type == 11 && strcmp(tags[i], "700") == 0 ) {
      strcpy(op, "(Report. ") ;
      op += 9 ; fdlen += 9 ;
      parens = TRUE ;
    }
    if ( (fstart = tag_data_address(tags[i], dir_start, data_start, &iflen))
      != NULL && iflen > 2 ) { 

      tag_occs[atoi(tags[i])]++ ;
      if ( diag ) {
	*op++ = '[' ; strcpy(op, tags[i]) ; op += 3 ; *op++ = ']' ;
	fdlen += 5 ;
      }
      ip = fstart + 2 ;		/* Skip version number and s/f mark */
      ilen = iflen - 3 ;	/* Version number, s/f/ mark and field mark */
      ilen = writefield(ip, ilen, pbuf) ;
      ip = pbuf ;

      if ( strcmp(tags[i], "150") == 0 && ( rec_type == 4 || rec_type == 12 ||
					   (rec_type == 6 && !J) ) ) {
	strcpy(op, "In: ") ; op += 4 ; fdlen += 4 ;
      }
      else if ( strcmp(tags[i], "210") == 0 ) {
	strcpy(op, "ed by ") ; op += 6 ; fdlen += 6 ;
      }
      else if ( strcmp(tags[i], "220") == 0 ) {
	strcpy(op, "trans by ") ; op += 9 ; fdlen += 9 ;
      }
      else if ( strcmp(tags[i], "170") == 0 ) {
	strcpy(op, "(Paper given at: ") ;
	op += 17 ; fdlen += 17 ;
	parens = TRUE ;
      }
      else if ( rec_type == 10 && strcmp(tags[i], "820") == 0 ) {
	strcpy(op, ", submitted ") ;
	op += 12 ; fdlen += 12 ;
      }
      if ( strcmp(tags[i], "210") != 0 && strcmp(tags[i], "220") != 0 ) {
	/* Not ed or trans */
	while ( ilen > 0 ) {
	  ch = *ip ;
	  ilen-- ;
	  if ( ch == SUBFIELD_MARK ) { /* Should only occur in 210, 220 */
	    *op++ = ';' ; *op++ = ' ' ;
	    fdlen += 2 ;
	    ip++ ; continue ;
	  }
	  *op++ = ch ; ip++ ; fdlen++ ;
	}
      }
      else {			/* ed or trans */
	in_word = FALSE ;		/* For removing punct after initials */
	word_length = 0 ;
	subfield_count = 1 ;

	while ( ilen > 0 ) {
	  ch = *ip ;
	  ilen-- ;
	  if ( ch == SUBFIELD_MARK ) {
	    subfield_count++ ;
	    in_word = FALSE ; word_length = 0 ;
	    *op++ = ';' ; *op++ = ' ' ;
	    fdlen += 2 ;
	    ip++ ;
	    continue ;
	  }
	  else if ( ch == '.' && word_length == 1 ) {
	    /* Depunctuate initials */
	    if ( isalpha(*(ip+1)) )	/* Need to put in a blank */
	      ch = ' ' ;
	    else {			/* Squash the dot */
	      ip++ ;
	      continue ;
	    }
	  }
	  if ( ch == ' ' ) {
	    in_word = FALSE ; word_length = 0 ;
	  }
	  else {
	    if ( in_word ) word_length++ ;
	    else if ( isalnum(ch) ) {
	      in_word = TRUE ;
	      word_length = 1 ;
	    }
	  }
	  *op++ = ch ; ip++ ; fdlen++ ;
	}
	if ( *(op - 1) == ' ' ) { op-- ; fdlen-- ; }
      }
      if ( strcmp(tags[i], "810") == 0 ) {
	/* Last four characters may be date of publication, save them */
	(void) memcpy(date_array, op - 4, 4) ;
	may_be_date = TRUE ;
      }
      *op++ = ',' ; *op++ = ' ' ; fdlen += 2 ;
    }
    if ( rec_type == 11 && i == 0 ) { /* Just done affiliation if any */
      while ( ! isalnum(*(op - 1)) ) {
	op-- ; fdlen-- ;
      }
      *op++ = ')' ; *op++ = ' ' ; fdlen += 2 ;
    }
  }


  /* Strip trailing spaces */
  while ( fdlen > 0 && ( *(op - 1) == ' ' || *(op - 1) == ',' ) ) {
    op-- ; fdlen-- ;
  }
  if ( parens ) {
    *op++ = ')' ; fdlen++ ;
    parens = FALSE ;
  }
  *op++ = FIELD_MARK ; fdlen++ ;
  reclen += fdlen ;

  /* (was an empty field here) */

  /* FIELD (5) CLASSIFICATION CODES */

  fdlen = 0 ;
  if ( (fstart = tag_data_address("120", dir_start, data_start, &iflen))
      != NULL && iflen > 2 ) {  /* 120: classification codes */

    ip = fstart + 2 ;		/*Skip version number and first subfield mark*/
    ilen = iflen - 3 ;		/* Version number, first subfield mark and
				   field mark */
    ilen = writefield(ip, ilen, pbuf) ;
    ip = pbuf ;

    {
      int goodclass = FALSE, cl_start = FALSE ;
      while ( ilen > 0 ) {
	ch = *ip ;
	if ( cl_start ) {
	  if ( strchr(sections, ch) ) goodclass = TRUE ;
	  cl_start = FALSE ;
	}
	ilen-- ;
	if ( ch == SUBFIELD_MARK ) {
	  ch = ' ' ; /* Assume OK to separate by blank*/
	  cl_start = TRUE ;
	}
	*op++ = ch ; ip++ ; fdlen++ ;
      }
      if ( goodclass == FALSE ) return -50 ;
    }
  }
  else if ( ! silent )
    fprintf(stderr, "%s no 120 classification code tag\n", fcno) ;
  *op++ = FIELD_MARK ; fdlen++ ;
  reclen += fdlen ;

  /* FIELD (6) (ASCII) DATE & do limits (was field 7: DATE code, LANG code) */

  fdlen = 0 ;

  if ( may_be_date ) {
    date = atoi(date_array) ;
    if ( date < 1960 || date > NEXT_YEAR ) {
      if ( ! silent ) fprintf(stderr, "%s strange date (%d)\n", fcno, date) ;
      date = 0 ;
    }
  }
  else {
    date = 0 ;
    /* if ( ! silent ) fprintf(stderr, "%s no date identified\n", fcno) ; */
    no_date++ ;
  }
  if ( date == 0 ) sprintf(op, "    %c", FIELD_MARK) ;
  else sprintf(op, "%.4d%c", date, FIELD_MARK) ;
  op += 5 ; fdlen += 5 ;
  reclen += fdlen ;

  if ( (fstart = tag_data_address("160", dir_start, data_start, &iflen))
      != NULL && iflen > 2 ) {  /* 160: language */

    tag_occs[atoi("160")]++ ;
    ip = fstart + 2 ;		/*Skip version number and first subfield mark*/
    ilen = iflen - 3 ;		/* Version number, first subfield mark and
				   field mark */
    ilen = writefield(ip, ilen, lang) ;
    ip = lang ;

    while ( ilen > 0 ) {
      ch = *ip ;
      if ( isupper(ch) ) *ip = tolower(ch) ;
      ip++ ; ilen-- ;
    }
    *ip = '\0' ;
  }
  else {			/* Assume eng */
    strcpy(lang, "eng") ;
  }

  /* Do limits */
  if ( strcmp(lang, "eng") == 0 ) limits |= 1 ;
  else limits |= 2 ;
  if ( date == 0 ) limits |= 128 ; /* No date */
  else {
    if ( date > NEXT_YEAR - 2 ) limits |= 4 ; /* New */
    if ( date > NEXT_YEAR - 4 ) limits |= 8 ; /* Recent */
  }
  if ( strcmp(fcno, firstnewfcno) > 0 ) /* Only works cos theyre numeric and
					   issued in sequence and all same length */
    limits |= 16 ;		/* Addition */
  else limits |= 32 ;		/* Previous */
  /* Could also do something with record type */
  /* Oct 97: moved limits to after last field */

  /* FIELD (7) (was field 8) FEATURE HEADINGS */
  /* Use 131 - Free indexing terms */
  /* This field is punctuated with semicolons and an indexing mark is inserted
     at the start of each subfield */

  fdlen = 0 ;
  if ( (fstart = tag_data_address("131", dir_start, data_start, &iflen))
      != NULL && iflen > 2 ) {  

    tag_occs[atoi("131")]++ ;
    ip = fstart + 2 ;		/* Skip version number & 1st subfield mark */

    ilen = iflen - 3 ;		/* Version number, subfield mark, field mark */

    ilen = writefield(ip, ilen, pbuf) ;
    ip = pbuf ;

    *op++ = '^' ; fdlen++ ;

    while ( ilen > 0 ) {
      ch = *ip ;
      ilen-- ;
      if ( ch == SUBFIELD_MARK ) {
	*op++ = ';' ; *op++ = ' ' ;
	fdlen += 2 ;
	ch = '^' ;
      }
      *op++ = ch ; ip++ ; fdlen++ ;
    }
    *op++ = '.' ; fdlen++ ;

    /* Uppercase 1st letter */
    ch = *(op - fdlen + 1) ;
    if (islower(ch)) *(op - fdlen + 1) = toupper(ch) ;
  }
  *op++ = FIELD_MARK ; fdlen++ ;
  reclen += fdlen ;

  /* FIELD (8) (was 9) ABSTRACT */

  fdlen = 0 ;
  if ( (fstart = tag_data_address("110", dir_start, data_start, &iflen))
      != NULL && iflen > 2 ) {  /* 110: abstract */

    tag_occs[atoi("110")]++ ;
    ip = fstart + 2 ;		/* Skip version number and subfield mark */
    ilen = iflen - 3 ;		/* Version number, subfield mark, field mark */
    ilen = writefield(ip, ilen, pbuf) ;
    ip = pbuf ;

    while ( ilen > 0 ) {
      ch = *ip ;
      ilen-- ;
      if ( ch == SUBFIELD_MARK ) {
	ip++ ; continue ; /* Shouldn't have any */
      }
      *op++ = ch ; ip++ ; fdlen++ ;
    }
    if ( isalnum(*(op - 1)) ) {
      *op++ = '.' ; fdlen++ ;
    }
  }
  /* else if ( ! silent ) fprintf(stderr, "%s no abstract\n", fcno) ; */

  *op++ = FIELD_MARK ; fdlen++ ;
  reclen += fdlen ;

  /* FIELD (9) (was 10) DESCRIPTORS */

  /* This field is punctuated with semicolons and an indexing mark is inserted
     at the start of each subfield */

  fdlen = 0 ;
  if ( (fstart = tag_data_address("130", dir_start, data_start, &iflen))
      != NULL && iflen > 2 ) {  

    tag_occs[atoi("130")]++ ;
    ip = fstart + 2 ;		/* Skip version number & 1st subfield mark */

    ilen = iflen - 3 ;		/* Version number, subfield mark, field mark */

    ilen = writefield(ip, ilen, pbuf) ;
    ip = pbuf ;

    *op++ = '^' ; fdlen++ ;

    while ( ilen > 0 ) {
      ch = *ip ;
      ilen-- ;
      if ( ch == SUBFIELD_MARK ) {
	*op++ = ';' ; *op++ = ' ' ;
	fdlen += 2 ;
	ch = '^' ;
      }
      *op++ = ch ; ip++ ; fdlen++ ;
    }
    *op++ = '.' ; fdlen++ ;

    /* Uppercase 1st letter */
    ch = *(op - fdlen + 1) ;
    if (islower(ch)) *(op - fdlen + 1) = toupper(ch) ;
  }
  *op++ = FIELD_MARK ; fdlen++ ;
  reclen += fdlen ;
  sprintf(op, "%.5d", limits) ;
  op += 5 ;
  *op++ = FIELD_MARK ;
  *op++ = RECORD_MARK ;
  reclen += 7 ;
  *op = '\0' ;

  return(reclen) ;

}

