#ifndef lint
static char *convert_marc_sccsid="@(#) convert_marc.c 1.3 94/11/08\n";
#endif

/******************************************************************************

  convert_marc_new  SW July 93

  As convert_marc but with control number as (what will be) field 1,
  lims, if any, at start as HASH-terminated digit string
  followed by year now stored as 'yyyy'

******************************************************************************/


/***************************************************************************

  Tape Conversion Program                        Julie Porteous  June 1988

  Renamed convert_marc Jan 1990

  Program for conversion of tapes from eight-bit MARC format to ascii okapi
  exchange format. Input data may or may not have segment control words, but
  must use the normal MARC record marks, field marks and subfield marks.
  Translated from Gill Venner's 1985 COBOL program. Record processing is more
  or less identical to that of the original program.

  Main difference is in the assembly of individual records from segments.
  Translation of characters from 8-bit to ascii is done for subfields as they
  are assembled into a field buffer, not for each block as was the case in the
  earlier version.

  Trivial revision Jan 92 & recompile SW
  Revised November 1989 - January 1990 SW

  The program is almost unreadable because of its origins (three
  translations from language to language). However, it works.
  This version uses standard input and output.

  There must be at least one command line argument -scw or -no_scw.
  These determine whether or not there are segment control words in the
  input. 
  Other command line arguments - see help string below.

  After extensive trials on a sample of the City University catalogue the
  program seems to be producing reasonable results.

  The program is largely controlled by #ifdef's. The only ones which work
  are CITY and BATH. One of these must be defined (can be done with -D in
  makefile).

  The process is roughly as follows:

  a record is read, either using segment control words or length as given
  in record label

  it is passed to process_record; tags are selected in accordance with
  the defined TAG_OCCS and initialised tag_values; characters are translated
  either by the function translate_char or (recently) in accordance with the
  hard-coded translation table; if the tag's indicators match the values
  given in tag_values the tag, indicators and field contents are stored
  in an element of the wanted_fields array (with "punctuation" - usually the
  three bytes following indicators in tag_values - incorporated - this is very
  obscure and needs careful reading before one dares amend tag_values).

  When the whole input record has been read, and tags which may be wanted
  stored in the wf array, the output record is assembled in Okapi field order.
  Terminal punctuation and indexing marks are added at this stage (hard coded).

  Finally, the output record is sent to stdout.

  Changes

  The program has been very considerably simplified (compare with convert_tape
  or the old COBOL program). There used to be a very messy attempt to remove
  duplicate subject headings - this has been replaced with an optional (but
  default) procedure called dedup(). Names from 7xx fields are concatenated
  to any 1xx name (7xx names in parentheses). The runtime file was always
  like this. The drawback of doing it at this stage is that the sorting
  (by title within author within reverse date) now sorts on a 7xx name if there
  is no 1xx. If CITY or BATH are defined, the control number goes in field 6
  by itself (originally copy numbers, instead of field 7. Any 505 contents
  notes now go in field 9.

  Error checking

  A lot of diagnostics and warnings may be obtained. Records are rejected
  if they appear to start or finish badly (get_next_record) or if there
  is no 001 field or a seriously defective 001, also if there is no 245.
  Spurious blanks at start of subfields are looked for.

  When processing a new file it is worth obtaining extensive output samples
  with the -samples flag, and comparing input and output. Special checks
  are not too difficult to write. The script dump_convert_marc_errors
  will nicely print stderr output collected in e.g. an emacs shell.
  The sorts of things which go wrong include: records starting/finishing
  badly - this is often due to spurious null characters, which can simply
  be removed by piping input through tr; 008 date in wrong position (not
  much can be done except try to get a date from a 260, which the program
  does); phoney indicators (replace them with '00'); ...

  Running the program

  There is one compulsory command line argument -scw or -no_scw

  Running with -help will list arguments etc

  Example% convert_marc -no_scw -samples 100 < marc_file >> converted_file

  or the output may be piped through the script sort_okapi, which puts
  records in reverse date order

****************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <values.h>

#undef OLD_CHECK_TAG
#define USING_TRANSLATION_TABLE

/* BATH or CITY (etc) must be defined in Makefile */

#define TEMP_LEN  9
#define SCW_LEN   5
#define RECSIZE   6000
#define FIELD_LEN 2048		/* Arbitrary, increase if crashes */
#define MAX_FIELDS  100		/* Arbitrary */

#define SC_NUM    14		/* number of special characters */
#define LAB_LEN   24		/* Length of record label */
#define DIR_LEN   12
#define TAG_LEN   3
#define DATA_LEN  4
#define SCP_LEN   5
#define FIELD_MARK   0x1E /* end of field in marc format (5 less than okapi) */
#define SUBFIELD_MARK   0x1F	/* subfield marker ... */
#define RECORD_MARK     0x1D

#define TRUE      1
#define FALSE     0


typedef struct {	/* store subfields for tags found in tags table */
  char tag[4] ;
  char ind1 ;
  char ind2 ;
  char field[FIELD_LEN] ;
} WANTED_FIELDS[MAX_FIELDS] ;
  
char help[] = 
"Input and output are from stdin/stdout. All errors & diagnostics are to\n\
stderr.\n\
\
Command line arguments: one of -scw or -no_scw must be present,\n\
the former if this is a tape-like file with segment control words,\n\
otherwise the latter.\n\n\
\
Other possible arguments include\n\
  -skip <no of characters to skip at start of input>\n\
    This is to allow skipping a \"header\" record or tape label\n\
    at the start of an input file.\n\
  -start <control number> (an LC number must be entered as lc<yy><nnnnnn>)\n\
  -recs <number of records to process>\n\
  -dump analytics|leftovers|505\n\
  -samples <frequency> (e.g. -samples 50 will dump every 50th record)\n\
  -diag <n> (various degrees of diagnostics, uses bits)\n\
  -silent (inhibits all stderr output except samples, dumps and stats)\n\
  -dedup <0|1> (tries to remove duplicate SHs if true (default))\n\
  -help\n\n" ;

int diag = 0 ;			/* Bits used, set by -diag flag */
int dump_analytics = FALSE, dump_505 = FALSE, dump_leftovers = FALSE ;

/* Global statistics */
int no_fcno = 0 ;
int phoney_fcno = 0 ;
int misc_errors = 0 ;
int no_245 = 0 ;
int badstarts = 0, badends = 0 ;
int no_date = 0, no_008_date = 0, spurious_008_date = 0, date_260 = 0 ;
int with_bad_008_length = 0 ;
int has_240 = 0, using_240 = 0 ;
int with_505 = 0, analyt_with_505 = 0, analytics = 0, with_leftover_tits = 0 ;
int no_class = 0 ;
int with_bad_inds = 0 ;
int with_leading_blanks = 0 ;
int with_eight_bit = 0, with_control = 0, with_null = 0 ;
int has_feature_hdgs = 0, has_topical_lcsh = 0, has_geog_lcsh = 0,
    has_pn_lcsh = 0, has_corp_lcsh = 0 ;
int no_subj = 0 ;
int dup_sh_recs = 0 ;		/* No of recs having duplicated SH */
int dup_sh_saved = 0 ;		/* Amount saved by removing dup SHs */

#ifdef CITY
int no_005 = 0 ;
#endif

#if defined CITY || defined BATH

#ifdef BATH
#define TAG_OCCS 30		/* TAGS 0 - 29 */
#endif
#ifdef CITY
#define TAG_OCCS 31
#endif

char * tag_values[] = {		/* Table of tags to be selected from MARC
				   record with indicators, subfields and
				   required punc */
#ifdef CITY
  "029 a ##b ##",
#endif
#ifdef BATH
  "082 a ##", 
#endif
  "083 a. #b ##",
#ifdef CITY
  "089 a ##d ##c ##",
#endif
  "1000a ##e ##f, #",
/* The 3 ifdef INITIALS haven't been tested, only tried without #define */
#ifdef INITITALS
  "1001a ##h2",
  "1002a ##h2",
  "1003a ##h2",
#else
  "1001a ##h, #",
  "1002a ##h, #",
  "1003a ##h, #",
#endif INITIALS
  "110 a ##c. #e (#",
  "111 a ##c. #e (#",
  "240 a3##",
  "245 a3##b : i; #j. #k = l. #",
  "248 g1##h3##i; #j. #",
  "250 a ##e, #",
  "260 b / c, #",
  "440 a3##",
  "490 a ##",
  "505 a7##",
  "6000a ##e ##f, #",
#ifdef INITIALS
  "6001a ##h2##",
  "6002a ##h2##",
  "6003a ##h2##",
#else
  "6001a ##h, #",
  "6002a ##h, #",
  "6003a ##h, #",
#endif
  "610 a ##c. #e (#",
  "611 a ##c. #e (#",
  "650 a. #x - y - z - ",
  "651 a. #x - y - z - ",
  "7000a ##e ##f, #",
#ifdef INITIALS
  "7001a ##h2##",
  "7002a ##h2##",
  "7003a ##h2##",
#else
  "7001a ##h, #",
  "7002a ##h, #",
  "7003a ##h, #",
#endif
  "710 a ##c. #e (#",
  "711 a ##c. #e (#"
  } ;
  
#endif defined BATH || defined CITY

int special_vals[] = { 66, 68, 70, 72, 74, 76, 98, 100, 102, 104, 106, 108, 112, 116 } ;

char * special_ch_vals[] =  {
  "L ",
  "O ",
  "D ",
  "TH",
  "AE",
  "OE",
  "l ",
  "o ",
  "d ",
  "th",
  "ae",
  "oe",
  "i ",
  "th"
  } ;
  

long record_count = 0 ;

char fcno[20] = "BOF" ;		/* Current record control number */
char start_fcno[20] ;		/* Control number to start at */
char first_fcno[20] ;		/* First in file */
int searching = FALSE ;		/* True if not starting at the beginning */
int silent = FALSE ;		/* True if not displaying much */
int dedup_sh = TRUE ;		/* Tries to remove dup SHs */

char rec_buf[RECSIZE] ;

#ifdef USING_TRANSLATION_TABLE

char * conversion_table[256] = {
  "\0","\001","\002","\003","\004","\005","\006","\007",
  "\010","\011","\012","\013","\014","\015","\016","\017",
  "\020","\021","\022","\023","\024","\025","\026","\027",
  "\030","\031","\032","\033","\034","*","#","$",
  "\040","\041","\042","\043","(dollars) ","\045","\046","\047",
  "\050","\051","\052","\053","\054","\055","\056","\057",
  "\060","\061","\062","\063","\064","\065","\066","\067",
  "\070","\071","\072","\073","\074","\075","\076","\077",
  "\100","\101","\102","\103","\104","\105","\106","\107",
  "\110","\111","\112","\113","\114","\115","\116","\117",
  "\120","\121","\122","\123","\124","\125","\126","\127",
  "\130","\131","\132","[","/","]","","",
  "'",	/* Should be 'sharp' but may be used for opening single quote */
         "\141","\142","\143","\144","\145","\146","\147",
  "\150","\151","\152","\153","\154","\155","\156","\157",
  "\160","\161","\162","\163","\164","\165","\166","\167",
  "\170","\171","\172","","","-","beta","gamma",
  /* \175 should be "alpha" but misused in City catalogue */
  "","","","","","","","",
  "","","","","","","","",
  "","","","","","","","",
  "","","","","","","","",
  "","L","O","D","Th","Ae","Oe","'",
  "",
#ifdef CITY  
  "", /* Should be " flat" but in CITY occurs spuriously before pub dates */
#else
  " flat",
#endif
  "","","","","","",
  "","l","o","d","th","ae","oe","",
  "i","(pounds) ","d","","","","","",
  "","","","","","","","",
  "","","","","","","","",
  "","","","","","","","",
  "","","","","","","","",
  "","","","","","","","",
  "","","","","","","","",
  "","","","","","","","",
  "","","","","","","",""
  } ;
#endif

/***************************************************************************

  convert_tape()

  Gets run time parameters and open files. While there are more records get
  next record and process.

  Reads next record from marc file, selects required tags and subfields,
  using table of tag_values.
  For each selected tag, translates the subfield to ascii, and adds appropriate
  punctuation after performing any exception processing (given by numeric
  "punctuation" in the tags table).
  Subfields are assembled and the whole field stored in struct of type
  WANTED_FIELDS.
  Fields within each record are separated by '#', empty fields are left blank,
  and the markers ^ _ | $ @ are inserted thus:

  ^ indicates title like data
  | indicates personal name
  $ indicates corporate name
  _ indicates no indexing required
  @ indicates editor/publisher data

*****************************************************************************/

main(argc, argv)
     int argc ;
     char * argv[] ;
{
  char temp[TEMP_LEN] ;
  char out_buf[RECSIZE] ;
  int rec_len = 0 ;
  int records = TRUE ;
  int good_recs = 0 ;
  int i ;
  int first = TRUE ;

  int argerr = FALSE ;		/* TRUE if argument error */
  char *thing ;
  int amount_to_skip = 0 ;		/* Characters to skip at BOF */
  int has_scw = 999 ;		/* Command line arg sets true or false */
  int recs_to_do = MAXINT ;	/* Overridden by command arg -recs */
  int freq = 0 ;		/* Output diagnostic every freqth record */

  i = 1 ;
  while ( i < argc ) {
    if ( strcmp(argv[i], "-skip") == 0 ) {
      amount_to_skip = atoi(argv[i + 1]) ;
      if ( amount_to_skip < 0 ) {
	fprintf(stderr, "Can't skip %d characters\n", amount_to_skip) ;
	exit(1) ;
      }
      i++ ;
    }
    else if ( strcmp(argv[i], "-start") == 0 ) {
      if ( strlen(argv[i + 1]) > 19 ) {
	fprintf(stderr, "Start control number too long: %s\n", argv[i + 1]) ;
	exit(1) ;
      }
      strcpy(start_fcno, argv[i + 1]) ;
      searching = TRUE ;
      i++ ;
    }
    else if ( strcmp(argv[i], "-no_scw") == 0 ) has_scw = FALSE ;
    else if ( strcmp(argv[i], "-scw") == 0 ) has_scw = TRUE ;
    else if ( strcmp(argv[i], "-dump") == 0 ) {
      thing = argv[i + 1] ;
      if ( strcmp(thing, "analytics") == 0 ) dump_analytics = TRUE ;
      else if ( strcmp(thing, "505") == 0 ) dump_505 = TRUE ;
      else if ( strcmp(thing, "leftovers") == 0 ) dump_leftovers = TRUE ;
      else {
	fprintf(stderr, "Unknown argument %s after -dump\n", thing) ;
	fprintf(stderr, "Must be one of analytics, 505, leftovers\n") ;
	exit(1) ;
      }
      i++ ;
    }
    else if (strcmp(argv[i], "-recs") == 0 ) {
      recs_to_do = atoi(argv[i + 1]) ;
      if ( recs_to_do <= 0 ) {
	fprintf(stderr, "Bad argument %s after -recs\n", argv[i + 1]) ;
	exit(1) ;
      }
      i++ ;
    }
    else if (strcmp(argv[i], "-diag") == 0 ) {
      diag = atoi(argv[i + 1]) ;
      i++ ;
    }
    else if (strcmp(argv[i], "-samples") == 0 ) {
      freq = atoi(argv[i + 1]) ;
      if ( freq <= 0 ) {
	fprintf(stderr, "Bad argument %s after -samples\n\
Must be a number. If the number is n every nth record will be output\n",
		argv[i + 1]) ;
	exit(1) ;
      }
      i++ ;
    }
    else if ( strcmp(argv[i], "-silent") == 0 ) silent = TRUE ;
    else if (strcmp(argv[i], "-help") == 0) {
      fprintf(stderr, "%s", help) ;
      exit(1) ;
    }
    else if (strcmp(argv[i], "-dedup") == 0 ) {
      if ( i == argc - 1 || ! isdigit(*argv[i + 1]) ) {
	fprintf(stderr, "Bad argument %s after -dedup\n\
Should be 0 or 1. If 1 (default) duplicate subject headings will be removed\n",
		argv[i + 1]) ;
	exit(1) ;
      }
      dedup_sh = atoi(argv[i + 1]) ;
      i++ ;
    }
    else {
      argerr = TRUE ;
      fprintf(stderr, "Unknown arg %s\n", argv[i]) ;
    }
    i++ ;
  }
  if ( argerr || argc < 2 || has_scw == 999 ) {
    fprintf(stderr, "%s", help) ;
    exit(1) ;
  }

  for ( i = 0 ; i< amount_to_skip ; i++ ) {
    if ( getchar() == EOF ) break ;
  }
  if (feof(stdin)) {
    fprintf(stderr, "EOF while skipping\n") ;
    exit(1) ;
  }

  /* get next record and process */

  while (record_count < recs_to_do) {
    *rec_buf = '\0' ;
    /* get next record ... break if at end of file */
    if ( (rec_len = get_next_record(rec_buf, has_scw)) == -1 ) break ;
    else if ( !silent && ! searching && (rec_len == -2 || rec_len == -3) )
      fprintf(stderr, "Last c/no %s\n", fcno) ;
    else {
      /* process next record */
      records = process_record(rec_buf, out_buf) ; /* Global searching is
						      used also (etc) */
      /* write output record */
      if (records == 1 && ! searching) {
	/* e.g. no tag "008" would have returned 8 etc... */
	printf("%s\n", out_buf) ;
	good_recs++ ;
	if ( first ) {
	  first = FALSE ;
	  strcpy(first_fcno, fcno) ;
	  if ( ! silent ) fprintf(stderr, "First (good) record is %s\n\n",
				  rec_buf) ;
	}
	if ( freq && good_recs % freq == 0 ) 
	  fprintf(stderr, "\nRecord %d. Input:\n%s\nOutput:\n%s\n",
		  good_recs, rec_buf, out_buf) ;
      }
    }
    if ( ! searching ) record_count++ ;
  } /* while more records */

  fprintf(stderr, "n\nEND OF RUN ***************************************\n\n");
  fprintf(stderr, "First record %s\n", first_fcno) ;
  fprintf(stderr, "Last record %s\n\n", fcno) ;
  fprintf(stderr, "\n\nSTATISTICS **************************************\n\n");
  fprintf(stderr, "Total number of records processed %d\n", record_count) ;
  fprintf(stderr, "Good records %d\n", good_recs) ;
  if ( no_fcno ) fprintf(stderr, "No control number, skipped, %d\n", no_fcno) ;
  if ( phoney_fcno ) fprintf(stderr, "Queer control number, skipped, %d\n",
			     phoney_fcno) ;
  if ( no_245 ) fprintf(stderr, "No 245 title, skipped, %d\n", no_245) ;
  if ( misc_errors )
    fprintf(stderr,
	    "Miscellaneous fatal record or processing errors, skipped, %d\n",
	    misc_errors) ;
  if ( badstarts || badends ) 
    fprintf(stderr, "Unprocessed records: %d bad starts, %d bad ends\n",
	    badstarts, badends) ;
  if ( with_bad_008_length )
    fprintf(stderr, "008 not length 40: %d\n", with_bad_008_length) ;
  fprintf(stderr,
     "No date %d\nNo plausible 008 date %d\nSpurious 008 date %d\nDate from 260 %d\n\n",
      no_date, no_008_date, spurious_008_date, date_260) ;
#ifdef CITY
  fprintf(stderr, "No 005 %d\n", no_005) ;
#endif
  fprintf(stderr, "With spurious indicator(s) (0 substituted) %d\n",
	  with_bad_inds) ;
  fprintf(stderr, "With leading blank(s) in subfield(s) %d\n",
	  with_leading_blanks) ;
  if ( diag & 8 ) 
    fprintf(stderr, "With eight-bit char(s) %d, with control char(s) %d, \
with null char(s) %d\n", with_eight_bit, with_control, with_null) ;
  fprintf(stderr, "Subject headings: feature headings %d recs\n",
	  has_feature_hdgs) ;
  fprintf(stderr, "                  topical LCSH %d recs\n",
	  has_topical_lcsh) ;
  fprintf(stderr, "                  geographical LCSH %d\n",
	  has_geog_lcsh) ;
  fprintf(stderr, "                  pers. name LCSH %d\n",
	  has_pn_lcsh) ;
  fprintf(stderr, "                  corp. name LCSH %d\n",
	  has_corp_lcsh) ;
  fprintf(stderr, "No subject headings at all %d\n", no_subj) ;
  fprintf(stderr, "%d records have duplicate subject headings\n",
	  dup_sh_recs ) ;
  /* fprintf(stderr, "Amount saved by removing duplicated: %d bytes\n",
	  dup_sh_saved) ; */

  fprintf(stderr,
     "\nWith uniform title %d, using uniform title as title %d\n",
	  has_240, using_240) ;
  fprintf(stderr,
     "Analytics %d, with 505 %d (total 505s %d)\n",
          analytics, analyt_with_505, with_505) ;
  fprintf(stderr,
     "With leftover series statements %d\n",
	  with_leftover_tits) ;
  fprintf(stderr,
     "No class number/shelf mark %d\n", no_class) ;
}





/* get_next_record 

   Reads next record from stdin into rec_buf and leaves file pointer at start
   of next record. This will be an segment control word if there are any,
   otherwise the start of the record label.
   Returns record_length on successful execution else returns -1 if EOF
   else -2 and error message if record starts badly. In this case it skips
   to the next record mark so that the next record can be read.
   If record "ends" with anything other than a record mark it prints an
   error message, skips to record mark and returns -3. In the latter case
   the record may be usable (but length is unknown).
*/

  
get_next_record(rec_buf, has_scw) /* get next logical record from file */
     char * rec_buf ;
     int has_scw ;		/* True if there are segment control words */
{
  char scw[SCW_LEN] ;
  char span ;
  register char *p, *op ;
  int slen ;
  int rlen = 0 ;
  register int ch, i ;
  int badstart = FALSE ;

  op = rec_buf ;
  if ( has_scw ) {
    p = scw ;
    for ( i = 0 ; i < SCW_LEN ; i++ ) {
      if ((ch = getchar()) == EOF ) break ;
      if ( ! isdigit(ch) ) badstart = TRUE ;
      *p++ = ch ;
    }
    if ( ch == EOF ) return -1 ;
    if ( badstart ) {
      badstarts++ ;
      if ( ! searching && ! silent )
	fprintf(stderr, "Record starts badly with %.5s, skipping..\n", scw) ;
      if (skip_to_end() == EOF ) return -1 ;
      else return -2 ;
    }
    span = scw[0] ;
    (void) sscanf(scw + 1, "%4d", &slen) ;
    
    /* read next segment into buffer */
    for (i = 0 ; i < slen - SCW_LEN ; i++ ) {
      if ( ( ch = getchar() ) == EOF ) break ;
      *op++ = ch ;
    }
    if ( ch == EOF ) return -1 ;     
    rlen += slen - SCW_LEN ;
    /* spanning indicator '1': record begins but doesn't end in this segment 
       spanning indicator '2': record neither begins or ends in this segment */
    if ( span == '1' ) {
      while (span != '3') {
	p = scw ;
	for ( i = 0 ; i < SCW_LEN ; i++ ) {
	  if ((ch = getchar()) == EOF ) break ;
	  *p++ = ch ;
	}
	if ( ch == EOF ) return -1 ;
	span = scw[0] ;
	(void) sscanf(scw + 1, "%4d", &slen) ;
	/* read segment and append into rec_buf */
	for (i = 0 ; i < slen - SCW_LEN ; i++ ) {
	  if ( ( ch = getchar() ) == EOF ) break ;
	  *op++ = ch ;
	}
	if ( ch == EOF ) return -1 ;     
	rlen += slen - SCW_LEN ;
      }
    }
  }
  else {			/* no segment control words */
    for ( i = 0 ; i < 5 ; i++ )  {
      if ((ch = getchar()) == EOF ) break ;
      if ( ! isdigit(ch) ) badstart = TRUE ;
      *op++ = ch ;
    }
    if ( ch == EOF ) return -1 ;
    if ( badstart ) {
      badstarts++ ;
      if ( ! searching && ! silent )
	fprintf(stderr,
		"Record starts badly with %.5s, skipping..\n", rec_buf) ;
      if (skip_to_end() == EOF) return -1 ;
      else return -2 ;
    }
    (void) sscanf(rec_buf, "%5d", &slen) ;
    for (i = 0 ; i < slen - 5 ; i++ ) {
      if ( ( ch = getchar() ) == EOF ) break ;
      *op++ = ch ;
    }
    if ( ch == EOF ) return -1 ;     
    rlen = slen ;
  } /* No segment control words */
    
  *op = '\0' ;
    
  if ( rec_buf[rlen - 1] != RECORD_MARK ) {
    badends++ ;
    if ( ! searching && ! silent ) {
      fprintf(stderr, "Record ends badly, record follows:\n%s\n", rec_buf) ;
      fprintf(stderr, "Skipped to record mark (or EOF)\n\n") ;
    }
    if ( skip_to_end() == 0 ) return -3 ;
    else return -1 ;
  }

  return rlen ;
}
  
skip_to_end()			/* Skip to next record mark */
{
  register int ch ;
  register int i = 0 ;

  while ( ( ch = getchar() ) != RECORD_MARK && ch != EOF ) i++ ;
  if ( ! silent )
    fprintf(stderr, "Skipped %d chars through record mark\n", i) ;
  if ( ch == RECORD_MARK ) return 0 ;
  else return EOF ;
}

/* process_record ... 2 june 88 

   Selects required tags and subfields using table of tag values. Adds
   appropriate punctuation after performing any exception processing which
   is given by numeric "punctuation" in the tags table.
   If fatal error on record returns value greater than 1. */


process_record(ibuf, obuf)
				/* marc record into okapi exchange format */
     char * ibuf ;
     char * obuf ;
{
  WANTED_FIELDS wf ;

  char * ip ;
  register char * op = obuf ;
  char * t, tag[4] ;
  char * b ;
  char tag_buf[FIELD_LEN] ;
  char * sp ;
  char sub_buf[FIELD_LEN] ;
  char buf[10] ;
  char wanted_delim[3] ;
  char * data_ptr ;
  register char * temp ;
  char *vtemp ;
  int jtemp ;
  char temp_buf[FIELD_LEN] ;
  char mark ;
  register char * this_tags_data ;
  register char ch ;
  static char lang[4] = "  " ;
  char punc[4] ;
  char ind1, ind2 ;
  char * sf ;
  char * tag_ptr ;
  char *start ;
  int length, newlength ;
  int cno_length = 0 ;

  int num_dirs ;		/* Number of directory subrecords */
  int sd_oset ;		/* offset of sub-directory in record  */
  int scp ;		/* starting character position relative to base */
  int data_len ;		/* Length of tag's data given by directory */
  int base_data_addr ;	/* base address of data - offset relative to start of 
			   record label */
  int sub_dir_start ;		/* Start of 1st analytic subdirectory record,
				   offset from start of record */
  int datecode = 0 ;
  unsigned limits = 0 ;
  int a_240 = 0 ;
  int a_245 = 0 ;
  int a_505 = 0 ;
  int count = 0 ;
  register int cc = 0 ;
  int tp = 0 ;
  int a = 0 ;			/* Indexes array of tags and data (wf[]) */
  int ntags  = 0 ;		/* Number of tags extracted */
  register int i, j ;

  int length_written ;

  int analyt = FALSE ;
  int leftover_tits = FALSE ;
  int bad_inds = FALSE ;
  int cno_error ;
  int no_digits ;		/* Booleans for control number checking */
  int leading_blanks = 0 ;
  int has_eight_bit = 0, has_control = 0, has_null = 0 ;
  int feature_hdgs = 0, topical_lcsh = 0, geog_lcsh = 0,
      pn_lcsh = 0, corp_lcsh = 0 ;
  int nsh = 0 ;			/* No of subj hdgs this rec */
  *tag_buf = ' ' ;

  /* get base address from record label then skip to first directory */
  (void) sscanf(ibuf + 12, "%5d", &base_data_addr) ;

  num_dirs = (base_data_addr - LAB_LEN) / DIR_LEN ;
  /* But will be fewer if there are subrecord directories (see 002) */

  /* get record control number from "001" tag */
  ip = ibuf + LAB_LEN ;
  next_tag(ip, tag) ;
  num_dirs-- ;
  if ( strcmp(tag, "001") != 0 ) {
    no_fcno++ ;
    if ( ! silent )
      fprintf(stderr, "No control number, last c/no %s\n", fcno) ;
    return 1000 ;
  }
  this_tags_data = ibuf + base_data_addr ;
  temp = fcno ;
  no_digits = TRUE ;
  cno_error = FALSE ;
  
  if ( *(this_tags_data + 2) == '-' ) {	/* Assume lc number */
    *temp++ = 'l' ; *temp++ = 'c' ;
    cno_length += 2 ;
  }
  while (( ch = *this_tags_data++ ) != FIELD_MARK && cno_length < 20 ) {
    if ( isdigit(ch) ) no_digits = FALSE ;
    if ( ch < '0' && ch != '-' ) {
      cno_error = TRUE ;
      continue ;
    }
    *temp++ = ch ;
    cno_length++ ;
  }
  if ( cno_length == 20 ) {
    temp-- ;
    if ( ! silent )
      fprintf(stderr, "%s control number too long, truncated\n", fcno) ;
  }
  *temp = '\0' ;
  if ( strncmpnc(fcno, "lc", 2) == 0 && *(fcno + 4) == '-' ) {
    if ( ! silent && ( diag & 32 ) )
      fprintf(stderr, "LC no %s changed to ", fcno) ;
    *(fcno + 4) = '0' ;		/* Replace hyphen */
    jtemp = strlen(fcno) ;
    if ( jtemp < 10 ) {		/* want lcyynnnnnn */
      strcpy(temp_buf, fcno + 5) ;
      for ( i = 0, vtemp = fcno + 5 ; i < 10 - jtemp ; i++ )
	*vtemp++ = '0' ;
      strcpy(vtemp, temp_buf) ;
    }
    if ( ! silent && ( diag & 32 ) ) fprintf(stderr, "%s\n", fcno) ;
  }
  if ( searching ) {
    if ( strcmp(fcno, start_fcno) == 0 ) {
      if ( ! silent ) fprintf(stderr, "Found start record %s\n", fcno) ;
      searching = FALSE ;
    }
    else return 1 ;
  }
  if ( no_digits ) {
    phoney_fcno++ ;
    if ( ! silent ) 
      fprintf(stderr, "Phoney control number %s, record follows\n%s\n\n",
	      fcno, ibuf) ;
    return 1000 ;
  }
  if ( cno_error )
    if ( ! silent ) 
      fprintf(stderr,
	      "Implausible character(s) removed from control number, \
record follows\n%s\n\n",
	      ibuf) ;

  ip += DIR_LEN ;
  next_tag(ip, tag) ;
  num_dirs-- ;

  /* check whether there is an analytical subdirectory present */
  if (strcmp(tag, "002") == 0) {
    scp = get_scp(ip) ;	/* Start character position of this tag's data */
    this_tags_data = ibuf + base_data_addr + scp ;
    sscanf(this_tags_data + 7, "%5d", &sub_dir_start) ;
    /* Gives offset in whole record of 1st subrecord directory */
    analyt = (base_data_addr - sub_dir_start) / DIR_LEN ;
    num_dirs -= analyt ;
    if ( analyt ) analytics++ ;
  }

#ifdef CITY
  while ( strcmp(tag, "005") < 0 ) {
    ip += DIR_LEN ;
    next_tag(ip, tag) ;		/* next tag should be "008", except OCLC-type
				   records, which have a 005 */
    num_dirs-- ;
  }
  if (strcmp(tag, "005") != 0) { /* Not a fatal error but no amendment date */
    if (diag & 1) 
      if ( ! silent ) 
	fprintf(stderr, "No 005 c/no %s tag %s\n", fcno, tag) ;
    no_005++ ;
  }
#endif CITY

  while ( strcmp(tag, "008") < 0 ) {
    ip += DIR_LEN ;
    next_tag(ip, tag) ;		/* next tag should be "008" */
    num_dirs-- ;
  }

  if (strcmp(tag, "008") != 0) { /* Not a fatal error but no date or lang */
    if ( ! silent ) fprintf(stderr, "No 008 c/no %s tag %s\n", fcno, tag) ;
    sprintf(lang, "%s", "   ") ;
  }
  else {
    /* process "008" tag */
    scp = get_scp(ip) ;
    data_len = get_data_len(ip) ;
    if ( data_len != 41 ) {
      with_bad_008_length++ ;
      if ( ! silent )
	fprintf(stderr,
		"%s incorrect 008 length %d, skipping %d bytes\n",
		fcno, data_len, data_len - 41) ;
      scp += data_len - 41 ;
    }
    data_ptr = ibuf + base_data_addr + scp + 36 - 1 ;
    /* advance to language code */
    /* get language code from positions 36 - 38 in "008" field */
    (void) sscanf(data_ptr, "%3s", lang) ;
    
    data_ptr = ibuf + base_data_addr + scp + 8 - 1 ;/* advance to date1 pos */
    sscanf(data_ptr, "%4d", &datecode) ;
    if ( datecode <= 0 ) sscanf(data_ptr + 4, "%4d", &datecode) ; 
    if ( datecode <= NEXT_YEAR - 1000 ) {
      if ( datecode > 800 ) {
	fprintf(stderr,
		"%s dodgy 008 date %d corrected? to %d\n",
		fcno, datecode, datecode + 1000) ;
	datecode += 1000 ;
      }
      else if ( datecode != 0 ) {
	fprintf(stderr, "%s 008 date %d set to 0\n",
		fcno, datecode) ;
	datecode = 0 ;	/* (Dec 93) */
      }
    }
    ip += DIR_LEN ;
    next_tag(ip, tag) ;
    num_dirs-- ;
  }				/* 008 exists */
  
  if ( datecode == 0 ) no_008_date++ ;
  else if ( datecode < 0 || datecode > NEXT_YEAR ) {
    if ( ! silent ) 
      fprintf(stderr, "%s spurious 008 date %d\n", fcno, datecode) ;
    spurious_008_date++ ;
    datecode = 0 ;
  }

  /* now process tags until end of record */

  *wanted_delim = ' ' ;

  while ( num_dirs >= 0 ) {

    if (*wanted_delim == SUBFIELD_MARK || *wanted_delim == '$' )
      sf = this_tags_data++ ;		/* skip to get-subfield */
    else {
      sp = sub_buf ;   	/* reset pointer for next subfield */

      if (check_tag(tag, &tp) == -1) { /* tag not found ... go to next tag */
	ip += DIR_LEN ;
	next_tag(ip, tag) ;
	num_dirs-- ;
	continue ;
      }
      scp = get_scp(ip) ;
      
      this_tags_data = ibuf + base_data_addr + scp ;

      ind1 = *this_tags_data++ ;
      ind2 = *this_tags_data++ ;
      if ( ind1 == ' ' ) ind1 = '0' ;
      if ( ind2 == ' ' ) ind2 = '0' ;
      if (!isdigit(ind1) || !isdigit(ind2)) {
	if ( ! silent ) 
	  fprintf(stderr, "%s tag %s bad indicators '%c%c', substituting 0\n",
		  fcno, tag, ind1, ind2) ;
	bad_inds = TRUE ;
	if ( !isdigit(ind1) ) ind1 = '0' ;
	if ( !isdigit(ind2) ) ind2 = '0' ;
      }
      if (tag_values[tp][3] != ' ') 
	if (check_first_indicator(ind1, &tp, tag) == -1) {
	  next_tag(ip, tag) ;
	  num_dirs-- ;
	  continue ; 
	}

      /* this_tag */
      this_tags_data++ ;	/* Skip subfield mark */
      sf = this_tags_data++ ;	/* sf points to subfield letter, data to data*/
    }


    /* get subfield */
    cc = 0 ;
    
    while (cc < FIELD_LEN) {
#ifdef USING_TRANSLATION_TABLE
      register int ch ;
      register char tch ;
      register char *p ;

      ch =  * (unsigned char *) this_tags_data++ ;
      if ( ch == ' ' ) {
	if ( cc == 0 ) {
	  leading_blanks++ ;
	  if ( ( diag & 4 ) && ! silent )
	    fprintf(stderr, "%s tag %s subfield starts with blank\n",
		    fcno, tag) ;
	  continue ;
	}
      }
      else if ( (diag & 8) ) {
	if ( ch > 0x7f ) {
	  has_eight_bit = TRUE ;
	  if ( (diag & 16) && ! silent ) 
	    fprintf(stderr,
		    "%s tag %s has eight-bit char %x, translation '%s'\n",
		    fcno, tag, ch, conversion_table[ch]) ;
	}
	else if ( ch == 0 ) {
	  has_null = TRUE ;
	  if ( (diag & 16) && ! silent ) 
	    fprintf(stderr, "%s tag %s has null character\n", fcno, tag) ;
	}
	else if ( ch < FIELD_MARK ) {
	  has_control = TRUE ;
	  if ( (diag & 16) && ! silent ) 
	    fprintf(stderr,
		    "%s tag %s has control char %x\n", fcno, tag, ch) ;
	}
      }
      p = conversion_table[ch] ;
      while ( tch = *p++ ) {
	if ( tch == '#' || tch == '$' ) break ;
	*sp++ = tch ;
	cc++ ;
      }
      if (tch == '#' || tch == '$') {
	*wanted_delim = tch ;
	break ;
      }
#else
      temp  = temp_buf ;
      count = translate_char(this_tags_data, temp) ;
      if (*temp_buf == '#' || *temp_buf == '$') {
	*wanted_delim = *temp_buf ;
	this_tags_data++ ;
	break ;
      }
      if (count) {
	while (*temp != '\0') *sp++ = *temp++ ; 
	this_tags_data += count ; /* Think this is wrong */
      }
      else {
	*sp++ = *temp ; 
	this_tags_data++ ;
      }
      cc++ ;
#endif
    }
  
    if (cc)  { *sp++ = '#' ; *sp++ = '\0' ; }

    sp = sub_buf ;
    if (cc)  {
      /* Check if this subfield is wanted..
	 This check should be done before translating, but it may be
	 a little difficult to change it.
      */
      tag_ptr  = tag_values[tp] + 4 ; /* character after tag and first
					 indicator */
      while (*sf != *tag_ptr && *tag_ptr != '\0') tag_ptr += 4 ; 
      if (*sf == *tag_ptr) { /* if 1st char of subfield found in tag_values */
	tag_ptr++ ;
	strncpy(punc, tag_ptr, 3) ; /* get punctuation from tag tables */

	if (! isdigit(*punc)) {

	  /* Here a clumsy check for date from 260 $c if we haven't got it
	     from 008 */
	  if ( datecode == 0 && *sf == 'c' && strcmp(tag, "260") == 0 ) {
	    temp = sub_buf ;
	    while ( ( ch = *temp++ ) != '\0' && ! isdigit(ch) ) ;
	    if ( isdigit(ch) && isdigit(*(temp+2)) )
	      sscanf(temp - 1, "%4d", &datecode) ;
	    if ( datecode <= NEXT_YEAR - 1000 ) {
	      if ( datecode > 800 ) {
		fprintf(stderr,
			"%s dodgy 260 date %d corrected? to %d\n",
			fcno, datecode, datecode + 1000) ;
		datecode += 1000 ;
	      }
	      else datecode = 0 ;	/* (Dec 93) */
	    }
	    if ( datecode ) date_260++ ;
	  }
	  save_subfield(tag_buf, punc, sub_buf) ;
	}
	else if (*punc == '3') { /* may have article */
	  if (ind2 == '0') {
	    if (strcmp(tag, "248") == 0) strcpy(punc, ": ^") ;
	    else              strncpy(punc, "$", 1) ;
	    save_subfield(tag_buf, punc, sub_buf) ;
	  }
	  else {
	    dont_index_article(tag_buf, sub_buf, ind2, tag) ;
	  }
	}
	else if (*punc == '2') { /* subfield is 100h */
	  initials(sub_buf, punc) ; 
	  save_subfield(tag_buf, punc, sub_buf) ;
	}
	else if (*punc == '1') { /* subfield is 248g */
	  /* dont_index_248g */
	  if (*tag_buf == ' ') *tag_buf = '\0'  ;
	  strcat(tag_buf, "_") ;
	  
	  temp = tag_buf ; while (*temp != '\0') temp++ ;
	  b   = sub_buf ;
	  while (*b != '\0') {
	    if (*b == '#') b++ ;	/* skip end of field mark within string */
	    else  *temp++ = *b++ ;
	  }
	  *temp = '\0' ;
	}
	else if (*punc == '7') { /* tag is 505 */
	  save_subfield(tag_buf, punc, sub_buf) ;
	}
	else {			/* unknown punc ... skip to next record */
	  misc_errors++ ;
	  if ( ! silent ) 
	    fprintf(stderr,"%s unknown punc '%s' - skipping to next record\n",
		    fcno, punc) ; 
	  return(7) ;
	}
      } /* check subfield */
    } /* if (cc) ... ones that aren't skip to next_subfield */
    
    if (*wanted_delim == '$') 
      continue ;				/* skip to get_subfield */
    else {
      if (*tag_buf != ' ') {
	strcpy(wf[a].tag, tag) ;
	wf[a].ind1 = ind1 ;
	wf[a].ind2 = ind2 ;
	
	strcpy(wf[a].field, tag_buf) ;
	strcat(wf[a].field, "#") ; 
	a++ ;
	*tag_buf = ' ' ;	/* re-initialise tag buffer */

	while (tp && strncmp(tag_values[tp], tag, 3) == 0 ) if (tp) tp-- ;

      }
      ip += DIR_LEN ;
      next_tag(ip, tag) ;
      num_dirs-- ;
    }
  }

  
  /* end of record ... assemble fields and output them */

  ntags = a ;
  strcpy(wf[a].tag, "999") ;	/* Should be redundant, was a stopper.
				   999 may be used as local field */

  *obuf = '\0' ;
  op = obuf ;

  /* Code for printing out the wf array */

  if ( (diag & 128) && !silent )
    for ( i = 0 ; i < ntags ; i++ ) {
      fprintf(stderr, "%s/%c%c/%s\n\n",
	      wf[i].tag, wf[i].ind1, wf[i].ind2, wf[i].field) ;
    }

  /* 1st field now control number */

  strcpy(op, fcno) ;
  op += strlen(fcno) ;

  op += finish_field(op, obuf) ;

  /* Field 2: year */
  if ( datecode > NEXT_YEAR || datecode < 0 ) datecode = 0 ;
  if ( datecode == 0 ) {
    if ( (diag & 4) && ! silent ) fprintf(stderr, "%s no date\n", fcno) ;
    no_date++ ;
  }

  /* Do limits & year */
  /* But note (nov 94) limits now go after last field, to suit new runtime
     conversion */
  if ( strcmp(lang, "eng") == 0 ) limits |= 1 ;
  else limits |= 2 ;
  if ( datecode == 0 ) limits |= 128 ; /* No date */
  else {
    if ( datecode > NEXT_YEAR - 2 ) limits |= 4 ; /* New */
    if ( datecode > NEXT_YEAR - 5 ) limits |= 8 ; /* Recent */
    if ( datecode > 1837 && datecode < 1902 )
      limits |= 16 ;		/* Victorian */
    if ( datecode > 1901 && datecode <= 1914 )
      limits |= 32 ;		/* Edwardian */
    if ( datecode >= 1914 && datecode <= 1945 )
      limits |= 64 ;		/* Wars and 30s */
    if ( datecode <= 1837 ) limits |= 256 ; /* Pre-Victorian */
    if ( datecode >= 1945 ) limits |= 512 ; /* Postwar */
  }

  if ( datecode > 0 ) {
    sprintf(op, "%4d", datecode) ;
    op += 4 ;
  }
  op += finish_field(op, obuf) ;

  /* FIELD 3 (was 1) - AUTHOR */

  /* Now has punctuation and includes all 7xx's, any 1xx comes first
     and 7xx's are parenthesised */

  mark = '\0' ;
  for ( i = 0 ; i < ntags ; i++ ) {
    if ( strcmp(wf[i].tag, "100") == 0 ) 
      if (wf[i].ind1 == '0') mark = '$' ;
      else                   mark = '|' ;
    else if ( strcmp(wf[i].tag, "110") == 0 || strcmp(wf[i].tag, "111") == 0 )
      mark = '$' ;
    if ( mark ) {
      *op++ = mark ;
      op += string_wfld(wf, op, i) ;
      break ;
    }
  }

  for ( i = 0 ; i < ntags ; i++ ) {

    if ( mark ) { *op++ = ';' ;	*op++ = ' ' ;} /* Last one will get stripped */
    mark = '\0' ;
    if ( strcmp(wf[i].tag, "700") == 0 )
      if ( wf[i].ind1 == '0') mark = '$' ;
      else mark = '|' ;
    else 
      if ( strcmp(wf[i].tag, "710") == 0 || strcmp(wf[i].tag, "711") == 0 ) 
	mark = '$' ;
    if ( mark ) {
      *op++ = '(' ; *op++ = mark ;
      op += string_wfld(wf, op, i) ;
      *op++ = ')' ;
    }
  }
  if ( op > obuf + 1 && ! mark && *(op - 1) == ' ' && *(op - 2) == ';' )
    op-= 2 ;			/* Overwrite surplus semicolon */
  op += finish_field(op, obuf) ; 

  /* field 4 (was 2) - title */
  
  a_240 = 0 ;			/* one or both of these will hold the index
				   of the appropriate tag, + 1 */
  a_245 = 0 ;

  for ( i = 0 ; i < ntags ; i++ ) {
    if      ( strcmp(wf[i].tag, "240") == 0 ) a_240 = i + 1 ;
    else if ( strcmp(wf[i].tag, "245") == 0 ) {
      a_245 = i + 1 ;
      break ;
    } /* doesn't work properly unless any 240 is before the 245 -
         they should be in this order in proper marc */
  }

  if (! a_245) {		/* diagnostics and skip to next record */
    no_245++ ;
    if ( ! silent )
      fprintf(stderr,
	    "%s no 245, skipping to next record, record follows\n%s\n\n",
	    fcno, ibuf) ;
    return(245) ;    /* return value greater than 1 ... go to next record */
  }

  if (a_240) {
    has_240++ ;
    if (wf[a_245 - 1].ind1 == '0') { /* 245 0x means no title access req.*/
      using_240++ ;
      a = a_240 - 1 ;
    }
    else a = a_245 - 1 ;
  }
  else a = a_245 - 1 ;

  *op++ = '^' ;
  op += string_dot_wfld(wf, op, a) ;

  op += finish_field(op, obuf) ;  


  /* field 5 (was 3) ... main edition and publication */

  for ( i = 0 ; i < ntags ; i++ ) {
    if ( (strcmp(wf[i].tag, "250") == 0 || strcmp(wf[i].tag, "260") == 0)
	 && wf[i].ind2 == '0' ) { /* ind2 = 0 means relates to work, not
				     part of work */
      /* shouldn't be more than one with ind2 = 0 but never mind */
      /* *op++ = '@' ;  Never been used, remove (Nov 94) */
      op += string_dot_wfld(wf, op, i) ;
      wf[i].tag[0] = ' ' ;	/* mark it used */
    }
  }

  op += finish_field(op, obuf) ;  


  /* field 5 (was 4) - series, part, titles etc... */

  /* series titles */

  for ( i = 0 ; i < ntags ; i++ ) {
    if ( (strcmp(wf[i].tag, "440") == 0 && wf[i].ind1 == '0') ||
	 ( strcmp(wf[i].tag, "490") == 0 && wf[i].ind2 == '0') ) {
      *op++ = '^' ;
      op += string_dot_wfld(wf, op, i) ; 
      wf[i].tag[0] = ' ' ;	/* mark it used */
    }
  }

  /* get part titles */

  for ( i = 0 ; i < ntags ; i++ ) {
    if ( strcmp(wf[i].tag, "248") == 0 ) {	/* found tag 248 */
      ind1  = wf[i].ind1 ;
      *op++ = '^' ;
      op += string_dot_wfld(wf, op, i) ; 
      
      for ( j = 0 ; j < ntags ; j++ ) {
	if (( strcmp(wf[j].tag, "440") == 0 && wf[j].ind1 == ind1 ) ||
	    ( strcmp(wf[j].tag, "490") == 0 && wf[j].ind2 == ind1 )) {
	  *op++ = '^' ;
	  op += string_dot_wfld(wf, op, j) ; 
	  wf[j].tag[0] = ' ' ;	/* mark it used */
	}
	else if (( strcmp(wf[j].tag, "250") == 0 ||
		   strcmp(wf[j].tag, "260") == 0) &&
		 ( wf[j].ind2 == ind1  )) {
	  *op++ = '@' ;
	  op += string_dot_wfld(wf, op, j) ; 
	  wf[j].tag[0] = ' ' ;	/* mark it used */
	  
	}
      }
    }
  }

  /* get leftovers */

  for ( i = 0 ; i < ntags ; i++ ) {
    if ( strcmp(wf[i].tag, "250") == 0 ) {
      *op++ = '@' ;
      op += string_dot_wfld(wf, op, i) ;
      leftover_tits = TRUE ;
    }
    else if ( strcmp(wf[i].tag, "440") == 0 ||
	      strcmp(wf[i].tag, "490") == 0 ) {
      *op++ = '^' ;
      op += string_dot_wfld(wf, op, i) ; 
      leftover_tits = TRUE ;
    }
/* 505s used to be used here if analytic - now see last field (contents) */
/*
      else if ( strcmp(wf[i].tag, "505") == 0 ) {
	a_505 = i + 1 ;
	if ( analyt ) {
	  *op++ = '^' ;
	  op += string_dot_wfld(wf, op, i) ; 
	}
      }
*/
  }

  op += finish_field(op, obuf) ; 

  /* field 7 (was 5) dewey number(s) and spine label */

#ifdef CITY			/* class. at 089 $a, spine label 089 $d $c */

  cc = 0 ;

  for ( i = 0 ; i < ntags ; i++ ) {
    if ( strcmp(wf[i].tag, "089") == 0 )  {
      op += string_wfld(wf, op, i) ;
      cc++ ;
      break ;
    }
  } 

  if ( ! cc ) {			/* no 089 found - look for 029 */
    cc = 0 ;
    for ( i = 0 ; i < ntags ; i++ ) {
      if ( strcmp(wf[i].tag, "029") == 0 )  {
	op += string_wfld(wf, op, i) ;
	cc++ ;
	break ;
      }
    } 
    if ( ! cc )	{		/* no 089 or 029 - do diagnostics */
      if ((diag & 2) && ! silent )
	fprintf(stderr, "%s no 089 or 029\n", fcno) ;
      no_class++ ;
    }
  }

#endif CITY

#ifdef BATH			/* class. udc at 082 $a */

  cc = 0 ;

  for ( i = 0 ; i < ntags ; i++ ) {
    if ( strcmp(wf[i].tag, "082") == 0 )  {
      op += string_wfld(wf, op, i) ;
      cc++ ;
      break ;
    }
  } 

  if ( ! cc ) {			/* no 082 found - do diagnostics */
    if ((diag & 2) && ! silent ) fprintf(stderr, "%s no 082\n", fcno) ;
    no_class++ ;
  }

#endif BATH

  op += finish_field(op, obuf) ; 

  
  /* field 8 (was 9) subject headings */

  start = op ; 

  nsh = 0 ;

  for ( i = 0 ; i < ntags ; i++ ) {
    if ( strcmp(wf[i].tag, "083") == 0 ) {
      *op++ = '^' ;
      op += string_dot_wfld(wf, op, i) ;
      feature_hdgs++ ;
      nsh++ ;
    }
  }

  for ( i = 0 ; i < ntags ; i++ ) {
    /* get 600s */
    if ( strcmp(wf[i].tag, "650") == 0 ) {
      *op++ = '^' ;
      op += string_dot_wfld(wf, op, i) ;
      topical_lcsh++ ;
      nsh++ ;
    }
    else if ( strcmp(wf[i].tag, "651") == 0 ) {
      *op++ = '^' ;
      op += string_dot_wfld(wf, op, i) ;
      geog_lcsh++ ;
      nsh++ ;
    }
    else if ( strcmp(wf[i].tag, "600") == 0 ) {
      if ( wf[i].ind1 == '0' ) *op++ = '$' ;
      else *op++ = '|' ;
      op += string_dot_wfld(wf, op, i) ;
      pn_lcsh++ ;
      nsh++ ;
    }
    else if ( strcmp(wf[i].tag, "610") == 0 || strcmp(wf[i].tag, "611") == 0 ){
      *op++ = '$' ;
      op += string_dot_wfld(wf, op, i) ;
      corp_lcsh++ ;
      nsh++ ;
    }
  }
  
  if ( nsh == 0 ) no_subj++ ;
  /* Experimental deduplication of SH's */
  if ( nsh > 0 && dedup_sh ) {
    length = op - start ;
    newlength = dedup(start, length, '^', "") ;
    if ( newlength > 0 && newlength < length ) {
      op = start + newlength ;
      dup_sh_recs++ ;
      dup_sh_saved += length - newlength ;
    }
  }
  /* */

  op += finish_field(op, obuf) ; 

  /* Field 9 - contents notes
     505 only to be used if analytic (see commented code under Field 4)
  */

  for ( i = 0 ; i < ntags ; i++ ) {
    register char * tempi ;
    if ( strcmp(wf[i].tag, "505") == 0 ) {
      a_505 = i + 1 ;
      *op++ = '^' ;
      start = op ;
      op += string_wfld(wf, op, i) ;
      /* Try to remove "Contents: " etc */
      if ( strncmpnc(start, "contents", 8) == 0 ) {
	tempi = start + 8 ;
	op = start ;
	while ( *tempi == ':' || *tempi == ' ' ) tempi++ ;
	while ( *tempi ) *op++ = *tempi++ ;
	*op = '\0' ;
      }
      break ;
    }
  }
  op += finish_field(op, obuf) ;

  sprintf(op, "%u%c%n", limits, FIELD_MARK, &length_written) ;
  op += length_written ;
  *op++ = RECORD_MARK ;
  *op = '\0' ;

  /* Stats and diagnostics */

  if ( bad_inds ) with_bad_inds++ ;
  if ( analyt && dump_analytics )
    fprintf(stderr,
	    "%s is analytic, record dump follows\n%s\n\n", fcno, ibuf) ;
  if ( leftover_tits ) with_leftover_tits++ ;
  if ( a_505 ) {
    with_505++ ;
    if ( analyt ) analyt_with_505++ ;
  }
  if ( a_505 && dump_505 )
    fprintf(stderr, "%s has 505 %s\n", fcno, wf[a_505 - 1].field) ;
  if ( leftover_tits && dump_leftovers )
    fprintf(stderr,
	    "%s has leftover series title(s) etc, record follows\n%s\n\n",
	    fcno, ibuf) ;
  if ( leading_blanks ) with_leading_blanks++ ;
  if ( feature_hdgs ) has_feature_hdgs++ ;
  if ( topical_lcsh ) has_topical_lcsh++ ;
  if ( geog_lcsh ) has_geog_lcsh++ ;
  if ( pn_lcsh ) has_pn_lcsh++ ;
  if ( corp_lcsh ) has_corp_lcsh++ ;

  if ( (diag & 8) && ! silent ) {
    if ( has_eight_bit ) with_eight_bit++ ;
    if ( has_control ) with_control++ ;
    if ( has_null ) with_null++ ;
  }

  return(1) ;
} /* process_record() */



/* next_tag */

next_tag(ip, tag)		/* gets next tag value */
     register char * ip ;
     register char * tag ;
{
  register int i ;

  for (i = 0 ; i < TAG_LEN ; i++ )  *tag++ = *ip++ ; 
  *tag = '\0' ;
}



/* get_scp */

get_scp(ip)			/* returns starting character position */
				/* ip is pointing to a directory entry */
     char * ip ;
{
  int scp ;

  ip += TAG_LEN + DATA_LEN ;	/* advance to starting character position */
  (void) sscanf(ip, "%5d", &scp) ;
  return scp ;
}
  

/* get_data_len */

get_data_len(ip)		/* Returns length of field
				   ip is pointing to a directory entry */
     char * ip ;
{
  int data_len ;

  ip += TAG_LEN ;
  (void) sscanf(ip, "%4d", &data_len) ;
  return data_len ;
}


/* check_first_indicator */

check_first_indicator(ind, tp, tag)	/* looks for indicator to match
					   corresponding tag */

     char ind ;
     int * tp ;
     char * tag ;
{
  while (*tp < TAG_OCCS) {/* search until found from current position to end */
    if (ind == tag_values[*tp][3]) {
      return(1) ;
    }
    (*tp)++ ;
    if (strncmp(tag, tag_values[*tp], 3) != 0) return(-1) ;
  }
  return(-1) ;
}

/* check_tag
   searches table of ascending tag values for current tag ...
   if found returns 1 else returns -1 */

#ifdef OLD_CHECK_TAG
check_tag(tag, tp)		/* searches tag values for matching tag */

     char * tag ;
     int * tp ;
{
  if (*tp >= TAG_OCCS) return(-1) ;
  while (*tp < TAG_OCCS && strncmp(tag, tag_values[*tp], 3) != 0) {
    if (strncmp(tag, tag_values[*tp], 3) < 0) 
      return(-1) ;
    if (strncmp(tag, tag_values[*tp], 3) > 0) {
      (*tp)++ ;
      if (*tp >= TAG_OCCS) return(-1) ;
    }
  }
  return(1) ;			/* matching tag found in table of tag values */
}  

#else

/* version starting from array base every time - is about as quick */

check_tag(tag, tp)		/* searches tag values for matching tag */
     register char * tag ;
     register int * tp ;
{
  register int t = 0 ;

  while (t < TAG_OCCS) {
    if (strncmp(tag, tag_values[t], 3) == 0 ) {
      *tp = t ;
      return(1) ;
    }
    t++ ;
  }
  return(-1) ;
}  
#endif OLD_CHECK_TAG

#ifndef USING_TRANSLATION_TABLE

/* translate_char()

   translation of characters from 8-bit mark format to ascii ... characters
   outside of the ascii character set are preceeded by '\'.
   different from gills program - read whole record into buffer and translate
   each character in turn, inserting special characters into buffer if
   necessary, don't bother with separate buffer to hold block accents */


translate_char (mchar, achar) /* convert from 8-bit to ascii */
     unsigned char * mchar ;
     char * achar ;
{
  int x = 0 ;			/* count of extra characters from accents */
  int s = 0 ;			/* count of special characters */
  char * a ;
  a = achar ;

  if (*mchar < 0x5e) {		/* punctuation, numbers and uc alphabetic */
    if (*mchar == 0x1e || *mchar == 0x1f) /* convert to ascii 0x23 and 0x24 */
      *mchar += 5 ;
    *a++ = *mchar ; 
  }
  else if (*mchar < 0x61) {	/* single dagger, eszett and sharp */
    *a++ = '\\' ;
    *a++ = *mchar ;
    x++ ;
  }
  else if (*mchar < 0x7b) {	/* lower case alphabetic */
    *a++ = *mchar ;
  }
  else if (*mchar < 0x80) {
    *a++ = '\\' ;
    *a++ = *mchar ;
    *a   = '\0' ;
    x++ ;
  }
  else if (*mchar == 0xe0) {	/* high tone */
    *a++ = '\\' ; 
    *a++ = 0xbb - 0x80 ;
    *a   = '\0' ;
    x++ ;
  }
  else if (*mchar == 0xfe) {	/* high comma centre */
    *a++ = '\\' ;
    *a++ = 0xbc - 0x80 ;
    *a   = '\0' ;
    x++ ;
  }
  else  {
    *a++ = '\\' ;
    x = subtract_backslash_and_save(mchar) ; 
    if (x) {	/* extra characters from table of special char values */
      *a++ = *mchar ;
      if (mchar[1] != '\0') *a++ = mchar[1] ;
    }
    else {	/* subtract 80 hex from character value so printable */
      *a++ = *mchar - 0x80 ; x++ ;
    }
    *a = '\0' ;
  }

  return(x) ;
}



/* subtract_backslash_and_save */

/* deals with special characters unprintable in 7-bit ascii ... for certain
   special characters (hex values in table special_vals) replaces character(s)
   with \ and character(s) from special_ch_vals */

subtract_backslash_and_save(a) /* deals with special case character conversion */
     char * a ;
{
  int x = 0, s = 0 ;

  while (s < SC_NUM ) {		/* is this character special hex value ??? */
    if (*a == special_vals[s]) break ;
    s++ ;
  }

  a++ ;

  if (s < SC_NUM) {		/* yes ... get alternative charcters from table */
    *a++ = special_ch_vals[s][0] ;
    x++ ;
    if (special_ch_vals[s][1] != ' ')  { *a++ = special_ch_vals[s][1] ; x++ ; }
    *a = '\0' ;
  }
  return(x) ;
}

#endif USING_TRANSLATION_TABLE				


/* save_subfield */

save_subfield(tag_buf, punc, sf_buf) /* appends subfield onto buffer
					containing all subfields for that tag*/

     char * tag_buf ;		/* buffer that subfields appended to */
     char * punc ;
     char * sf_buf ;		/* sub field buffer */
{
  char * temp ;
  char * tp ;

  if (*tag_buf == ' ' && punc[1] != '(') strcpy(punc, "#") ;
  if (*tag_buf == ' ') *tag_buf = '\0' ;

  strcat(tag_buf, punc) ;

  tp   = tag_buf ;
  while (*tp != '\0') tp++ ;
  temp = sf_buf ;
  while (*temp != '\0') {
    if (*temp == '#') temp++ ;	/* skip end of field mark within string */
    else  *tp++ = *temp++ ;
  }

  if (punc[1] == '(' ) *tp++ = ')' ; 
  *tp = '\0' ;
}



/* dont_index_article
   put "don't index" markers around article at start of 240, 245, 248
   (only used when no $g field), 440, 490 */

dont_index_article(tag_buf, sub_buf, ind2, tag)
     /* insert don't index markers */

     char * tag_buf ;
     char * sub_buf ;
     char ind2 ;
     char * tag ;
{
  int i ;

  if (*tag_buf == ' ') *tag_buf = '\0' ;

  if (ind2 == '1') 
    strncat(strcat(strncat(strcat(tag_buf, "_"), sub_buf, 1), "^"), sub_buf + 1 , 499) ;
  else if (ind2 == '2')
    strncat(strcat(strncat(strcat(tag_buf, "_"), sub_buf, 2), "^"), sub_buf + 2 , 498) ;
  else if (ind2 == '3')
    strncat(strcat(strncat(strcat(tag_buf, "_"), sub_buf, 3), "^"), sub_buf + 3 , 497) ;
  else if (ind2 == '4')
    strncat(strcat(strncat(strcat(tag_buf, "_"), sub_buf, 4), "^"), sub_buf + 4 , 496) ;
  else if (ind2 == '5')
    strncat(strcat(strncat(strcat(tag_buf, "_"), sub_buf, 5), "^"), sub_buf + 5 , 495) ;
  else if ( ! silent )
    fprintf(stderr,
	    "%s tag %s ind2 %c - ind2 out of range\n", fcno, tag, ind2) ;
}




/* initials 
   take initials from name: 100 (personal name), 600
   (personal name subject heading),
   700 (personal name added entry heading)
   keep hyphens (-) drop anything beginning with a lower case letter
   (de, von etc) */

initials(sub_buf, punc)		/* take initials from 100, 600, 700 */
     char * sub_buf ;		/* current position in subfield buffer */
     char * punc ;
{
  char * sp ;
  char w_delim[3] ;		/* wanted delimiter */
  char init[100], * ip ;
  char initials[100], * ptr ;
  int i ;

  ip  = init ;
  sp  = sub_buf ;
  ptr = initials ;

  *w_delim = ' ' ;
  while ( *w_delim != '#' && strncmp(w_delim, ".#", 2) != 0 && *sp != '\0') {

    ip = init ;
    i  = 0 ;

    /* init next */

    while ( *sp != ' ' && *sp != '.' && *sp != '-' && *sp != '#' &&
	    *sp != '\0' && strncmp(sp, ". ", 2) != 0 &&
	    strncmp(sp, ".#", 2) != 0 ) {
      if (i++ < 5) *ip++ = *sp++ ;
      else sp++ ;
    }
    if ( strncmp(sp, ". ", 2) == 0 || strncmp(sp, ".#", 2) == 0 ) {
      strncpy(w_delim, sp, 2) ;
      sp += 2 ;
    }
    else { *w_delim = *sp++ ; w_delim[1] = '\0' ; }
    *ip = '\0' ; 

    /* init check */
    while (*init == '\\') {	/* to get over names beginning with accent */
      if  (init[2] == '\\') *init = init[4] ;
      else                  *init = init[2] ; 
    }
    if (*init > 'z' || *init == ' ') { /* mainly lower case alphabetic */
      if (*w_delim == '#' || (strncmp(w_delim, ".#", 2) == 0)) break ;
      else continue ;        	       /* top of loop ... next name  */
    }
    *ptr++ = *init ;		/* put initial into subfield buffer */
    if (*w_delim == '-')	/* retain hyphen */
      *ptr++ = '-' ;
    else *ptr++ = ' ' ;
  }

  /* init end */
  ptr-- ; *ptr = '\0' ;
  strcpy(sub_buf, initials) ;	/* return initials to calling function */
  *punc = ' ' ; punc[1] = '\0' ; 

}

/* string_wfld
   concatenate wanted field at end of ouput string. doesn't copy end of field
   marker (shouldn't be there !!).
   returns count of number of characters copied. */

string_wfld(wf, op, a)/* string wanted field onto end of output string */
     WANTED_FIELDS wf ;
     register char * op ;
     int a ;
{
  register char * p ;
  register int count = 0 ;

  p = wf[a].field ;

  while (*p != '\0') {
    if (*p == '#') p++ ;
    else { *op++ = *p++ ; count++ ; }
  }
  wf[a].tag[0] = ' ' ; 
  *op = '\0' ;
  return(count) ;
}


/* string_dot_wfld
   same as string_wfld but adds punctuation at end */

string_dot_wfld(wf, op, a)		/* punctuate end of outputted string */
     WANTED_FIELDS wf ;
     register char * op ;
     int a ;
{
  register char * p ;
  register int count = 0 ;

  p = wf[a].field ;

  while (*p != '\0') {
    if (*p == '#') p++ ;
    else { *op++ = *p++ ; count++ ; }
  }
  wf[a].tag[0] = ' ' ; 

  if (*(op - 1) == ' ')  while (--*op == ' ') count-- ; 
  if (*op == '.' || *op == ',' || *op == '!' || *op == '?' || *op == ']' || *op == ':' ) 
    { *op++ = ' ' ; count++ ; }
  else { *op++ = '.' ; *op++ = ' ' ; count+= 2 ; }
  *op = '\0' ;
  return(count) ;
}



/* finish field */

finish_field(op, obuf)	/* field mark appended to output string */
     char * op ;
     char * obuf ;
{
  int count = 0 ;

  if      (op == obuf)       { *op++     = FIELD_MARK ; count++ ; }
  else if (*(op - 1) == ' ') *(op - 1) = FIELD_MARK ;
  else                       { *op++     = FIELD_MARK ; count++ ; }
  *op = '\0' ;
  return(count) ;
}

/*****************************************************************************

  dedup()                                                        17 Nov, 1989
                                                                 (A. Goker)
  Function which removes duplicate substrings from a string.
  Substrings are identified by a preceding "start character"
  and end with one of the characters specified in "endlist".
  "Endlist" might typically be ";:.,".

  Eg: "^Information retrieval; ^Technical processes; ^information retrieval."
                        or
      "Information retrieval; ^information retrieval; ^Technical processes."
                      both convert to 
      "Information retrieval; ^Technical processes."
  
  The function constitutes of two passes.
  Pass 1: Reads through the string and records the starting position and
          length of each substring.
  Pass 2: Starting from the last substring compares it with each preceding
          substring. If they are the same then eliminate the rightmost of 
	  the two substrings.

*****************************************************************************/


#define MAXSUBS 128		/* Maximum number of substrings to process */


dedup(s,l,startch, endlist)
     char *s;			/* pointer to string */
     int l;			/* length of string */
     char startch;		/* the character which starts the string */
     char *endlist ;		/* Null-terminated list of characters which
				   end a substring */
{

    /* .......................................................................
       "sublist" is an array of "substring"s
       "substring" is a structure holding the following details for 
       each substring:
         - start position of each substring eg. the position of the 1st letter
	 - the length of each substring

    ........................................................................*/

     struct substring {
       char *startpos;
       int slength;
     };

     struct substring   sublist[MAXSUBS];
     register char *p, ch ;
     char *endptr, *reppos1, *reppos2, *endpos ;
     int noofsubs, newslength, in_substring, last, othersubs, lastsubs;
     int noofchars, now_break;

     char *ss;
     int i;

     noofsubs=0;
     p=s;
     newslength = l;
     endptr = s + l;
     endpos = endptr;
     in_substring = FALSE;
     now_break = FALSE;

     ch = *p;
     if ( (l > 0) && ch != startch) {        /* even if there is no */
        sublist[noofsubs].startpos = p;      /* startch at start of string, */
        sublist[noofsubs].slength = 0;       /* still start a substring */ 
        in_substring = TRUE;
     }

     /* .....................................................................
	PASS 1
	Note the starting position and length of each substring before 
	comparing the substrings. Following is a summary of actions enabling
	this.

	char-type (ch)    in_substring  action
        --------------    ------------  ------
	start-char        F             start a substring, set in_substring T
	    "             T             finish substring, start next substring
	end-char          F             does not happen (not tested)
	    "             T             finish substring, set in_substring F
        others            F             continue as normal 
                                    ie. advance the pointer p (pt to next char)
	    "             T             increment length of substring
	
	Note: end-char is only test if in_substring=T

     ......................................................................*/

     while (p < endptr) {
       
       ch=*p++;
       if (ch == startch) {
	 if (in_substring) { 
	   noofsubs++;		/* finish substring */
	   if (noofsubs >= MAXSUBS) /* exit if too many substrings */
	     return (-1);
         }
	 sublist[noofsubs].startpos = p; /* start new substring at p */
	 sublist[noofsubs].slength = 0;
	 in_substring = TRUE;
       }

       else if ( isintab(ch, endlist) && in_substring ) {
	   noofsubs++;		/* finish substring */
	   if (noofsubs >= MAXSUBS) /* exit if too many substrings */
	     return (-1);
	   in_substring = FALSE;
	   }
		
       else if (!in_substring);	/* do nothing */
            else sublist[noofsubs].slength++;
    
     }

     if (in_substring) {
       noofsubs++;
       if (noofsubs >= MAXSUBS)	/* exit if too many substrings */
	 return (-1);
     }

     /* ....................................................................

	PASS 2
	Starting with last substring, compare it with the preceding 
	substring. If the two substrings are identical (irrespective of
	upper/lower case) then remove the rightmost (string currently 
	pointing to) of the substrings. Do this by rewriting the portion of
	the original string "s" to the right of the current substring on
	top of the current substring.

	Eg: If string "s" constitutes of substrings s1,s2,s3,s4,s5 AND
	    current substring is s4                                AND
	    s4 is identical to s2                                  THEN
	    replace s5 over s4 inorder to get the new string s=s1,s2,s3,s5
       
	Repeat the above with substring before the current substring, 
	making that the current substring, until no more substrings to
	compare.

     .....................................................................*/
 

     last = noofsubs-1;         /* index to last substring */
     for (lastsubs=last; lastsubs >= 1; lastsubs--) {
       for (othersubs = lastsubs-1; othersubs >= 0; othersubs--) {
	 now_break = FALSE;
	 if (sublist[lastsubs].slength == sublist[othersubs].slength) {
	   if ( strncmpnc (sublist[lastsubs].startpos,
			   sublist[othersubs].startpos, 
			   sublist[lastsubs].slength) == 0 ) {
	     reppos1 = sublist[lastsubs - 1].startpos +
	               sublist[lastsubs - 1].slength;
	     reppos2 = sublist[lastsubs].startpos +
	               sublist[lastsubs].slength;
	     noofchars = endpos - reppos2;
	     repover(reppos1, reppos2, noofchars);
	     newslength -= reppos2 - reppos1;
	     endpos = s + newslength;
	     now_break=TRUE;
	     break;
	   }
	   if (now_break == TRUE) break;

	 }
	 if (now_break == TRUE) break;
       }
     }

     return (newslength);

}

repover (s1,s2,length)  	/* function replaces string s2 over s1. */
     register char *s1;		/* s1 and s2 actually point to different */
     register char *s2;		/* parts of the same string s. s2 should */
				/* point somewhere to the right of s1. */
     register int  length;	/* length is the no of chars to replace */
{

  while (length-- > 0) 
    *s1++ = *s2++ ;

}

/*****************************************************************************/

isintab(ch, table)         /* Returns 0 if ch not in terminated string  */ 
                           /* table, otherwise position ( 1.. ) of ch in tab */
				/* Bug: null character cannot be found */
				/* Correction Jan 1990 to handle empty table */
     register char ch, *table ;
{
  register int i = 1 ; 

  while ( ch != *table ) { 
    if ( *table == '\0' ) return(0) ;
    table++ ;
    i++ ; 
  }
  return(i) ;
}

/*****************************************************************************

  strncmpnc(s1, s2, n)  Byte comparison as strncmp but case independent

*****************************************************************************/

strncmpnc(s1, s2, n)		/* As strncmp but case-independent */
     register char *s1, *s2 ;
     register int n ;
{
  register int diff = 0, c1, c2 ;
  while ( n-- ) {
    if ( isupper( c1 = *s1++ ) ) c1 = tolower(c1) ;
    if ( isupper( c2 = *s2++ ) ) c2 = tolower(c2) ;
    if ( diff = c1 - c2 ) break ;
  }
  return(diff) ;
}

