/***************************************************************************

  This program is more or less as written by Julie Porteous in 1988, a
  translation from Gill Venner's DEC COBOL program of 1983 (and earlier).
  It is unlikely to be used again in this form. SW Jan 1990.

  Tape Conversion Program                        Julie Porteous  June 1988

  Program for conversion of tapes from eight-bit MARC format to ascii okapi
  exchange format.
  Translated from Gill Venner's 1985 COBOL program. Record processing is more
  or less identical to that of the original program.

  Main difference is in the assembly of individual records from segments.
  Translation of characters from 8-bit to ascii is done for subfields as they
  are assembled into a field buffer, not for each block as was the case in the
  earlier version.

  To get stderr to go to file and still get interactive display on screen run
  using the following command line:-

  marmot% ( tape > /dev/tty) >& stderr_filename

****************************************************************************/

#include <ctype.h>
#include <stdio.h>

#define TEMP_LEN  9
#define SCW_LEN   5
#define RECSIZE   6000
#define TAG_OCCS  36		/* tag fields 0 - 36 */
#define SUBF_LEN  1024	        /* over 1024 and marc subfield is truncated */
#define FIELD_LEN 1024
#define MAX_SUBF  50
#define S_HDR_LEN 1024		/* maximum length of subject heading (field 9) */
#define SC_NUM    14		/* number of special characters */
#define DIRECTORY 12
#define TAG_LEN   3
#define DATA_LEN  4
#define SCP_LEN   5
#define DIAG_LEN  70		/* max length of diagnostics record */
#define HASH      0x1E		/* end of field in marc format (5 less than okapi) */
#define DOLLAR    0x1F		/* subfield marker ... */
#define TRUE      1		/* defined here since no stdeqs.h */
#define FALSE     0

#define INPUTFILEMSG "enter eight-bit filename: "
#define TMPFILEMSG   "enter output filename stem: "
/* Above are hangovers from DEC COBOL */


typedef struct {		/* store subfields for tags found in tags table */
  char tag[4] ;
  char ind1 ;
  char ind2 ;
  char field[SUBF_LEN] ;
} WANTED_FIELDS[MAX_SUBF] ;
  
typedef struct{			/* store subject headings */
  char mark ;
  char buf[FIELD_LEN] ;
  char punc[4] ;
} SUB_HDR[30] ;		/* no particular reason for size ... comes from Gill's program ... */

SUB_HDR sh ;

char next_year[] = "1990" ;	/* Do you need function to get 'current' next year ?? */

char * tag_values[] = {		/* table of tags to be selected from MARC record */
  "082 a # b$##c6",		/* with indicators, subfields and required punc */
  "083 a. #",
  "1000a # e # f, #",
  "1001a # h2",
  "1002a # h2",
  "1003a # h2",
  "110 a # c. #e (#",
  "111 a # c. #e (#",
  "240 a3#",
  "245 a3  b : i; #j. #k = l. #",
  "248 g1  h3  i; #j. #",
  "250 a # e, #",
  "260 b / c, #",
  "440 a3",
  "490 a #",
  "505 a7",
  "6000a # e # f, #",
  "6001a # h2",
  "6002a # h2",
  "6003a # h2",
  "610 a # c. #e (#",
  "611 a # c. #e (#",
  "650 a. #x--#y--#z--#",
  "651 a. #x--#y--#z--#",
  "7000a # e # f, #",
  "7001a # h2",
  "7002a # h2",
  "7003a # h2",
  "710 a # c. #e (#",
  "711 a # c. #e (#",
  "960 a # b # p#",
  "961 a # b # p#",
  "962 a # b # p#",
  "964 a # b # p#",
  "965 a # b # p#",
  "966 a # b # p#",
  "970 c # "
  } ;
  
int special_vals[] = { 66, 68, 70, 72, 74, 76, 98, 100, 102, 104, 106, 108, 112, 116 } ;

char * special_ch_vals[] =  {
  "L ",
  "O ",
  "D ",
  "TH",
  "AE",
  "OE",
  "l ",
  "o ",
  "d ",
  "th",
  "ae",
  "oe",
  "i ",
  "th"
  } ;
  

static long record_count = 0 ;

char fcno[20] ;		/* store curent record control number */
char DREC[DIAG_LEN] ;

char rec_buf[RECSIZE] ;	/* max rec size is 6000 but average 4000 chars */


/***************************************************************************

  convert_tape()

  Gets run time parameters and open files. While there are more records get
  next record and process.

  Reads next record from 8-bit marc file, selects required tags and subfields,
  using table of tag_values.
  For each selected tag, translates the subfield to ascii, and adds appropriate
  punctuation after performing any exception processing (given by numeric
  "punctuation" in the tags table).
  Subfields are assembled and the whole field stored in struct of type
  WANTED_FIELDS.
  Fields within each record are separated by '#', empty fields are left blank,
  and the markers ^ _ | $ @ are inserted thus:

  ^ indicates title like data
  | indicates personal name
  $ indicates corporate name
  _ indicates no indexing required
  @ indicates editor/publisher data

*****************************************************************************/

main(argc, argv)
     int argc ;
     char * argv[] ;
{
  FILE * in_fp, * out_fp ;
  char temp[TEMP_LEN] ;
  char out_buf[RECSIZE] ;
  char out_name[30] ;
  char in_name[30] ;
  int rec_len = 0 ;
  int records = TRUE ;

  /* get input file name and open */
  printf("%s ", INPUTFILEMSG) ; scanf("%s", in_name) ;
  if ( (in_fp = fopen(in_name, "r")) == NULL) {
    printf("can't open file: %s\n", in_name) ;
    exit() ;
  }

  /* output file name is stem and TMP */
  printf("%s", TMPFILEMSG) ; scanf("%s", temp) ;
  strcat(strcat(out_name, temp), "TMP") ;   /* open output file */
  out_fp = fopen(out_name, "w") ;

  /* get next record and process */

  while (records) {
    *rec_buf = '\0' ;
    /* get next record ... break if at end of file */
    if ( (rec_len = get_next_record(in_fp, rec_buf)) == -1 ) break ;
    /* process next record */
    records = process_record(rec_buf, out_buf) ;
    /* write output record */
    if (records == 1) {
      /* e.g. no tag "008" would return 8 etc... */
      fprintf(out_fp, "%s\n", out_buf) ;
    }
/*    printf("obuf = '%s'\n", out_buf) ;
/*    if ( (record_count % 300) == 0) {	/* print every 1000th record */
/*      fprintf(stderr, "obuf = '%s'\n", out_buf) ;
    } 
*/
    record_count++ ;

    fflush(stderr) ;
    fflush(out_fp) ;
  } /* while more records */
  printf("Total number of records processed = %d\n", record_count) ;
}





/* get_next_record  ... 1 jun 88 

   Reads next record from file and leaves file pointer at start of next
   records segment control word.
   Returns record_length on successful execution else returns -1. */

  
get_next_record(fp, rec_buf)	/* get next logical record from file */
     FILE * fp ;
     char * rec_buf ;
{
  char scw[SCW_LEN] ;
  char span ;
  char * p ;
  char * s_len = scw + 1 ;
  int slen ;
  int i ;
  int rlen = 0 ;

  if (next_scw(fp, scw) == -1) return(-1) ; /* EOF */
      
  span = scw[0] ;
  slen = atoi(s_len) - SCW_LEN ;

  /* read next segment into buffer */
  for (i = 0, p = rec_buf ; i < slen ; i++, p++, rlen++ ) *p = getc(fp) ;
     
  /* spanning indicator '1' : record begins but doesn't end in this segment 
     spanning indicator '2' : record neither begins or ends in this segment */
  while (span == '1' || span == '2') {
    if (next_scw(fp, scw) == -1) return(-1) ; /* EOF */
    span = scw[0] ;
    slen = atoi(s_len) - SCW_LEN ;
    /* read segment and append into rec_buf at p */
    for (i = 0 ; i < slen ; i++, p++, rlen++ ) *p = getc(fp) ;
    if (span != '2') break ;
  }
  return(rlen) ;
}
  
    


/* next_scw */
  
next_scw(fp, scw)		/* get next segment control word */
     FILE * fp ;
     char * scw ;
{
  char * p ;
  int c ;
  int i ;

  for (i = 0, p = scw ; i < SCW_LEN ; i++, p++ ) {
    *p = getc(fp) ;
    if (! i) {			/* check whether spanning indicator is valid */
      p = scw ;
      while ((*p < '0' || *p > '3') && *p != EOF) *p = getc(fp) ;
    }
    if (*p == EOF) return(-1) ;
  }
  return(0) ;
}








/* process_record ... 2 june 88 

   Selects required tags and subfields using table of tag values. Adds
   appropriate punctuation after performing any exception processing which
   is given by numeric "punctuation" in the tags table.
   If error on record returns value greater than 1 and skips to next record.   */


process_record(ibuf, obuf)	/* marc record into okapi exchange format */
     char * ibuf ;
     char * obuf ;
{
  WANTED_FIELDS wf ;

  char * ip ;
  char * op ;
  char * t, tag[3] ;
  char * b ;
  char tag_buf[FIELD_LEN] ;
  char * sp ;
  char sub_buf[SUBF_LEN] ;
  char buf[10] ;
  char wanted_delim[3] ;
  char * data_ptr ;
  char * temp ;
  char temp_buf[SUBF_LEN] ;
  char buf_082[10][40] ;	/* dimensions from Gill's program - store for dewey no. */
				/* N.B. assumes no dewey no. more than 30 chars */
  char mark ;
  char * this_tags_data ;
  char lang[4] ;
  char date1[5] ;
  char DCed[3] ;		/* save DC edition information for tag "082c" */
  char punc[4] ;
  char site_code[3] ;
  char ind1, ind2 ;
  char * sf ;
  char * tag_ptr ;

  int a505 = FALSE ;
  int translated  = FALSE ;
  long sd_oset ;		/* offset of sub-directory in record  */
  long scp ;			/* starting character position relative to base */
  long base_data_addr ;		/* base address of data - relative to start of  */
				/* record (record label) */
  int datecode = 0 ;
  int site  = 0 ;
  int site1 = 0 ;
  int a_240 = 0 ;
  int a_245 = 0 ;
  int count = 0 ;
  int cc = 0 ;
  int tp = 0 ;
  int a  = 0 ;
  int i ;
  int analyt ;

  *tag_buf = ' ' ;

  /* get base address from record label then skip to first directory */
  ip = ibuf + 12 ;
  for (i = 0, b = buf ; i < 5 ; i++ ) *b++ = *ip++  ; *b = '\0' ;
  base_data_addr = atoi(buf) ;
  ip  += 7 ;			/* skip the rest of the record label */

  /* getrecord control number from "001" tag */

  next_tag(ip, tag) ;
  get_scp(ip, &scp) ;
  this_tags_data = ibuf + base_data_addr ;
  temp = fcno ;
  while (*this_tags_data != HASH) *temp++ = *this_tags_data++ ; *temp = '\0' ;
  ip += DIRECTORY ;
  next_tag(ip, tag) ;

  /* check whether there is an analytical subdirectory present */
  if (strcmp(tag, "002") == 0) {
    get_scp(ip, &scp) ;		/* advances ip to start of next directory entry */

    sd_oset = scp + base_data_addr + 1 ;	/* offset of sub-directory in record */

    data_ptr = ibuf + sd_oset + 7 ;	/* advance to field containing sub-directory scp */
    for (i = 0, b = buf ; i < SCP_LEN ; i++ ) *b++ = *data_ptr++ ; *b = '\0' ;

    analyt = (atoi(buf) / DIRECTORY ) - 1 ; /* get number of fields before subfield */

    ip += DIRECTORY ;
    next_tag(ip, tag) ;		/* next tag should be "008" */
  }

  if (strcmp(tag, "008") != 0) {/* bad record ... return 8 and get next record */
    printf("No tag 008 for record fcno = %s - skipping to next record\n", fcno ) ;
    fprintf(stderr, "tag 008 missing fcno (%s) tag (%s)\nmark record = %s\n\n",
	    fcno, tag, rec_buf) ;
    return(8) ;			/* return value > 1 so don't write record to file */
  }

  
  /* process "008" tag */
  get_scp(ip, &scp) ;

  data_ptr = ibuf + base_data_addr + scp + 36 - 1 ; /* advance to language code */
  /* get language code from positions 36 - 38 in "008" field */
  for (i = 0, b = lang ; i < 3 ; i++ ) *b++ = *data_ptr++ ; *b = '\0' ;
  
  data_ptr = ibuf + base_data_addr + scp + 8 - 1  ; /* advance to date1 code */
  /* get date1 from positions 8 - 11 */
  for (i = 0, b = date1 ; i < 4 ; i++ ) *b++ = *data_ptr++ ;  *b = '\0' ;

  /* perform dte */
  if ( atoi(date1) < 1750 )
    fprintf(stderr, "date (%s) < 1750 ... fcno (%s)\nmark record = %s\n",
	    date1, fcno, rec_buf) ;
  else if ( strcmp(date1, next_year) > 0 )
    fprintf(stderr, "date (%s) > next_year(%s) ... fcno (%s)\nmark record = %s\n",
	    date1, next_year, fcno, rec_buf) ;
  else
    datecode = atoi(date1) - 1750 ;

  /* now process tags until end of record */

  ip += 12 ;
  next_tag(ip, tag) ;

  *wanted_delim = ' ' ;

  while (*tag != HASH && *tag != '#') {	/* '#' is the end of record marker */

    if (*wanted_delim == DOLLAR || *wanted_delim == '$' )
      sf = this_tags_data++ ;		/* skip to get-subfield */
    else {
      sp = sub_buf ;   	/* reset pointer for next subfield */

      if (check_tag(tag, &tp) == -1) { /* tag not found ... go to next tag */
	ip += DIRECTORY ;
	next_tag(ip, tag) ;
	continue ;
      }
      get_scp(ip, &scp) ;
      
      this_tags_data = ibuf + base_data_addr + scp ;

      ind1 = *this_tags_data++ ;
      ind2 = *this_tags_data++ ;

      if (tag_values[tp][3] != ' ') 
	if (check_first_indicator(ind1, &tp, tag) == -1) {
	  next_tag(ip, tag) ;
	  continue ; 
	}

      /* this_tag */
      this_tags_data++ ;
      sf = this_tags_data++ ;
    }


    /* get subfield */
    cc = 0 ;
    
    while (cc < SUBF_LEN) {
      temp  = temp_buf ;
      count = translate_char(this_tags_data, temp) ;
      if (*temp_buf == '#' || *temp_buf == '$') {
	*wanted_delim = *temp_buf ;
	this_tags_data++ ;
	break ;
      }
      if (count) {
	while (*temp != '\0') *sp++ = *temp++ ; 
	this_tags_data += count ;
      }
      else {
	*sp++ = *temp ; 
	this_tags_data++ ;
      }
      cc++ ;
    }
  
    if (cc)  { *sp++ = '#' ; *sp++ = '\0' ; }

    sp = sub_buf ;
    if (cc)  {
      tag_ptr  = tag_values[tp] + 4 ;	/* character after tag and first indicator */
      
      while (*sf != *tag_ptr && tp < TAG_OCCS && *tag_ptr != '\0') {
	tag_ptr += 4 ; 
	if (*tag_ptr != ' ') continue ;
	else break ; 
      }
      if (*sf == *tag_ptr) { /* if 1st char of subfield found in tag_values */
	tag_ptr++ ;
	strncpy(punc, tag_ptr, 3) ; /* get punctuation from tag tables */

	if (! isdigit(*punc)) {
	  save_subfield(tag_buf, punc, sub_buf) ;
	}
	else if (*punc == '3') { /* may have article */
	  if (ind2 == '0') {
	    if (strcmp(tag, "248") == 0) strcpy(punc, ": ^") ;
	    else              strncpy(punc, "$", 1) ;
	    save_subfield(tag_buf, punc, sub_buf) ;
	  }
	  else {
	    dont_index_article(tag_buf, sub_buf, ind2, tag) ;
	  }
	}
	else if (*punc == '2') { /* subfield is 100h */
	  initials(sub_buf, punc) ;
	  save_subfield(tag_buf, punc, sub_buf) ;
	}
	else if (*punc == '1') { /* subfield is 248g */
	  /* dont_index_248g */
	  if (*tag_buf == ' ') *tag_buf = '\0'  ;
	  strcat(tag_buf, "_") ;
	  
	  temp = tag_buf ; while (*temp != '\0') temp++ ;
	  b   = sub_buf ;
	  while (*b != '\0') {
	    if (*b == '#') b++ ;	/* skip end of field mark within string */
	    else  *temp++ = *b++ ;
	  }
	  *temp = '\0' ;
	}
	else if (*punc == '6') { /* subfield is 082c */
	  /* look_for_DC_edition */
	  DCed[0] = sub_buf[0] ;
	  DCed[1] = sub_buf[1] ;
	  DCed[3] = '\0' ;
	}
	else if (*punc == '7') { /* tag is 505 */
	  a505 = TRUE ;
	  save_subfield(tag_buf, punc, sub_buf) ;
	}
	else {			/* unknown punc ... skip to next record */
	  printf("fcno (%s) unknown punc (%s) - skipping to next record\n", fcno, punc) ; 
	  return(7) ;		/* return value greater than 1 ... go to next record */
	}
      } /* check subfield */
    } /* if (cc) ... ones that aren't skip to next_subfield */
    
    if (*wanted_delim == '$') 
      continue ;				/* skip to get_subfield */
    else {
      if (*tag_buf != ' ') {
	a++ ;
	strcpy(wf[a].tag, tag) ;
	wf[a].ind1 = ind1 ;
	wf[a].ind2 = ind2 ;
	
	strcpy(wf[a].field, tag_buf) ;
	strcat(wf[a].field, "#") ; 
	*tag_buf = ' ' ;	/* re-initialise tag buffer */

	while (tp && strncmp(tag_values[tp], tag, 3) == 0 ) if (tp) tp-- ;

      }
      ip += DIRECTORY ;
      next_tag(ip, tag) ;
      
    }
  }

  
  /* end of record ... assemble fields and output them */

  a++ ;				/* wf subscript */
  strcpy(wf[a].tag, "999") ;

  *obuf = '\0' ;
  op = obuf ;

  

  /* FIELD 1 - AUTHOR */

  a = 0 ;
  while ( strcmp(wf[a].tag, "111") <= 0 ) {
    temp = wf[a].field ;
    if ( strcmp(wf[a].tag, "100") == 0 ) {
      if (wf[a].ind1 == '0') *op++ = '$' ;
      else                   *op++ = '|' ;
      op += string_wfld(wf, op, a) ;
    }
    else if ( strcmp(wf[a].tag, "110") == 0 || strcmp(wf[a].tag, "111") == 0 ) {
      *op++ = '$' ;
      op += string_wfld(wf, op, a) ;
    }
    a++ ;
  }
  op += finish_field(op, obuf) ; 


  /* FIELD 2 - TITLE */
  
  a_240 = 0 ;
  a_245 = 0 ;
  a     = 0 ;

  while ( strcmp(wf[a].tag, "245") <= 0 )  {
    if      ( strcmp(wf[a].tag, "240") == 0 ) a_240 = a ;
    else if ( strcmp(wf[a].tag, "245") == 0 ) a_245 = a ;
    a++ ;
  }

  if (! a_245) {		/* DIAGNOSTICS and skip to next record */
    diag(" no 245") ;
    fprintf(stderr, "%s\nmark record = %s\n\n", DREC, rec_buf) ;
    return(245) ;		/* return value greater than 1 ... go to next record */
  }

  if (a_240) {
    if (wf[a].ind1 == '0') wf[a_245].tag[0] = ' ' ;
    else                   wf[a_240].tag[0] = ' ' ;
  }
  a = 0 ;
  while  ( strcmp(wf[a].tag, "245") <= 0 ) {
    temp = wf[a].field ;
    if ( strcmp(wf[a].tag, "245") == 0 || strcmp(wf[a].tag, "240") == 0 ) {
      *op++ = '^' ;
      op += string_dot_wfld(wf, op, a) ;
    }
    a++ ;
  }
  op += finish_field(op, obuf) ;  


  /* FIELD 3 ... MAIN ED_PUB */

  while ( strcmp(wf[a].tag, "260") <= 0 ) {
    if (wf[a].tag[0] != ' ') {
      if ( (strcmp(wf[a].tag, "250") == 0 || strcmp(wf[a].tag, "260") == 0)
	  && wf[a].ind2 == '0' ) {
	    temp = wf[a].field ;
	    *op++ = '@' ;
	    op += string_dot_wfld(wf, op, a) ;
	  }
    }
    a++ ;
  }
  op += finish_field(op, obuf) ;  


  /* FIELD 4 - SERIES, PART, TITLES ETC... */

  while ( strcmp(wf[a].tag, "490") <= 0 ) {
    if ( wf[a].tag[0] != ' ' ) {
      if ( (strcmp(wf[a].tag, "440") == 0 && wf[a].ind1 == '0')
	  ||
	  ( strcmp(wf[a].tag, "490") == 0 && wf[a].ind2 == '0') ) {
	    *op++ = '^' ;
	    op += string_dot_wfld(wf, op, a) ; 
	  }
    }
    a++ ;
  }

  /* get 248 etc */
  a = 0 ;
  while ( strcmp(wf[a].tag, "999") != 0 ) { /* search through all tags in wf */

    if ( strcmp(wf[a].tag, "248") == 0 ) {	/* found tag 248 */
      ind1  = wf[a].ind1 ;
      *op++ = '^' ;
      op += string_dot_wfld(wf, op, a) ; 
      
      i = a ;			/* temporarily save value of 'a' */
      while ( strcmp(wf[a].tag, "999") != 0 ) { /* search the remaining tags */
	/* get 248 remainder */
	if (wf[a].tag[0] != ' ') {
	  if      (( strcmp(wf[a].tag, "440") == 0 && wf[a].ind1 == ind1 ) ||
                   ( strcmp(wf[a].tag, "490") == 0 && wf[a].ind2 == ind1 )) {
		     *op++ = '^' ;
		     op += string_dot_wfld(wf, op, a) ; 
		   }
	  else if (( strcmp(wf[a].tag, "250") == 0 || strcmp(wf[a].tag, "260") == 0) &&
		   ( wf[a].ind2 == ind1  )) {
		     *op++ = '@' ;
		     op += string_dot_wfld(wf, op, a) ; 
		   }
	}
	a++ ;
      }
      a = i ;			/* restore value of 'a' */
    }
    a++ ;
  }

  /* get leftovers */
  a = 0 ;
  while ( strcmp(wf[a].tag, "999") != 0 ) {
    if (wf[a].tag[0] != ' ') {
      if ( strcmp(wf[a].tag, "250") == 0 ) {
	diag_tag(wf, a) ;
	*op++ = '@' ;
	op += string_dot_wfld(wf, op, a) ; /* punctuate end of outputted string */
      }
      else if ( strcmp(wf[a].tag, "440") == 0 || strcmp(wf[a].tag, "490") == 0 ) {
	diag_tag(wf, a) ;
	*op++ = '^' ;
	op += string_dot_wfld(wf, op, a) ; 
      }
      else if ( strcmp(wf[a].tag, "550") == 0 && analyt) {
	*op++ = '^' ;
	op += string_dot_wfld(wf, op, a) ; 
      }
    }
    a++ ;
  }
  op += finish_field(op, obuf) ; 


  /* FIELD (5) ADDED NAMES */

  a = 0 ;
  while ( strcmp(wf[a].tag, "999") != 0 ) {
    /* get 7-xxx */
    if ( wf[a].tag[0] != ' ' ) {
      if ( strcmp(wf[a].tag, "700") == 0 )
	if ( wf[a].ind1 == '0') {
	  *op++ = '$' ;
	  op += string_wfld(wf, op, a) ;
	}
	else {
	  *op++ = '|' ;
	  op += string_wfld(wf, op, a) ;
	}
      else 
	if ( strcmp(wf[a].tag, "710") == 0 || strcmp(wf[a].tag, "711") == 0 ) {
	  *op++ = '$' ;
	  op += string_wfld(wf, op, a) ;
	}
    }
    a++ ;
  } /* for all tags */

  op += finish_field(op, obuf) ; 

  
  /* FIELD (6) DEWEY NUMBER(S) and SPINE LABEL */
  /* tag 082 is the Dewey decimal classification number (current edition). */

  cc = 0 ;	/* cc = count of 082 tags found */
  a  = 0 ;
  /* perform get first 082 */
  while ( strcmp(wf[a].tag, "082") <= 0 && ! cc ) {
    if ( strcmp(wf[a].tag, "082") == 0 )  {
      op += string_wfld(wf, op, a) ;
      wf[a].tag[0] = ' ' ; 
      strcpy(buf_082[cc], wf[a].field) ;
     cc++ ;
    }
    a++ ;
  } /* ... get first 082 */

  /* get 970 */
  a = 0 ;
  while ( strcmp(wf[a].tag, "970") <= 0 ) {
    if ( strcmp(wf[a].tag, "970") == 0 ) {
      *op++ = ' ' ;
      op += string_wfld(wf, op, a) ;
    }
    a++ ;
  }

  if ( ! cc ) {			/* no 082s found - do diagnostics */
    diag(" no 082") ;
    fprintf(stderr, "%s\nmark record = %s\n\n", DREC, rec_buf) ;
    return(82) ; 		/* go to next record */
  }
  else {			/* get other 082 tags */
    a = 0 ;
    while ( strcmp(wf[a].tag, "082") <= 0 ) {
      /* get other_82s */
      /* ignore duplicate dewey numbers for PCL */

      if ( strcmp(wf[a].tag, "082") == 0 ) {
	i = 0 ;
	while ( wf[i].tag[0] != ' ' &&  i < cc ) {
	  /* check 082 */
	  if ( strcmp(wf[a].field, buf_082[i]) ) wf[a].tag[0] = ' ' ;
	  i++ ;
	}
	if ( strcmp(wf[a].tag, "082") == 0 ) {
	    cc++ ;
	    strcpy(buf_082[cc], wf[i].field) ;
	    op += string_wfld(wf, op, i) ;
	  }
      }
      a++ ;
    } /* ... while tag not greater than 082 */
  } 
  op += finish_field(op, obuf) ; 

  
  /* FIELD (7) ACCESSION NUMBERS */
  /* very PCL specific ... for Charles' tapes etc should leave blank field */

  a = 0 ;

  /* perform get-96x */

  while ( strcmp(wf[a].tag, "969") <= 0 ) {
    if ( strncmp(wf[a].tag, "96", 2) == 0 ) { /* tag stem comparison */
      /* get site code */
      if      ( wf[a].tag[2] == '0' ) site++ ;
      else if ( wf[a].tag[2] == '1' ) site1 = 2 ;
      else if ( wf[a].tag[2] == '2' ) site += 4 ;
      else if ( wf[a].tag[2] == '4' ) site += 8 ;
      else if ( wf[a].tag[2] == '5' ) site += 16 ;
      else if ( wf[a].tag[2] == '6' ) site1 = 2 ;

      temp = wf[a].field ;	/* concatenate field and ' ' at end of output  */
      while (*temp != '\0') {
	if (*temp == '#') temp++ ;
	else *op++ = *temp++ ;
      }
      *op++ = ' ' ;
    }
    a++ ;
  }
  op += finish_field(op, obuf) ;

  
  /* FIELD (8) STE code, DATE code, DC edition, (from 082$c), LANG code, CONTROL no. */

  site += site1 ;

  if (site < 10) *op++ = '0' ;	/* leading zero if site code is single digit */

  sprintf(op, "%d", site) ; 
  while (*op != '\0') op++ ;  
  if (! datecode) { *op++ = '0' ; *op++ = '0' ; } /* expects 3 digit field */
  sprintf(op, "%d", datecode) ;
  while (*op != '\0') op++ ;  *op = '\0' ;
  strcat(strcat(strcat(op, lang), DCed), fcno) ;
  while (*op != '\0' && *op != '#') op++ ; *op = '\0' ;
  
  op += finish_field(op, obuf) ;
  

  /* FIELD (9) SUBJECT HEADINGS */

  count = 0 ;

  /* for each tag field "651", "650", "600", "610", "611" split seperate subject
     headings and punctuation
     Assemble into array of structs type SUB_HDR ... global variable */
  a = 0 ;
  while ( strcmp(wf[a].tag, "651") <= 0 ) {
    if ( strcmp(wf[a].tag, "650") == 0 || strcmp(wf[a].tag, "651") == 0 ) {
      mark = '^' ;
      subj(wf, wanted_delim, &count, a, mark) ; 
    }
    else {
      if ( strcmp(wf[a].tag, "600") == 0 )
	if ( wf[a].ind1 == '0' ) {
	  mark = '$' ;
	  subj(wf, wanted_delim, &count, a, mark) ; 
	}
	else {
	  mark = '|' ;
	  subj(wf, wanted_delim, &count, a, mark) ; 
	}
      else if ( strcmp(wf[a].tag, "610") == 0 || strcmp(wf[a].tag, "611") == 0 ) {
	mark = '$' ;
	subj(wf, wanted_delim, &count, a, mark) ;  
      }
    }
    a++ ;
  } /* ... while < "651" */

  /* process subject headings for tag field "083" ... into array of struct type SUB_HDR */
  a = 0 ;
  while ( strcmp(wf[a].tag, "083") <= 0 ) {
    if ( strcmp(wf[a].tag, "083")  == 0 ) {
      mark = '^' ;
      subj(wf, wanted_delim, &count, a, mark) ;
    }
    a++ ;
  }

  /* concatenate subjectheadingrs at end of output buffer */
  a = 0 ;
  while ( a < count ) {		/* count = no. subject headings found */
    strcpy(punc, sh[a].punc) ;
    /* do something with "--" ... change to " - " */
    if ( *punc == '-' )   strcpy(punc, " - ") ; 
    else
      if ( *punc == ',' ) strcpy(punc, ", #") ;
      else                strcpy(punc, ". #") ;
    /* string subject headings into output buffer */
    if ( sh[a].mark != '#' )	/* ignore mark character if end of field */
      *op++ = sh[a].mark ;
    temp = sh[a].buf ;		/* string subject heading itself - ignore '#' */
    while ( *temp != '\0' ) { if (*temp == '#') temp++ ; else *op++ = *temp++ ; } *op = '\0' ;
    temp = punc ;		/* string punctuation - break and ignore '#' */
    while ( *temp != '#' && *temp != '\0' ) *op++ = *temp++ ; *op = '\0' ;
    a++ ;
  } /* concatenate subject headings */
  
  op += finish_field(op, obuf) ;


  /* FIELD (10) CONTENTS NOTE (non-analytics) */
  /* COBOL prog had call to get-505 and finish field before check-whats-left
     if analyt = 0 perform get-505 varying a from1 by 1 until wtag(a) = "999"
     finish-field */
  
  count = FALSE ;

  /* check what is left */
  a = 0 ;
  while ( strcmp(wf[a].tag, "999") < 0 ) {
    if ( wf[a].tag[0] != ' ' ) {
      if ( (strcmp(wf[a].tag, "100") == 0 ||
	    strcmp(wf[a].tag, "110") == 0 ||
	    strcmp(wf[a].tag, "111") == 0 ||
	    strcmp(wf[a].tag, "240") == 0 ||
	    strcmp(wf[a].tag, "245") == 0  ) && analyt )
	count = TRUE ;
      else
	diag_tag(wf, a) ;
    }
    a++ ;
  }
  
  if (count) {
    fprintf(stderr, "fcno (%s) is analytic", fcno) ;
    if (a505) fprintf(stderr, " with 505\n") ;
    else      fprintf(stderr, "\n") ;
    fprintf(stderr, "%s\n\n", DREC) ;
  }

  return(1) ;
} /* process_record() */







/* next_tag */

next_tag(ip, tag)		/* gets next tag value */
     char * ip ;
     char * tag ;
{
  int i ;
  char * t ;
  char * p ;

  t = tag ; p = ip ;
  for (i = 0 ; i < TAG_LEN ; i++ )  *t++ = *p++ ; 
  *t = '\0' ;
}



/* get_scp */

get_scp(ip, scp)		/* returns starting character position */
     char * ip ;
     long * scp ;
{
  int i ;
  char buf[5] , * b ;

  b = buf ;
  ip += TAG_LEN + DATA_LEN ;	/* advance to starting character position */

  for (i = 0 ; i < SCP_LEN ; i++ ) *b++ = *ip++ ;
  *b  = '\0' ;

  *scp = atoi(buf) ;
}
  



/* check_first_indicator */

check_first_indicator(ind, tp, tag)	/* looks for indicator to match corresponding tag */
     char ind ;
     int * tp ;
     char * tag ;
{
  while (*tp < TAG_OCCS) {	/* search until found from current position to end */
    if (ind == tag_values[*tp][3]) {
      return(1) ;
    }
    (*tp)++ ;
    if (strncmp(tag, tag_values[*tp], 3) != 0) {
      if (strcmp(fcno, "0002111667") == 0) printf("skipping ... fcno %s\n", fcno) ; 
      return(-1) ;
    }
  }
  return(-1) ;
}


		      
  
/* check_tag
   searches table of ascending tag values for current tag ... if found returns 1
   else returns -1 */

check_tag(tag, tp)		/* searches tag values for matching tag */
     char * tag ;
     int * tp ;
{
  while (strncmp(tag, tag_values[*tp], 3) != 0 && *tp < TAG_OCCS) {
    if (strncmp(tag, tag_values[*tp], 3) < 0) 
      return(-1) ;
    if (strncmp(tag, tag_values[*tp], 3) > 0) {
      (*tp)++ ;
      if (*tp > TAG_OCCS) return(-1) ;
    }
  }
  return(1) ;			/* matching tag found in table of tag values */
}  




/* translate_char()

   translation of characters from 8-bit mark format to ascii ... characters outside of
   the ascii character set are preceeded by '\'.
   different from gills program - read whole record into buffer and translate each
   character in turn, inserting special characters into buffer if necessary, don't
   bother with seperate buffer to hold block accents */


translate_char (mchar, achar) /* convert from 8-bit to ascii */
     unsigned char * mchar ;
     char * achar ;
{
  int x = 0 ;			/* count of extra characters from accents */
  int s = 0 ;			/* count of special characters */
  char * a ;
  a = achar ;

  if (*mchar < 0x5e) {		/* punctuation, numbers and UC alphabetic */
    if (*mchar == 0x1e || *mchar == 0x1f) /* convert to ascii 0x23 and 0x24 */
      *mchar += 5 ;
    *a++ = *mchar ; 
  }
  else if (*mchar < 0x61) {	/* single dagger, eszett and sharp */
    *a++ = '\\' ;
    *a++ = *mchar ;
    x++ ;
  }
  else if (*mchar < 0x7b) {	/* lower case alphabetic */
    *a++ = *mchar ;
  }
  else if (*mchar < 0x80) {
    *a++ = '\\' ;
    *a++ = *mchar ;
    *a   = '\0' ;
    x++ ;
  }
  else if (*mchar == 0xe0) {	/* high tone */
    *a++ = '\\' ; 
    *a++ = 0xbb - 0x80 ;
    *a   = '\0' ;
    x++ ;
  }
  else if (*mchar == 0xfe) {	/* high comma centre */
    *a++ = '\\' ;
    *a++ = 0xbc - 0x80 ;
    *a   = '\0' ;
    x++ ;
  }
  else  {
    *a++ = '\\' ;
    x = subtract_backslash_and_save(mchar) ; 
    if (x) {	/* extra characters from table of special char values */
      *a++ = *mchar ;
      if (mchar[1] != '\0') *a++ = mchar[1] ;
    }
    else {	/* subtract 80 hex from character value so printable */
      *a++ = *mchar - 0x80 ; x++ ;
    }
    *a = '\0' ;
  }

  return(x) ;
}



/* subtract_backslash_and_save */

/* deals with special characters unprintable in 7-bit ascii ... for certain
   special characters (hex values in table special_vals) replaces character(s)
   with \ and character(s) from special_ch_vals */

subtract_backslash_and_save(a) /* deals with special case character conversion */
     char * a ;
{
  int x = 0, s = 0 ;

  while (s < SC_NUM ) {		/* is this character special hex value ??? */
    if (*a == special_vals[s]) break ;
    s++ ;
  }

  a++ ;

  if (s < SC_NUM) {		/* yes ... get alternative charcters from table */
    *a++ = special_ch_vals[s][0] ;
    x++ ;
    if (special_ch_vals[s][1] != ' ')  { *a++ = special_ch_vals[s][1] ; x++ ; }
    *a = '\0' ;
  }
  return(x) ;
}


				

				


/* save_subfield */

save_subfield(tag_buf, punc, sf_buf) /* appends subfield onto buffer containing all
					     subfields for that tag */
     char * tag_buf ;		/* buffer that subfields appended to */
     char * punc ;
     char * sf_buf ;		/* sub field buffer */
{
  char * temp ;
  char * tp ;


  if (*tag_buf == ' ' && punc[1] != '(') strcpy(punc, "#") ;
  if (*tag_buf == ' ') *tag_buf = '\0' ;

  strcat(tag_buf, punc) ;

  tp   = tag_buf ;
  while (*tp != '\0') tp++ ;
  temp = sf_buf ;
  while (*temp != '\0') {
    if (*temp == '#') temp++ ;	/* skip end of field mark within string */
    else  *tp++ = *temp++ ;
  }
  *tp = '\0' ;

  if (punc[1] == '(' ) *tp++ = ')' ; *tp = '\0' ; 
}



/* dont_index_article
   put "don't index" markers around article at start fo 240, 245, 248 (only used
   when no $g field), 440, 490 */

dont_index_article(tag_buf, sub_buf, ind2, tag)	/* insert don't index markers */
     char * tag_buf ;
     char * sub_buf ;
     int ind2 ;
     char * tag ;
{
  int i ;

  if (*tag_buf == ' ') *tag_buf = '\0' ;

  if (ind2 == '1') 
    strncat(strcat(strncat(strcat(tag_buf, "_"), sub_buf, 1), "^"), sub_buf + 1 , 499) ;
  else if (ind2 == '2')
    strncat(strcat(strncat(strcat(tag_buf, "_"), sub_buf, 2), "^"), sub_buf + 2 , 498) ;
  else if (ind2 == '3')
    strncat(strcat(strncat(strcat(tag_buf, "_"), sub_buf, 3), "^"), sub_buf + 3 , 497) ;
  else if (ind2 == '4')
    strncat(strcat(strncat(strcat(tag_buf, "_"), sub_buf, 4), "^"), sub_buf + 4 , 496) ;
  else if (ind2 == '5')
    strncat(strcat(strncat(strcat(tag_buf, "_"), sub_buf, 5), "^"), sub_buf + 5 , 495) ;
  else 
    printf("fcno (%s) tag (%s) ind2 (%c) - ind2 out of range\n", fcno, tag, ind2) ;
}




/* initials 
   take initials from name: 100 (personal name), 600 (personal name subject heading),
   700 (personal name added entry heading)
   keep hyphens (-) drop anything beginning with a lower case letter (de, von etc) */

initials(sub_buf, punc)		/* take initials from 100, 600, 700 */
     char * sub_buf ;		/* current position in subfield buffer */
     char * punc ;
{
  char * sp ;
  char w_delim[3] ;		/* wanted delimiter */
  char init[100], * ip ;
  char initials[100], * ptr ;
  int i ;

  ip  = init ;
  sp  = sub_buf ;
  ptr = initials ;

  *w_delim = ' ' ;
  while ( *w_delim != '#' && strncmp(w_delim, ".#", 2) != 0 && *sp != '\0') {

    ip = init ;
    i  = 0 ;

    /* init next */

    while ( *sp != ' ' && *sp != '.' && *sp != '-' && *sp != '#' && *sp != '\0' &&
	   strncmp(sp, ". ", 2) != 0 && strncmp(sp, ".#", 2) != 0 ) {
	     if (i++ < 5) *ip++ = *sp++ ;
	     else sp++ ;
	   }
    if ( strncmp(sp, ". ", 2) == 0 || strncmp(sp, ".#", 2) == 0 ) {
      strncpy(w_delim, sp, 2) ;
      sp += 2 ;
    }
    else { *w_delim = *sp++ ; w_delim[1] = '\0' ; }
    *ip = '\0' ; 

    /* init check */
    while (*init == '\\') {	/* to get over names beginning with accent */
      if  (init[2] == '\\') *init = init[4] ;
      else                  *init = init[2] ; 
    }
    if (*init > 'Z' || *init == ' ') { /* mainly lower case alphabetic */
      if (*w_delim == '#' || (strncmp(w_delim, ".#", 2) == 0)) break ;
      else continue ;        	       /* top of loop ... next name  */
    }
    *ptr++ = *init ;		/* put initial into subfield buffer */
    if (*w_delim == '-')	/* retain hyphen */
      *ptr++ = '-' ;
    else *ptr++ = ' ' ;
  }

  /* init end */
  ptr-- ; *ptr = '\0' ;
  strcpy(sub_buf, initials) ;	/* return initials to calling function */
  *punc = ' ' ; punc[1] = '\0' ; 


}



/* string_wfld
   Concatenate wanted field at end of ouput string. Doesn't copy end of field
   marker (shouldn't be there !!). Removes extra spaces at end.
   Returns count of number of characters copied. */

string_wfld(wf, op, a)		/* string wanted field onto end of output string */
     WANTED_FIELDS wf ;
     char * op ;
     int a ;
{
  char * p ;
  int count = 0 ;

  p = wf[a].field ;

  while (*p != '\0') {
    if (*p == '#') p++ ;
    else { *op++ = *p++ ; count++ ; }
  }
  wf[a].tag[0] = ' ' ; 
  *op = '\0' ;
  return(count) ;
}





/* string_dot_wfld
   same as string_wfld but adds punctuation at end */

string_dot_wfld(wf, op, a)		/* punctuate end of outputted string */
     WANTED_FIELDS wf ;
     char * op ;
     int a ;
{
  char * p ;
  int count = 0 ;

  p = wf[a].field ;

  while (*p != '\0') {
    if (*p == '#') p++ ;
    else { *op++ = *p++ ; count++ ; }
  }
  wf[a].tag[0] = ' ' ; 

  if (*(op - 1) == ' ')  while (--*op == ' ') count-- ; 
  if (*op == '.' || *op == ',' || *op == '!' || *op == '?' || *op == ']' || *op == ':' ) 
    { *op++ = ' ' ; count++ ; }
  else { *op++ = '.' ; *op++ = ' ' ; count+= 2 ; }
  *op = '\0' ;
  return(count) ;
}



/* finish field */

finish_field(op, obuf)		/* '#' end of field appended to output string */
     char * op ;
     char * obuf ;
{
  int count = 0 ;

  if      (op == obuf)       { *op++     = '#' ; count++ ; }
  else if (*(op - 1) == ' ') *(op - 1) = '#' ;
  else                       { *op++     = '#' ; count++ ; }
  *op = '\0' ;
  return(count) ;
}




/* subj()   28 jun 88 

   direct copy from gill's program "subj section" ... that attempts to cope with
   duplicates ... doesn't work very well.
   process subject headings ... punctuation and marks for individual headings into
   global array of structs type sub_hdr.  */

subj(wf, delim, count, a, mark) /* get subjects ... attempt to skip duplicates */
     WANTED_FIELDS wf ;
     char * delim ;
     int * count ;
     int a ;
     char mark ;
{
  char buf[FIELD_LEN] ;
  char * b ;
  char * temp ;
  int i = 0 ;
  int j = 0 ;

  strcpy(delim, "  ") ;		/* initialise delimiter */
  temp = wf[a].field ;

  while (*temp != '#' && *temp != '\0') temp++ ; temp++ ;  /* skip first '#' */

  while ( *delim != '#' && strcmp(delim, ".#") != 0 && strcmp(delim, ",#") != 0 ) {

    b = buf ;
    while (strncmp(temp, ". ", 2) != 0 && strncmp(temp, ".#", 2)  != 0 &&
	   strncmp(temp, ",#", 2) != 0 && strncmp(temp, "--", 2)  != 0 &&
	   strcmp(temp, ",. ", 3) != 0 && *temp != '#' ) {
	     *b++ = *temp++ ;
	     i++ ;
	   }
    *b = '\0' ;

    if (*temp == '#')                    { *delim = '#' ; temp++ ; }
    else if ( strncmp(temp, ",. ", 3) == 0 ) { strcpy(delim, ",. ") ; temp += 3 ;  }
    else if ( strncmp(temp, "---", 3) == 0 ) { strcmp(delim, "--") ;  temp += 3 ;  }
    else                                 { strncpy(delim, temp, 2) ; temp += 2 ; }
		   
    if (*temp == '#') temp++ ;

    if ( i > S_HDR_LEN - 1 ) {
      printf("fcno (%s) subject heading truncated since exceeds 1024 \n", fcno) ;
      i    = S_HDR_LEN - 1 ;
      b    = buf + i + 1 ;
      *b++ = '#' ;
    }	       

    /* save subject heading  ... only save field 083 if unique */
	   
    if ( strcmp(wf[a].tag, "083") == 0 ) {
      j = 0 ;
      while (j <= *count) {	/* for all the other subject fields */
	if (strcmp(buf, sh[j].buf) == 0 ) break ;
	j++ ;
      }
    }

    /* field is 083 and unique or other than 083 */
    if ( strcmp(buf, sh[j].buf) != 0 ) {
      sh[*count].mark = mark ;
      strcpy(sh[*count].buf, buf) ;
      strcpy(sh[*count].punc, delim) ;
      *count += 1 ;
    }
  }
  wf[a].tag[0] = ' ' ;

} /* subj */

	       

	       

	       

/* diag_tag */

diag_tag(wf, a)
     WANTED_FIELDS wf ;
     int a ;
{
  char * d ;

  strcat(strcpy(DREC, fcno), " ") ;
  strcat(strcat(DREC, wf[a].tag), " ") ;
  
  d = DREC ; while (*d != '\0') d++ ;
  *d++ = wf[a].ind1 ;
  *d++ = ' ' ;
  *d++ = wf[a].ind2 ;
  *d   = '\0' ;

  return(1) ;
}
     


/* diag */


diag(diag_message)
     char * diag_message ;
{

  strcat(strcat(strcpy(DREC, fcno), " "), diag_message) ;

}
