/*****************************************************************************

  iso2709subs.c Jan 92 SW

  Functions for reading and converting ISO2709 databases

  Databases in a MARC-like ISO 2709 format may come on tapes with a
  slightly variable blocklength. On a unix system the variable blocklength
  tapes can be satisfactorily read using cat /dev/rmt* > whatever.

  Each record consists of a 24-character record label, a directory and some
  data fields. The directory and individual data fields, though not the record
  label, are terminated by a field mark (hex 1E or CTRL-^). The whole record
  ends with a record mark (hex 1D or CTRL-]). Each data field starts with a
  digit (often 1) which is a version number, and each subfield starts with
  a subfield mark (hex 1F or CTRL-_).

  Bytes 0-4 of the record label are the length of the record (inclusive of
  label and all field, subfield and record marks). Byte 5 contains N for new
  (it could maybe contain C for changed). Bytes 12-16 contain the offset of
  the start of the data fields from the beginning of the record label. The
  rest of the record label is unimportant.

  Byte 24 up to a field mark consists of a sequence of directory entries,
  one for each data field, 12 bytes long. Each directory entry conatins three
  subfields. There is one directory entry for each field in the record, and
  entries are in the same order as the fields to which they refer (although
  this is not important). The directory subfields are (bytes 0-2) a three-
  digit tag which identifies the field, (bytes 3-6) the field's length, (bytes
  7-11) the offset relative to the start of the data fields. Thus a suitable
  procedure for locating the data for a specified tag is: search directory
  for the tag, read the offset subfield and add it to the offset of the start
  of the data fields (bytes 12-16 of the record label). This gives the offset
  of the required data from the start of the whole record. If the length is
  required this is given in bytes 3-6 of the directory entry.

******************************************************************************/

#include <stdio.h>
#include "iso2709.h"

/*****************************************************************************

  get_next_record

  Reads next record from stdin into the given buffer, returning its length,
  or a negative result on error. It returns EOF if there is an incomplete
  or empty record at the end of the input. Calling program must look at
  the value of res (see below).

  On entry, the next character read (from stdin) is assumed to be the start
  of a record. 

  As soon as 5 characters have been read this is taken as the record length
  (if they have a plausible value - see below).

  If the first 6 chars don't look plausible input is read to record mark,
  and an error code and message are returned. (Must be 5 digits and letter,
  likely '0'dddd'N'). In this case the record is unusable but the next one
  should be OK.

  Input is read until the indicated no of characters or record mark is read,
  whichever happens first.

  If the last character read was not a record mark, if it was a digit
  it is pushed back on the input stream (assumed count right but
  record mark lost) and the returned length adjusted. Otherwise the
  character is retained as part of the record, but a warning is issued
  if it is not a field mark.

  Result codes (bits):

     0: OK - record mark and length coincide
     1: EOF before end of rec (rec no good);
     2: record seems to start badly; (rec no good)
     4: rec ended because of record mark before indicated length; (rec should
        be OK)
     8: read length but no record mark, last char pushed back (rec OK?)
    16: as 8 but last char retained (rec dubious unless field mark)
    32: null char(s) elided (not counted in length)
    64: newline, ff, cr chars elided and length reduced
   128: control chars (0x1 - 0x1a except 0x9 0x0a 0x0d 0x0c)
   256: non-ascii character present
   512: record too long (ignore buffer contents)

  In cases 8 and 16 the next record is likely to start badly.

  There is no check for overflow of the buffer if it is less than
  REC_MAX_LEN (16K).

*****************************************************************************/

get_next_record(buf, res, silent)
     char *buf ;
     int *res ;			/* Error codes */
     int silent ;		/* If not silent reports to stderr */
{
  register char *op = buf ;
  register int ch ;
  register int incount = 0, outcount = 0 ;
  int indlen ;			/* Indicated record length */
  int fin = 0 ;
  int i ;
  int result = 0 ;		/* Bits used */
  char *label = buf ;
  char status ;

  *res = 0 ;

  /* First try to read 1st 6 chars */
  for ( i = 0 ; i < 6 && (ch = getchar()) != EOF ; i++ ) {
    *op++ = ch ;
    incount++ ;
    outcount++ ;
  }
  if ( ch == EOF ) {
    *res |= 1 ;
    return EOF ;
  }

  sscanf(label, "%5d", &indlen) ;
  status = buf[5] ;
  if ( indlen < 40 || ! isalpha(status) ) {
    *res |= 2 ;
    if ( ! silent )
      fprintf(stderr,
	      "Implausible rec label %.6s, skipping to eorec\n",
	      label) ;
    return skip_thro_recmark() ;
  }
  
  while ( (ch = getchar()) != R_MARK && ch != EOF && incount < indlen) {
    if ( ch == 0 ) {
      *res |= 32 ;
      /* Don't count as incount */
      if( ! silent ) fprintf(stderr, "Null char elided\n") ;
      continue ;
    }
    incount++ ;
    if ( ch == '\n' || ch == '\r' || ch == '\f' || ch == '\t' ) {
      *res |= 64 ;
      if ( ! silent )
	fprintf(stderr, "White space elided\n") ;
      if ( ch == '\t' ) {
	ch = ' ' ;
	*op++ = ch ;
	outcount++ ;
      }
      else continue ;
    }
    else if ( ch <= 0x1a ) {
      *res |= 128 ;
      if ( ! silent )
	fprintf(stderr, "Non-ascii character\n") ;
    }
    else if ( ch >= 0x7f ) {
      *res |= 256 ;
      if ( ! silent )
	fprintf(stderr, "Non-ascii character\n") ;
    }
    *op++ = ch ;
    outcount++ ;
  }
  /* Now ch is R_MARK or EOF */
  if ( ch == EOF ) {		/* EOF before R_MARK, rec no good */
    *res |= 1 ;
    return EOF ;
  }
  else if ( ch == R_MARK ) {
    if ( incount == indlen - 1 ) {	/* Normal case */
      *op++ = ch ;
      outcount++ ;
      incount++ ;
      return outcount ;
    }
    else {
      *res |= 4 ;
      if ( ! silent ) {
	fprintf(stderr,
		"Rec ended with record mark before indicated length:\n") ;
	fprintf(stderr, "length %d but ind length %d\n",
		incount + 1, indlen) ;
      }
      *op++ = ch ;
      incount++ ;
      outcount++ ;
    }
  }
  else {			/* Read correct len, but no R_MARK */
    if ( ! silent )
      fprintf(stderr, "Rec ended with (char) %x and no record mark\n",
	      ch) ;
    if ( isdigit(ch) ) {	/* Assume start of next rec */
      if ( ! silent )
	fprintf(stderr, "--assume start of next rec, pushed back\n") ;
      *res |= 8 ;
      if ( ungetc(ch, stdin) == EOF )
	fprintf(stderr, "Can't unget\n") ;
    }
    else if ( ch == F_MARK ) {	/* Rec prb OK */
      if ( ! silent )
	fprintf(stderr, "--field mark, assume rec OK\n") ;
      *res |= 16 ;
      *op++ = ch ;
      outcount++ ;
    }
    else {			/* Likely incomplete record */
      if ( ! silent ) fprintf(stderr, "--record dubious\n") ;
      *res |= 16 ;
      *op++ = ch ;
      outcount++ ;
    }
  }
  return outcount ;
}

skip_thro_recmark()		/* Throw away input up to record mark */
{
  register int ch, numchars = 0 ;

  fprintf(stderr, "Doing skip_thro_recmark()\n") ;
  while ( (ch = getchar()) != EOF && ch != R_MARK )
    numchars++ ;
  return numchars ;
}


/*****************************************************************************

  check_rec  SW Jan 92

  Unfinished

  Reads ISO2709 record (without segment control words) into a buffer and
  performs a number of checks. It is assumed the first 24 bytes (record label)
  are already at the start of the read buffer

  Record is read up to length as given in record label, or until record
  mark has been read.

  If record mark doesn't agree with length this is reported, next 5 bytes
  are read and checked as digits. If they are digits and the length they
  indicate is > 24
*****************************************************************************/

char buf[REC_MAX_LENGTH] ;

check_rec(buf, result)
     char *buf ;
     int *result ;
{
  char *label_p = buf ;
  char *length_p = buf ;
  char *data_offset_p = buf + 12 ;
  char *dir_p = buf + 24 ;
}  


/*****************************************************************************

  get_data_start()  SW Mar 92

  Gets the address of data start in a 2709 record

*****************************************************************************/

char *
get_data_start(rec)
     char *rec ;
{
  struct leader *leader = (struct leader *) rec ;
  int base_offset ;
  
  sscanf(leader->base, "%5d", &base_offset) ;
  return rec + base_offset ;
}

/***************************************************************************

   find_tag()

   Must be called with buf pointing to a directory record.
   Looks for next directory entry for "tag".
   Returns NULL or the address of the first directory entry for the sought tag.

****************************************************************************/

char *
find_tag(tag, buf, data_start)
     register char *tag, *buf, *data_start ;
{
  int found = 0 ;
  char *dirp = buf ;

  while (! found && dirp < data_start) {
    if ( strncmp(dirp, tag, TAG_LEN) == 0 ) found = 1 ;
    else dirp += DIR_LEN ;
  }
  if ( found ) return(dirp) ;
  else return(NULL) ;
}

/*****************************************************************************

  tag_data_address(tag, dirp, data_start, &length)

  Given dirp pointing to a directory entry returns the start and length of
  the data for the tag whose directory entry this is, or NULL if there is
  no further directory entry for the tag. Also returns the position dirp
  of the tag's directory entry.

*****************************************************************************/


char *
tag_data_address(tag, dirp, data_start, length)
     char *tag, **dirp, *data_start ;
     int *length ;
{
  int offset ;			/* Of data from data_start */

  if ( ( *dirp = find_tag(tag, *dirp, data_start)) == NULL )
    return(NULL) ;
  sscanf(*dirp + 3, "%4d", length) ;
  sscanf(*dirp + 7, "%5d", &offset) ;
  return(data_start + offset) ;
}
		 

/* reperror()  Report tag not found */

reperror(tag, recnum)
     char *tag ;
     int recnum ;
{
  fprintf(stderr, "No %s, rec %d\n") ;
  return -1 ;
}

#if DONE

/***************************************************************************

  writefield_inspec(ip, l, op) performs a number of standard substitutions
  on the contents of a field (of length l), returning the new length of the
  field. It:

    elides ~ ^ @ _
    replaces $ by (dollars) followed by a blank if necessary
    replaces # or \234 by (pounds) followed by a blank if necessary
    replaces ` by ' and \ by / 
    elides and reports non-ascii and control characters other than subfield
      marks (except field, subfield and record marks are left alone).

  writefield() is called before the individual processing for specific fields.

***************************************************************************/

writefield(ip, l, op, silent)
     register char *ip, *op ;
     register int l ;
     int silent ;
{
  register int outl = 0 ;
  register int i ;
  register char ch ;
  char *start = ip ;

  for ( i = 0 ; i < l ; i++ ) {
    ch = *ip++ ;
    if ( ch == '~' || ch == '^' || ch == '@' || ch == '_' ) {
      if ( ! silent ) fprintf(stderr, "%c elided\n", ch) ;
      continue ;
    }
    else if ( ch == '`' ) ch = '\'' ;
    else if ( ch == '\\' ) ch = '/' ;
    else if ( ch == '$' ) {
      strcpy(op, "(dollars)") ;
      op += 9 ; outl += 9 ;
      if ( *ip != ' ' ) {
	*op++ = ' ' ; outl++ ;
      }
    }
    else if ( ch == '#' || ch == '\234' ) {
      strcpy(op, "(pounds)") ;
      op += 8 ; outl += 8 ;
      if ( *ip != ' ' ) {
	*op++ = ' ' ; outl++ ;
      }
    }
    else if ( ! isascii(ch) ) {
      if ( ! silent ) fprintf(stderr,"Non-ascii character elided. Context is\n%.*s\n",
	      l, start) ;
      continue ;
    }
    else if ( iscntrl(ch) && ch != F_MARK && ch != R_MARK
	     && ch != SF_MARK) {
      if ( ! silent ) fprintf(stderr,"Control character elided. Context is\n%.*s\n",
	      l, start);
      continue ;
    }
    else {
      *op++ = ch ;
      outl++ ;
    }
  }

  return(outl) ;
}



#endif DONE
