/******************************************************************************

  simple_extract    SW Nov 02

  Simple version of "words3" text parser. Allows choice of stemfunctions
  but splitting procedure is hard coded and there are no facilites for
  outputting positional information.


******************************************************************************/

#include "simple_extract.h"

/******************************************************************************

  BSS character classification. See charclass.h This array is generated by
  setup_toktab.c

******************************************************************************/

int Toktab[256] = 
{0,0,0,0,0,0,0,0,0,131076,131076,0,131076,131076,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131092,116,294916,0,8,0,1048578,294916,65540,327684,0,0,20,4,84,4,16386,16386,16386,16386,16386,16386,16386,16386,16386,16386,116,116,32772,4,32772,116,8,642,770,770,770,4738,770,770,770,4738,770,770,6914,770,770,642,770,770,770,6914,770,642,770,8962,8962,8706,2818,4,0,262148,8,524296,32772,1154,1282,1282,1282,5250,1282,1282,1282,5250,1282,1282,7426,1282,1282,1154,1282,1282,1282,7426,1282,1154,1282,9474,9474,9218,3330,4,8,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0} ;


/******************************************************************************

  para.c     SW March 94

  Routines for paragraph identification

  find_para(s, l, rule) called at start of line in s will scan up to l chars
  looking for the first line which starts a paragraph. It returns the length
  scanned up to the start of the paragraph (or l if it hasn't found one).
  It will not return a paragraph of length 0 unless it is called with
  l = 0 ( i.e. it is assumed that the start is known to be the beginning of a
  para). Typically, it would be called repeatedly, advancing each time by
  the previous call's return value.

  Several rules are built in. If rule & INDENT a paragraph starts at
  the beginning of a line which has some alphanumeric content and is
  indented by more than the previous content-bearing line. If rule &
  GAP a paragraph starts at a content bearing  line preceded by one or
  more "empty" lines. Other rules may be coded in as required. They should
  be represented by single bits. Some rules may need too much context
  to be implemented in this function.

  Different rules are needed for some texts. For example, for plays the
  indent criterion is wrong and should be replaced by first word upper case.

  Notes.

  The INDENT rule is not clever. If fails to detect one-line paras. A concept
  of "prevailing indentation" may be useful, but beware.

  Nov 02: modifying for "simple"; removed UC rule and diags

******************************************************************************/

int
simple_find_para(char *s, int l, int rule)
{
  int retval = 0 ;
  int ll = 0 ;			/* Linelength */
  int indent = 0 ;
  int newindent ;
  int newlines = 0 ;
  int interline = 1 ;		/* Conceivably this could be passed or
				   determined within this function */
  int firstline = TRUE ;	/* Never want para before 1st line */
  int content ;
  char ch ;
  char *p = s ;
  char *linestart, *content_start, *pp ;
  int i ;

  while ( l > 0 ) {
    /* Find next line */
    ll = ch = 0 ;
    linestart = p ;
    while ( l-- > 0 && (ch = *p++) != '\n' ) ll++ ;
    if ( ch == '\n' ) ll++ ;
    content = FALSE ;
    newindent = 0 ;
    /* Examine line */
    for ( i = 0, pp = linestart ; i < ll ; i++, pp++ ) {
      ch = *pp ;
      if ( content == FALSE ) {
	if ( Toktab[(int)ch] & TOK_START ) { /* ? */
	  content = TRUE ;
	  content_start = pp ;
	}
	else if ( ch == ' ' ) newindent++ ;
	else if ( ch == TABCHAR ) newindent += TABSIZE ;
      }
    }
    if ( content == FALSE ) newlines++ ;
    /* Check rules */
    else {
      if ( ! firstline ) {
	if (((rule & GAP) && newlines >= interline) ||
	    ((rule & INDENT) && newindent > indent) )
	  break ;
      }
      newlines = 0 ;
      indent = newindent ;
      firstline = FALSE ;
    }
    retval += ll ;
  }
  return retval ;
}

/******************************************************************************

  sentence.c     SW April 94

  Routines for sentence identification

  find_sentence(s, l, rule) will scan up to l chars looking for the
  start of the next sentence (or end of input). It returns the length
  scanned. Typically, it would be called repeatedly, advancing each time by
  the previous call's return value.

  Several rules are supposed to be built in. The rule which is used should
  depend on the nature of the input.

  The most general rule is for data of field_type TEXT. 

  The rule is approximately as follows.

  <PUNCT> = /[.;:?!]/, <CLOSING> = /[\])\'\"], <SPACE> = /[ \n\t]/

  Consider for sentence end (success) at the end of
  <PUNCT><CLOSING>*<SPACE><SPACE>* and at end of input.

  If end of input, success.
  If there has been no alphanumeric content so far, failure.
  If there is only a little left, don't bother (success at end of text).
  If <PUNCT> is other than full stop, success.
  Go forward and look at next text token.
    (a text token is, roughly, any alphanumeric sequence)
  If it doesn't start with capital letter, failure.
  Backtrack and look at text token preceding full stop.
  If length <= 5 and initial capital (or all caps), failure.
    (prevents success after "Ms. ", "James A.P. ")
  If length = 1, failure.
    (prevents success after "b.b.c. ")
  Otherwise success.

  Notes

  A paragraph end (in real life) always ends a sentence. This routine doesn't
  look for paragraph ends. It should not be called across paragraph
  boundaries (see para.c).

  The length returned contains (of course) any white space etc in the input
  before the first text token. It also contains any trailing space after
  the terminating punctuation.

  Obviously it cannot distinguish between ".. U.S. Navy .." and
  ".. in the U.S. The Navy .. " (won't break either). Nor will it start
  a sentence after a full stop when the next token starts with a digit
  (or lower case letter). It will not work well with text which is all caps.

  The boundary conditions haven't been thoroughly tested.

  Obviously, somewhat cleverer rules can be devised, but it's probably not
  worth it. Speed is quite important.

  NB Only the "TEXT" rule has been written (the value of "rule" is
  ignored). Other data-types include TITLE, SUBJECT_HEADINGS, NAMES,
  where special delimiters are used. These cases are much simpler.

  Apr 02: maxlength seems to be unused at present

  Nov 02: modified a bit from BSS find_sentence
  Note that sentence is supposed to be truncated if maxtoks have been output
  (this was added some time ago -- not sure if it works)

******************************************************************************/

int
simple_find_sentence(char *s, int l, int rule, int maxtoks, int maxlength)
{
  int retval = 0 ;
  int rretval ;
  int c ;
  char *p = s ;
  char *pp, *tok_start ; 
  int ll ;
  int ini_cap ;
  int content = 0 ;		/* Avoid trivial sentences */
  int toklen ;
  int numtoks = 0 ;
  int intoken = FALSE ;

  /* TEXT rule only written */

  switch (rule) {

  default:
    while ( l-- > 0 ) {
      retval++ ;
      if ( Toktab[(c = *p++)] & TOK_START ) {
	content++ ;
	if ( ! intoken ) {
	  numtoks++ ;
	  intoken = TRUE ;
	}
      }
      else if ( intoken && (Toktab[c] & TOK_FIN) ) {
	intoken = FALSE ;
	if ( numtoks == maxtoks ) {
	  /* Should look back for last linefeed or something */
	  ll = l ; pp = p ; rretval = retval ;
	  /* If wanting to return numtoks have to decrement here */
	  while ( --retval > 0 ) if ( *(--pp) == '\n' ) break ;
	  if ( retval == 0 ) retval = rretval ;
	  else break ;
	}
      }
      if ( l <= LITTLE_BIT_LEFT ) { /* Don't bother with the rest */
	retval += l ; l = 0 ;
	continue ;
      }
      /* Scan for punct */
      if ( ! content || ! (Toktab[c] & TOK_STOP) ) continue ;
      /* First look at continuation */
      /* First requirement is that the next bit should be either space or
	 some sort of closing quote or bracket then space. */
      ll = l ; pp = p ;		/* Save l & p in case have to backtrack */
      while ( ll > 0 && (Toktab[(int)(*pp)] & TOK_CLOSING) ) {
	ll-- ; pp++ ; 
      }
      if ( ll == 0 || ! tok_space(*pp) ) {
	retval += l - ll ; l = ll ; p = pp ; 
	continue ;
      }
      if ( c == '.' ) {
	/* This is the most likely case, and the most messy */
	/* Then find next TOK_START (now pointing to SPACE after STOP) */
	pp++ ; ll-- ;
	while ( ll > 0 && ! (Toktab[(int)(*pp)] & TOK_START) ) {
	  pp++ ; ll-- ;
	}
	if ( ll == 0 || ! tok_upper(*pp) ) continue ;
	/* If here, have got DOT [CLOSING] SPACE [NON-A/N] UC */
	/* Then have to backtrack & look at preceding token */
	pp = p - 2 ; 
	/* Starting with the character before the '.', locate end, and
	   then beginning of previous token */
	while ( pp >= s && ! (Toktab[(int)(*pp)] & TOK_START) )
	  pp-- ;
	/* Note the above depends on tokens being identifiable backwards
	   or forwards, indifferently */
	ini_cap = TRUE ;
	toklen = 1 ;
	while ( pp >= s && ! (Toktab[(int)(*pp)] & TOK_FIN) ) {
	  toklen++ ;
	  pp-- ;
	}
	/* pp now pointing to char before start of token */
	tok_start = pp + 1 ;	/* Only for debugging I think */
	if ( tok_upper(*tok_start) ) ini_cap = TRUE ;
	else ini_cap = FALSE ;
	if ( (toklen <= 5 && ini_cap) || toklen == 1 ) continue ;
      }				/* Case '.' */
      /* Now if we've got here there is a sentence */
      while ( l > 0 &&
      (Toktab[(int)(*p)] & (TOK_SPACE|TOK_STOP|TOK_CLOSING))) {
	retval++ ;		/* THINK this is right */
	l-- ; p++ ;
      }
      break ;
    }
    return retval ;
  }
}

/******************************************************************************

  simple_w3_rewrite

  Rewrites input converting to LC as sequence of "words" each terminated by
  a single '\0'

  Returns number of output strings

  Modified from BSS phase2_token()

  Handles supposed "initialisms", retains '.' in numbers and makes a crude
  attempt at handling hyphens. Hyphens are retained between digits, squashed
  in positions 2-4 in a word, treated as blank in other positions.

  Apostrophes are retained if in position 2 in a word, otherwise squashed.

  Ampersand if surrounded by blanks is replaced by "and"

  The output should contain only alphanumerics, the occasional apostrophe,
  and '.' and '-' within numerics.

  Note outputblank and extrablank are identical in this version and
  output a null char instead of a blank.

******************************************************************************/


#define SPACE 1
#define INIT 2
#define ALLDIGIT 4

int
simple_w3_rewrite(char *inptr, int inlength, char *t, int *skipping)
{
  int ch, nch ;			/* Current and next character */
  int flags = SPACE ;
  int l = 0 ;			/* Output length */
  int tt ;
  int wordlen = 0 ;
  int wordcount = 0 ;

#define outputchar(x) *t++ = x;l++;if (wordlen++==0) wordcount++; 

#define outputblank *t++='\0';l++;wordlen=0;flags=SPACE;

#define extrablank outputblank

  if ( inlength > 0 ) nch = *inptr++ ;
  while ( inlength-- > 0 ) {
    ch = nch ; nch = ( inlength > 0 ) ? *inptr++ : '\0' ; 

    if ( *skipping ) {
      if ( ch == '>' ) *skipping = 0 ;
      continue ;
    }

    tt = Toktab[ch] ;
    if ( tt & TOK_ALPH ) {
      if (flags & ALLDIGIT) {
	if ( *(t - 1) == '-' ) *(t - 1) = ' ' ;
	flags &= ~ALLDIGIT ;
      }
      ch = tolower(ch) ;
      if (flags & INIT) {
	if ( tok_alnum(nch) || nch == '\'' ) { extrablank ; }
      }
      else if ( ((flags & SPACE) || *(t-1) == '-') &&
	       (nch == ' ' || nch == '.')) {
	flags |= INIT ;
      }
      flags &= ~SPACE ; outputchar(ch) ;
    } /* (alph) */
    else if ( tt & TOK_DIG ) {
      if ( flags & INIT ) { extrablank ; }
      if ( flags & SPACE ) flags |= ALLDIGIT ;
      flags &= ~SPACE ; outputchar(ch) ;
    } /* (dig) */
    else if ( isspace(ch) ) {
      if (!(flags & (INIT|SPACE)) &&  !(*(t-1) == '-' && (flags & ALLDIGIT))) {
	outputblank ;
      } 
    } /* otherwise blank is skipped */ 
    else if (ch == '.') {
      /* If ALLDIGIT and next char is digit output unchanged else if next is
	 apunct or initflag or spaceflag ignore, otherwise output space */
      if ( (flags & ALLDIGIT) && tok_digit(nch) ) {
	outputchar(ch) ; flags &= ~SPACE ; 
      }
      else if ( !( (Toktab[nch] & TOK_APUNCT) || ( flags & (INIT|SPACE)))){
	extrablank ;
      }
      /* else ignore */
    } /* '.' */
    else if ( ch == ',' ) {
      flags &= ~INIT ;
      if ( ! ( (Toktab[nch] & TOK_APUNCT) || ( flags & SPACE ) ||
	      ( tok_digit(nch) && tok_digit((*(t-1))) ) ) ) {
	extrablank ;
      }
    }
    else if ( tt & TOK_MPUNCT ) {
      flags &= ~INIT ;
      if ( ! ( (flags & SPACE) || (Toktab[nch] & TOK_MPUNCT) ) ) {
	extrablank ;
      }
    }
    else if ( ch == '-' ) {
      /* Dec 95: bug in incomplete date ranges e.g. 1914-
	 Added the Toktab[nch) & TOK_DIG
	 but really want lookahead to next A/N.
	 Not sure if the *(t-1)=ch is sound either. */
      if ( (flags & ALLDIGIT) && tok_digit(nch) ) {
	if ( flags & SPACE ) { flags &= ~SPACE ; *(t-1) = ch ; }
	else if ( *(t - 1) != '-' ) { outputchar(ch) ; }
      }
      else if ( (flags & SPACE) || (Toktab[nch] & TOK_APUNCT) ) ;
      else if ( nch == '-' ) { extrablank ; }
      else { 
	/* Inserting here Apr 93 a crude version of the old do_hyphs() */
	if ( wordlen >= 4 ) { extrablank ; }
	/* else squash it */
      }
      flags &= ~INIT ;
    }

    else if ( ch == '/' ) {	/* Cf. hyphen */
      if ( ! (flags & SPACE) &&  tok_alnum(nch) &&
	  tok_alnum(*(t-1)) ) {	extrablank ; }
      /* else squash */
    }
    else if ( ch == '\'' && ( l == 1 || *(t - 2) == ' ' ) &&
	     ! ( flags & INIT ) && tok_alnum(nch) &&
	     tok_alpha(*(t - 1)) ) {
      outputchar(ch) ;
    }
    else if ( ch == '&' && ((flags & SPACE) || (flags & INIT))
	     && nch == ' ' ) {
      if ( ( flags & INIT ) ) {	extrablank ; }
      outputchar('a') ; outputchar('n') ; outputchar('d') ;
      flags &= ~(SPACE|INIT|ALLDIGIT) ;
    }
    else if ( ch == '<' ) { *skipping = TRUE ; continue ; }

    /* else ignore */
  }
  if ( *(t - 1) != '\0' ) { outputblank ; }
  /* fprintf(stderr, "length=%d\n", l) ;*/
  return wordcount ;
}

/******************************************************************************

  simple_extract_words() 

  from BSS extract()  originally SW June 92

  The function is for extracting from one "paragraph" of text, not more than
  one field of a record & not more than MAX_TEXT in length, preferably a single
  "sentence".

  Return value: number of items output.

  Calls simple_w3_rewrite followed by stopping, then stemming.
  Output is placed in the supplied buffer as a sequence of terminated
  strings.

  Jan 99: added int * arg "skipping" which is passed to and set/reset
  by simple_w3_rewrite. Bodge to get round only first "sentence" of <....>
  was stopped.

******************************************************************************/

int
simple_extract_words(char *inbuf, int inlen,
		     int (*stemfn)(char *, int, char *), int (*stopfn)(char *),
		     char *outbuf, int *outlen,
		     int *skipping )
{
  int nterms, outterms ;
  int outtermlen, intermlen ;
  char *ip, *op ;
  int i ;

  nterms = simple_w3_rewrite(inbuf, inlen, outbuf, skipping) ;
  /* Terms are now a sequence of nterms strings in outbuf */
  for ( i = 0, ip = op = outbuf, outterms = *outlen = 0 ;
	i < nterms ;
	i++, ip += intermlen + 1 ) {
    intermlen = strlen(ip) ;
    if ( stopfn(ip) ) continue ;
    outtermlen = stemfn(ip, strlen(ip), op) ;
    op += outtermlen ;
    *op++ = '\0' ;
    *outlen += outtermlen + 1 ;
    outterms++ ;
  }
  return outterms ;
}

/******************************************************************************

  simple_split_text

  Slightly modified from BSS extract_text_nosource()
  Returns number of terms placed in output buf

  Input is a text of given length, for example the indexable portion
  of a single document.

  For each paragraph, for each sentence calls simple_extract_words()
  and accumulates the output as a sequence of strings in the supplied
  buffer.

******************************************************************************/

int
simple_split_text(char *inptr, int inlen, int (*stemfn)(char *, int, char *),
		  int (*stopfn)(char *), char *outptr)
{
  char *sp ;
  int pl, sl ;
  int numterms = 0 ;

  int skipping ;
  int outlen ;
  int pn = 0, sn = 0 ;		/* debug only */
  while ( inlen > 0 ) {
    sn=0 ;
    pl = simple_find_para(inptr, inlen, INDENT|GAP) ; 
    pn++ ;
    skipping = 0 ;
    inlen -= pl ; sp = inptr ; inptr += pl ;
    while ( pl > 0 ) {
      sl = simple_find_sentence(sp, pl, 0, TEXTPOS_MAX_T, 0) ;
      sn++ ;
      numterms += simple_extract_words(sp, sl, stemfn, stopfn, outptr,
				       &outlen, &skipping) ;
      pl -= sl ;
      sp += sl ;
      outptr += outlen ;
    }
  }
  return numterms ;
}

/******************************************************************************

  This main() is given as an example and for testing. The document to be
  processed is on stdin and output is to stdout. Note that it works on
  an indefinitely large amount of input but the last term in each
  buffer except the last and the first term in each buffer except the
  first are likely to be truncated or curtailed.

******************************************************************************/
char *Usage = "Usage: %s -help | -stopfile <name> -stemfunction <name>\n\
Parses text on stdin into one term per line. The stoplist is applied\n\
before stemming. Stemfunctions are psstem, sstem, wstem or nostem.\n\
Defaults are stopfile /dev/null and stemfunction nostem.\n" ;


int psstem(char *, int, char *) ;
int sstem(char *, int, char *) ;
int wstem(char *, int, char *) ;
int nostem(char *, int, char *) ;

int
main(int argc, char **argv)
{
  int i ;
  char *stopfilename = "/dev/null", *stemfuncname = "nostem" ;
  int (*stemfn)(char *, int, char *) ;
  int nchars ;
  char *inbuf, *outbuf ;
  int nstops ;
  int nterms ;
  char *op ;

  for ( i = 1 ; i < argc ; i++ ) {
    if ( strcmp(argv[i], "-stopfile") == 0 && i < argc - 1 )
      stopfilename = argv[++i] ;
    /* else if ( strcmp(argv[i], "-stopfunction") == 0 && i < argc - 1 )
       stopfuncname = argv[++i] ; */
    else if ( strcmp(argv[i], "-stemfunction") == 0 && i < argc - 1 )
      stemfuncname = argv[++i] ;
    else if ( strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "-help") == 0 ) {
      fprintf(stderr, Usage, argv[0]) ;
      exit(1) ;
    }
    else {
      fprintf(stderr, Usage, argv[0]) ;
      exit(1) ;
    }
  }
    
  if ( strcmp(stemfuncname, "nostem") == 0 ) stemfn = nostem ;
  else if ( strcmp(stemfuncname, "wstem") == 0 ) stemfn = wstem ;
  else if ( strcmp(stemfuncname, "sstem") == 0 ) stemfn = sstem ;
  else if ( strcmp(stemfuncname, "psstem") == 0 ) stemfn = psstem ;
  else {
    fprintf(stderr, "%s: can't find stem function %s\n",
	    argv[0], stemfuncname) ;
    fprintf(stderr, Usage, argv[0]) ;
    exit(1) ;
  }
  inbuf = malloc(MAXDOCSIZE) ;
  outbuf = malloc(MAXDOCSIZE) ;
  nstops = setup_stoplist(stopfilename) ; /* Defaults to /dev/null */
  while ( !feof(stdin) ) {
    nchars = fread(inbuf, 1, MAXDOCSIZE - 1, stdin) ;
    inbuf[nchars] = '\0' ;
    nterms = simple_split_text(inbuf, nchars, stemfn, simple_stop, outbuf) ;
    /* fprintf(stderr, "nterms=%d\n", nterms) ;*/
    for ( i = 0, op = outbuf ; i < nterms ; i++ ) {
      printf("%s\n", op) ;
      op += strlen(op) + 1 ;
    }
    /* fprintf(stderr, "Done %d terms\n", nterms) ; */
  }
  exit(0) ;
}
