/***************************************************************************/
/*                                                                         */
/* This program compares the rankings from 1 system with another.     	   */
/*									   */
/* USAGE compare <qrels_file> <sys1_rankings> <sys2_rankings> [-q|-s|-c|-C]*/
/*									   */
/* Assumes that the qrels file is in query order and that sys1 and sys2    */
/* are also in query order.						   */
/*									   */
/* Blaim : Martin Smith							   */
/*								  	   */
/***************************************************************************/



#include <stdio.h>
#include <stdlib.h>

#define no_summary_lines 500
#define summary_line_size 140


/* Use a linked list to store all the document ids for reldocs from qrels,    */
/* docs from ranking for sys1 and docs from rankings for sys2 for every query */
   
#define idlen 25   
typedef struct dnode *recptr;
typedef struct dnode {
			char id[idlen];
			recptr dnext;
		     } rec;




/* A structure is used to group linked variables together */
/* This one holds recall precision info                   */

#define no_precision_levels 11  /* Must be > 3 to be effective */
#define no_doc_levels 9		/* Links into the doc_levels global variable - see below */

typedef struct rp *rpinfo;
typedef struct rp {
			float	ip [no_precision_levels];	/* Interpolated precision  recall values */
			float	dp [no_doc_levels];		/* Document precision levels */
			float	avg_precision;
			float	r_precision;	
			long	retrieved;
			long	retrieved_relevant;
		  } rpinf;



/* Global Variable ooops... Needed for Document levels - never changes */

int	doc_levels [no_doc_levels] = {5, 10, 15, 20, 30, 100, 200, 500, 1000};


/* Structure used to hold totals for averages for summary of run */

typedef struct avg *avginfo;
typedef struct avg {
			float	tot_no_relevant;
			float	tot_precision_diff;
			float	tot_precision_at_100_diff;
			float	tot_common100;
			float	tot_common1000;
			float	tot_unique1100;
			float	tot_unique11000;
			float	tot_unique2100;
			float	tot_unique21000;
			int	better_sys1;
			int	better_sys2;
			float	tot_better_diff_sys1;
			float	tot_better_diff_sys2;
			float	tot_better_rel_sys1;
			float	tot_better_rel_sys2;
			float	tot_better_unique100_sys1;
			float	tot_better_unique100_sys2;
			float	tot_better_unique1000_sys1;
			float	tot_better_unique1000_sys2;
			float	tot_comp100;
			float 	tot_comp1000;
			float	tot_diff100;
			float	tot_diff1000;
		    } avginf;


/* Procedure Prototypes */

void	mainloop(FILE *, FILE *, FILE *, int);
recptr	getrels(FILE *, recptr, int*, int*, char *, int*);
recptr	getdocs(FILE *, recptr, int*, int, int*, char *);
void	compare(int, recptr, recptr, recptr, int, char *, rpinfo, rpinfo, avginfo);
void 	recall_precision(recptr, recptr, int, rpinfo);
int  	common(recptr, recptr , recptr , int);
int 	unique(recptr, recptr, recptr, int);
void 	print_tables(int, long, rpinfo, rpinfo);
void	deletedocs(recptr);
int	member(char *, recptr);
int	length(recptr);
void 	init_rpinfo(rpinfo);



/* main - open up the files if they exist, set the system going and then */
/* close it all down again when things are done                          */

int main (int argc, char **argv)
{
	int 	exit_flag;
	FILE 	*qrel;
	FILE	*sys1;
	FILE	*sys2;
	int	detail;	/*set to 1 for -s flag or 2 for -q flag or 3 for -c flag*/
	
	exit_flag = 0;
	
	detail = 0;
	
	/* Check for 4 parameters - program_name, rels_file, sys1_ranking, sys2_ranking */
	if ((argc < 4) || (argc > 5))
	{
		fprintf(stderr,"Too many or too few parameters - usage is compare <qrels> <sys1> <sys2> [-s|-q|-c|-C]\n");
		fprintf(stderr,"Switch must be -c for just recall & precision summary, -C for recall & precision for each query,\n");
		fprintf(stderr,"               -s for summary or -q for detail for each query.\n");
		exit_flag = 1;
	}
	else
	/* Check for the existance of qrels file and open */
	if ((qrel = fopen(argv[1],"r")) == NULL)
	{	fprintf(stderr,"Can't open the qrels file %s.\n",argv[1]);
		exit_flag = 1;
	}
	else
	/* Check for the existance of sys1 file and open */
	if ((sys1 = fopen(argv[2],"r")) == NULL)
	{	fprintf(stderr,"Can't open the sys1 file %s.\n",argv[2]);
		exit_flag = 1;
	}
	else
	/* Check for the existance of sys2 file and open */
	if ((sys2 = fopen(argv[3],"r")) == NULL)
	{	fprintf(stderr,"Can't open the sys2 file %s.\n",argv[3]);
		exit_flag = 1;
	}
	else
	/* Check to see if args are OK */
	if ((argc == 5) && (strcmp(argv[4],"-q") != 0) && (strcmp(argv[4],"-s") != 0) && (strcmp(argv[4],"-c") != 0) && (strcmp(argv[4],"-C") != 0))
	{
		fprintf(stderr,"Invalid switch must be -c for just recall & precision summary, -C for recall & precision detail, -s for summary or -q for detail for each query.\n");
		exit_flag = 1;
	}
	else
	/*get on with comparing system 1 v system 2 if all OK */
	{
		if ((argc == 5) && (strcmp(argv[4],"-s") == 0))
			detail = 1;
		if ((argc == 5) && (strcmp(argv[4],"-q") == 0))
			detail = 2;
		if ((argc == 5) && (strcmp(argv[4],"-c") == 0))
			detail = 3;
		if ((argc == 5) && (strcmp(argv[4],"-C") == 0))
			detail = 4;
		printf("Comparing %s *(sys1) against %s **(sys2)\n",argv[2],argv[3]);
		fprintf(stderr,"\nPlease wait ");
		mainloop(qrel,sys1,sys2,detail);
		
		/* close every file */
		fclose(qrel);
		fclose(sys1);
		fclose(sys2);
	}
	
	exit (exit_flag);
}




/* mainloop - we now have the files so we can get each query in turn */
/* and process it.                                                   */

void mainloop(FILE *qrel, FILE *sys1, FILE *sys2, int detail)
{
	/* Keep a list of document identifiers for each query */
	recptr	qrels;
	recptr	sys1docs;
	recptr	sys2docs;
	
	/* An array to store the summary lines to print out at the end */
	char 	summary_line [no_summary_lines] [summary_line_size];
	int	count1;
	int	count2;
	
	/* we need to keep a track of where we are in the files */
	int	qno;
	int	qrelqno;
	char	qreldoc[idlen];
	int	relmarker;
	int 	sys1qno;
	char	sys1doc[idlen];
	int	sys2qno;
	char	sys2doc[idlen];
	int	qrel_eof_flag;
	int 	sys1_eof_flag;
	int	sys2_eof_flag;

        /* somewhere to store the recall and precision and averages data */
        
        rpinfo	sys1rp;
        rpinfo	sys2rp;
        avginfo	avg;

	
	/* Filler data for scanf */
	char	filler1[idlen];
	char	filler2[idlen];
	char	filler3[idlen];
	char	filler4[idlen];
	
	/* Initialise */
	
	sys1rp = (rpinfo) malloc (sizeof(rpinf));
	sys2rp = (rpinfo) malloc (sizeof(rpinf));
	init_rpinfo(sys1rp);
	init_rpinfo(sys2rp);
	
	avg = (avginfo) malloc (sizeof(avginf));
	avg->tot_no_relevant = 0;
	avg->tot_precision_diff = 0.0;
	avg->tot_precision_at_100_diff = 0.0;
	avg->tot_common100 = 0.0;
	avg->tot_common1000 = 0.0;
	avg->tot_unique1100 = 0.0;
	avg->tot_unique11000 = 0.0;
	avg->tot_unique2100 = 0.0;
	avg->tot_unique21000 = 0.0;
	avg->better_sys1 = 0;
	avg->better_sys2 = 0;
	avg->tot_better_diff_sys1 = 0.0;
	avg->tot_better_diff_sys2 = 0.0;
	avg->tot_better_rel_sys1 = 0.0;
	avg->tot_better_rel_sys2 = 0.0;
	avg->tot_better_unique100_sys1 = 0.0;
	avg->tot_better_unique100_sys2 = 0.0;
	avg->tot_better_unique1000_sys1 = 0.0;
	avg->tot_better_unique1000_sys2 = 0.0;
	
	
	count1 = 0;
	
	/* Read first record of files to get some data */
	qrel_eof_flag = fscanf(qrel, "%d %s %s %d", &qrelqno, filler1, qreldoc, &relmarker);
	sys1_eof_flag = fscanf(sys1, "%d %s %s %s %s %s", &sys1qno, filler1, sys1doc, filler2, filler3, filler4);
	sys2_eof_flag = fscanf(sys2, "%d %s %s %s %s %s", &sys2qno, filler1, sys2doc, filler2, filler3, filler4);
	
	/* Now we can process them */
	while (qrel_eof_flag != -1)
	{
		/* Initialise the variables */
		qrels = NULL;
		sys1docs = NULL;
		sys2docs = NULL;
		qno = qrelqno;
		
		/*get relevant docs into qrels list */
		qrels = getrels(qrel, qrels, &qrel_eof_flag, &qrelqno, qreldoc, &relmarker);
	
		/* Next read in all docs in ranking for sys1 */
		sys1docs = getdocs(sys1, sys1docs, &sys1_eof_flag, qno, &sys1qno, sys1doc);
		
		/* Next read in all docs in ranking for sys2 */
		sys2docs = getdocs(sys2, sys2docs, &sys2_eof_flag, qno, &sys2qno, sys2doc);
		
		/* Got all data for this query now we can compare */
		compare(qno, qrels, sys1docs, sys2docs, detail, summary_line [count1], sys1rp, sys2rp, avg);
		count1++;
		
		/* Delete the doc ids and start new query */
		deletedocs(qrels);
		deletedocs(sys1docs);
		deletedocs(sys2docs);
		fprintf(stderr,".");
		fflush(stderr);	
	}
	fprintf(stderr," Done\n\n");
	/* Now print out all the summaries for the run */
	
	/* Calc rp averages */
	if (count1 > 0)
	{
		for (count2 = 0; count2 < no_precision_levels; count2++)
		{
			sys1rp->ip[count2] /= count1;
			sys2rp->ip[count2] /= count1;
		}
		for (count2 = 0; count2 < no_doc_levels; count2++)
		{
			sys1rp->dp[count2] /= count1;
			sys2rp->dp[count2] /= count1;
		}
		sys1rp->avg_precision /= count1;
		sys2rp->avg_precision /= count1;
		sys1rp->r_precision /= count1;
		sys2rp->r_precision /= count1;
	
		/* Print them if requested */
		if (detail > 0)
			print_tables(0, avg->tot_no_relevant, sys1rp, sys2rp);
	

		/* then print out summary if requested */
		if (detail < 3)
		{
			printf("\n");
			printf("      | No. |  Ret Rel  | Avg Precision |    %%Diff    || Precision @100|    %%Diff    ||   Common    || Unique Sys1 || Unique Sys2\n");
			printf("Topic | Rel |sys1 |sys2 |  sys1 |  sys2 |s1>s2 | s2>s1||  sys1 |  sys2 |s1>s2 | s2>s1|| 100  | 1000 || 100  | 1000 || 100  | 1000\n");
			printf("------+-----+-----+-----+-------+-------+------|------++-------+-------+------|------++------+------++------+------++------+------\n");
			for (qno = 0; qno < count1; qno++)
				printf("%s\n",summary_line[qno]);
			printf("------+-----+-----+-----+-------+-------+------+------++-------+-------+------+------++------+------++------+------++------+------\n");
		
		
			/* Calc summary line averages */
	
			avg->tot_no_relevant /= count1;
			avg->tot_precision_diff /= count1;
			avg->tot_precision_at_100_diff /= count1;
			avg->tot_common100 /= count1;
			avg->tot_common1000 /= count1;
			avg->tot_unique1100 /= count1;
			avg->tot_unique11000 /= count1;
			avg->tot_unique2100 /= count1;
			avg->tot_unique21000 /= count1;

			printf("  Avg |%5.1f|%5.1f|%5.1f| %1.4f| %1.4f| %8.2f    || %1.4f| %1.4f| %8.2f    ||%6.2f|%6.2f||%6.2f|%6.2f||%6.2f|%6.2f\n", avg->tot_no_relevant, ((float) sys1rp->retrieved_relevant / count1), ((float) sys2rp->retrieved_relevant / count1), sys1rp->avg_precision, sys2rp->avg_precision, avg->tot_precision_diff, sys1rp->dp[5], sys2rp->dp[5], avg->tot_precision_at_100_diff, avg->tot_common100, avg->tot_common1000, avg->tot_unique1100, avg->tot_unique11000, avg->tot_unique2100, avg->tot_unique21000);

			if (avg->better_sys1 > 0)
			{
				avg->tot_better_diff_sys1 /= avg->better_sys1;
				avg->tot_better_rel_sys1 /= avg->better_sys1;
				avg->tot_better_unique100_sys1 /= avg->better_sys1;
				avg->tot_better_unique1000_sys1 /= avg->better_sys1;
			}
			if (avg->better_sys2 > 0)
			{
				avg->tot_better_diff_sys2 /= avg->better_sys2;
				avg->tot_better_rel_sys2 /= avg->better_sys2;
				avg->tot_better_unique100_sys2 /= avg->better_sys2;
				avg->tot_better_unique1000_sys2 /= avg->better_sys2;
			}
	
			printf("\n\n\t\t\t| Sys 1 | Sys 2 |\n");
			printf("\t\t\t|       |       |\n");
			printf("Topic Count\t\t|  %5d|  %5d|\n", avg->better_sys1, avg->better_sys2);
			printf("Average Percent Diff\t|%7.2f|%7.2f|\n", avg->tot_better_diff_sys1, avg->tot_better_diff_sys2);
			printf("Average Number Relevant\t|%7.2f|%7.2f|\n", avg->tot_better_rel_sys1, avg->tot_better_rel_sys2);
			printf("Average Unique @100\t|%7.2f|%7.2f|\n", avg->tot_better_unique100_sys1, avg->tot_better_unique100_sys2);
			printf("Average Unique @1000\t|%7.2f|%7.2f|\n\n\n", avg->tot_better_unique1000_sys1, avg->tot_better_unique1000_sys2);

			printf("\t\t\t|Topics |At 100 |At 1000|\n");
			if ((count1 - (avg->better_sys1 + avg->better_sys2)) > 0)
				printf("Comparable topics\t|%7d|%7.2f|%7.2f|\n", (count1 - (avg->better_sys1 + avg->better_sys2)), (avg->tot_comp100 / (count1 - (avg->better_sys1 + avg->better_sys2))), (avg->tot_comp1000 / (count1 - (avg->better_sys1 + avg->better_sys2))));
			else
				printf("Comparable topics\t|      0|      0|      0|\n");
			if ((avg->better_sys1 + avg->better_sys2) > 0)
				printf("Different topics\t|%7d|%7.2f|%7.2f|\n\n\n", (avg->better_sys1 + avg->better_sys2), (avg->tot_diff100 / (avg->better_sys1 + avg->better_sys2)), (avg->tot_diff1000 / (avg->better_sys1 + avg->better_sys2)));
			else
				printf("Different topics\t|      0|      0|      0|\n");
			
		}
	}
	else
	{
		fprintf(stderr,"Compare did not have enough information to do any comparisons!");
	}

}



/* Procedure fills up a list with relevant documents for query from qrels */

recptr getrels(FILE *qrel, recptr q1, int *eof_flag, int *qno, char *docid, int *relmarker)
{
	recptr	q2;
	int	qrelqno;
	
	/* Filler data for scanf */
	char	filler1[idlen];
	
	qrelqno = *qno;
	/* Read in all those records with qno = qrelqno */
	while ((*eof_flag != -1) && (*qno == qrelqno))
	{
		/* only interested if it is relevant */
		if (*relmarker == 1)
		{
			/* add to list (top) */
			q2 = (recptr) malloc (sizeof(rec));
			strcpy(q2->id,docid);
			q2->dnext = q1;
			q1 = q2;
		}
		
		/* Read next record of files to get some data */
		*eof_flag = fscanf(qrel, "%d %s %s %d", qno, filler1, docid, relmarker);
	}
	return q1;
}



/* Procedure fills up a doc list with items from rankings (in order) */
recptr getdocs(FILE *sys, recptr d1, int *eof_flag, int qno, int *dqno, char *docid)
{
	recptr 	d2;
	recptr 	d3;
	recptr 	d4;

	/* Fillers for scanf */
	char	filler1[idlen];
	char	filler2[idlen];
	char	filler3[idlen];
	char	filler4[idlen];

	/* Read in all the document ranks for this qno */
	while ((*eof_flag != -1) && (*dqno <= qno))
	{
		if (*dqno == qno)
		{
		/* add first record */
		if (d1 == NULL)
		{
			d1 = (recptr) malloc (sizeof(rec));
			strcpy(d1->id,docid);
			d1->dnext = NULL;
		}
		/*or add to end of list */
		else
		{
			/* traverse list to end */
			d2 = d1;
			while (d2 != NULL)
			{
				d3 = d2;
				d2 = d2->dnext;
			}
			/* when at end of list add the record */
			d4 = (recptr) malloc (sizeof(rec));
			strcpy(d4->id,docid);
			d4->dnext = NULL;
			d3->dnext = d4;
		}
		}
		/* Read next record */
		*eof_flag = fscanf(sys, "%d %s %s %s %s %s", dqno, filler1, docid, filler2, filler3, filler4);
	}
	return d1;
}



/* main work horse - does the comparison of two lists given rel docs list */

void compare(int qno, recptr r1, recptr d1, recptr d2, int detail, char *summaryln, rpinfo sys1totrp, rpinfo sys2totrp, avginfo avg)
{
	rpinfo	sys1rp;								/* RP info for this ranking for sys1 */
	rpinfo	sys2rp;								/* for sys2 - both are added to sysNtotrp */
	int	no_relevant;							/* number of rel docs */
	float	diffa;								/* Needed to avoid div by zero */
	float	diffb;								/* for two differences */
	char	flag [3];							/* Indicates best system */
	float	common100;							/* %common rel at 100 */
	float	common1000;							/* %common rel at 1000 */
	float	unique1100;							/* %unique sys1 at 100 */
	float	unique11000;							/* %unique sys1 at 1000 */
	float	unique2100;							/* sys2 100 */
	float	unique21000;							/* sys2 1000 */
	int	count;								/* loop counter */
	float	total;								/* keep a total */
	char	diffas [14];							/* % diff string */
	char	diffbs [14];							/* % diff string */

	
	/* Initialise */
	sys1rp = (rpinfo) malloc (sizeof(rpinf));
	sys2rp = (rpinfo) malloc (sizeof(rpinf));
	init_rpinfo(sys1rp);
	init_rpinfo(sys2rp);
	no_relevant = length(r1);
	
	/* Only work stuff out if we have relevant docs */
	if (no_relevant > 0)
	{
		/* Next work through list to get r-p values */
		recall_precision(r1, d1, no_relevant, sys1rp);
		
		/* Next work through list to get r-p values */
		recall_precision(r1, d2, no_relevant, sys2rp);
	}


	/* If we have detail set to 2 or 4 then we print out the individual query data */
	if ((detail == 2) || (detail == 4))
		print_tables(qno, no_relevant, sys1rp, sys2rp);

	/* Add the above levels to the totals */
	for (count = 0; count < no_precision_levels; count++)
	{
		sys1totrp->ip[count] += sys1rp->ip[count];
		sys2totrp->ip[count] += sys2rp->ip[count];
	}
	for (count = 0; count < no_doc_levels; count++)
	{
		sys1totrp->dp[count] += sys1rp->dp[count];
		sys2totrp->dp[count] += sys2rp->dp[count];
	}
	sys1totrp->avg_precision += sys1rp->avg_precision;
	sys2totrp->avg_precision += sys2rp->avg_precision;
	sys1totrp->r_precision += sys1rp->r_precision;
	sys2totrp->r_precision += sys2rp->r_precision;
	sys1totrp->retrieved += sys1rp->retrieved;
	sys2totrp->retrieved += sys2rp->retrieved;
	sys1totrp->retrieved_relevant += sys1rp->retrieved_relevant;
	sys2totrp->retrieved_relevant += sys2rp->retrieved_relevant;
	
	
	
	
	/* Work out common and unique records at various levels */
	
	if (detail < 3)
	{
		common100 = (float) common(r1,d1,d2,100);
		unique1100 = (float) unique(r1,d1,d2,100);
		unique2100 = (float) unique(r1,d2,d1,100);
	}
	else
	{
		common100 = 0;
		unique1100 = 0;
		unique2100 = 0;
	}
	total = common100 + unique1100 + unique2100;
	if (total > 0)
	{
		common100 = (common100 / total) * 100;
		unique1100 = (unique1100 / total) * 100;
		unique2100 = (unique2100 / total) * 100;
	}
	else
	{
		common100 = 0;
		unique1100 = 0;
		unique2100 = 0;
	}
	
	if (detail < 3)
	{
		unique11000 = (float) unique(r1,d1,d2,1000);
		common1000 = (float) common(r1,d1,d2,1000);
		unique21000 = (float) unique(r1,d2,d1,1000);
	}
	else
	{
		common1000 = 0;
		unique11000 = 0;
		unique21000 = 0;
	}
	total = common1000 + unique11000 + unique21000;
	if (total > 0)
	{
		common1000 = (common1000 / total) * 100;
		unique11000 = (unique11000 / total) * 100;
		unique21000 = (unique21000 / total) * 100;
	}
	else
	{
		common1000 = 0;
		unique11000 = 0;
		unique21000 = 0;
	}
	
/* I feel that these figures should be used as they are representitive of all the relevant documents 

	if (no_relevant > 0)
	{
		common100 = ((float) common(r1,d1,d2,100) / (float) no_relevant) * 100;
		common1000 = ((float) common(r1,d1,d2,1000) / (float) no_relevant) * 100;
		unique1100 = ((float) unique(r1,d1,d2,100) / (float) no_relevant) * 100;
		unique11000 = ((float) unique(r1,d1,d2,1000) / (float) no_relevant) * 100;
		unique2100 = ((float) unique(r1,d2,d1,100) / (float) no_relevant) * 100;
		unique21000 = ((float) unique(r1,d2,d1,1000) / (float) no_relevant) * 100;
	}
	else
	{
		common100 = 0.0;
		common1000 = 0.0;
		unique1100 = 0.0;
		unique11000 = 0.0;
		unique2100 = 0.0;
		unique21000 = 0.0;
	}
*/
	

	
	/* Next work out the difference between the two systems */
	if ((sys1rp->avg_precision > 0) && (sys2rp->avg_precision > 0))
		if (sys2rp->avg_precision < sys1rp->avg_precision)
			diffa = ((sys2rp->avg_precision - sys1rp->avg_precision) / sys1rp->avg_precision) * 100.00;
		else
			diffa = ((sys1rp->avg_precision - sys2rp->avg_precision) / sys2rp->avg_precision) * -100.00;
	else
		if (sys1rp->avg_precision > 0)
			diffa = -99.99;
		else
			diffa = 99.99;
			
	if (diffa <= 0)
		sprintf(diffas," %5.2f|      ", (diffa * -1));
	else
		sprintf(diffas,"      | %5.2f", diffa);

	if ((sys1rp->dp[5] > 0) && (sys2rp->dp[5] > 0))
		if (sys2rp->dp[5] < sys1rp->dp[5])
			diffb = ((sys2rp->dp[5] - sys1rp->dp[5]) / sys1rp->dp[5]) * 100.00;
		else
			diffb = ((sys1rp->dp[5] - sys2rp->dp[5]) / sys2rp->dp[5]) * -100.00;
	else
		if (sys1rp->dp[5] > 0)
			diffb = -99.99;
		else
			diffb = 99.99;

	if (diffb <= 0)
		sprintf(diffbs," %5.2f|      ", (diffb * -1));
	else
		sprintf(diffbs,"      | %5.2f", diffb);


	/* Add an indicator just in case and total up*/
	if ((diffa <= -20.0) || ((sys1rp->retrieved_relevant > 0) && (sys2rp->retrieved_relevant == 0)))
	{
		strcpy(flag," *");
		avg->better_sys1++;
		avg->tot_better_diff_sys1 += (-1 * diffa);
		avg->tot_better_rel_sys1 += no_relevant; /* Perhaps should be sysrp1->retrieved_relevant */
		avg->tot_better_unique100_sys1 += unique1100;
		avg->tot_better_unique1000_sys1 += unique11000;
		avg->tot_diff100 += common100;
		avg->tot_diff1000 += common1000;
	}
	else
	if ((diffa >= 20.0) || ((sys2rp->retrieved_relevant > 0) && (sys1rp->retrieved_relevant == 0)))
	{
		strcpy(flag,"**");
		avg->better_sys2++;
		avg->tot_better_diff_sys2 += diffa;
		avg->tot_better_rel_sys2 += no_relevant; /* perhaps should be sysrp2->retrieved_relevant */
		avg->tot_better_unique100_sys2 += unique2100;
		avg->tot_better_unique1000_sys2 += unique21000;
		avg->tot_diff100 += common100;
		avg->tot_diff1000 += common1000;
	}
	else
	{
		strcpy(flag,"  ");
		avg->tot_comp100 += common100;
		avg->tot_comp1000 += common1000;
	}
		

        /* Put all the summary info into a string to print at the end */
	sprintf(summaryln, "%s %3d| %4d| %4ld| %4ld| %1.4f| %1.4f|%s|| %1.4f| %1.4f|%s||%6.2f|%6.2f||%6.2f|%6.2f||%6.2f|%6.2f",flag, qno, no_relevant, sys1rp->retrieved_relevant, sys2rp->retrieved_relevant, sys1rp->avg_precision, sys2rp->avg_precision, diffas, sys1rp->dp[5], sys2rp->dp[5], diffbs, common100, common1000, unique1100, unique11000, unique2100, unique21000);


	/* Update all the totals for the averge info */
	
	avg->tot_no_relevant += no_relevant;
	if (diffa < 0)
		avg->tot_precision_diff += diffa * -1;
	else
		avg->tot_precision_diff += diffa;
	if (diffb < 0)
		avg->tot_precision_at_100_diff += diffb * -1;
	else
		avg->tot_precision_at_100_diff += diffb;
	avg->tot_common100 += common100;
	avg->tot_common1000 += common1000;
	avg->tot_unique1100 += unique1100;
	avg->tot_unique11000 += unique11000;
	avg->tot_unique2100 += unique2100;
	avg->tot_unique21000 += unique21000;

	/* OK now get rid of any extra bits of variables we might not need */
	free(sys1rp);
	free(sys2rp);
	
}




/* Calc recall and precision for this ranking */

void recall_precision(recptr r1, recptr d1, int no_rel, rpinfo sysrp)
{
	recptr 	d2;		/* We are looking at this list */
	float	recall = 0;	/* Recall measure (no_ret_rel / no_rel) */
	float	precision = 0;	/* Precision measure (no_ret_rel) / no_ret) */
	int	count;		/* Loop counter */
	float	totprec = 0;	/* Totaler for precision */
	float	prevrecall = 0; /* Previous recall */
	
	d2 = d1;
	/* Trapse through the list */
	while (d2 != NULL)
	{
		/* got a document */
		sysrp->retrieved++;
		
		/* If we find a relevant document then */
		if (member(d2->id,r1) > 0)
		{
			sysrp->retrieved_relevant++;		
			recall =  (float) sysrp->retrieved_relevant / no_rel;
		}
		
		/* Change the precision */
		precision =  (float) sysrp->retrieved_relevant /  sysrp->retrieved;
		
		/* A change in recall means a change in avergae precision totaler */
		if (recall > prevrecall)
		{		
			totprec += precision;
			prevrecall = recall;
		}
		
		/* put the precision in the no_precision_levels point recall scale */
		for (count = 0; count < no_precision_levels; count++)
			/* if its the highest precision for that point */
			if ((recall >= ((float) count / (no_precision_levels - 1))) && (sysrp->ip[count] < precision))
				sysrp->ip[count] = precision;			
				
		/* Now look to see if we are at the N docs precision level as defined in doclevels */
		for (count = 0; count < no_doc_levels; count++)
			/* If we are at a point then save the precision */
			if (doc_levels[count] == sysrp->retrieved)
				sysrp->dp[count] = precision;
				
		/* This saves the R-precision score */
		if (no_rel == sysrp->retrieved)
			sysrp->r_precision = precision;
			
		/* Move on one */
		d2 = d2->dnext;
	}

	/* Work out the avg non-interpolated precision */
	sysrp->avg_precision = totprec / (float) no_rel;

}




/* Looks at the number of common records given a threshold */

int common(recptr r, recptr d1, recptr d2, int threshold)
{
	recptr	d3;	/* The list to traverse */
	int	m;	/* The number common within thershold */
	int	n;	/* How far down the list we are */

	d3 = d1;
	m = 0;
	n = 1;
	while ((d3 != NULL) && (n <= threshold))
	{
		/* Common if the doc is relevant and apears under the threshold of the other list */
		if ((member(d3->id,r) > 0) && (member(d3->id,d2) > 0) && (member(d3->id,d2) <= threshold))
			m++;
		/* Next item please */
		n++;
		d3 = d3->dnext;
	}
	return m;
}




/* Looks at the number of unique records given a threshold */

int unique(recptr r, recptr d1, recptr d2, int threshold)
{
	recptr	d3;	/* List to traverse */
	int	m;	/* Number unique */
	int	n;	/* Where are we in the list */

	d3 = d1;
	m = 0;
	n = 1;
	while ((d3 != NULL) && (n <= threshold))
	{
		/* Unique if its relevant and not in the other list or its is in list and over threshold */
		if ((member(d3->id,r) > 0) && ((member(d3->id,d2) == 0) || (member(d3->id,d2) > threshold)))
			m++;
		/* Next item to process */
		n++;
		d3 = d3->dnext;
	}
	return m;
}



/* print out tables of data */

void print_tables(int qno, long no_relevant, rpinfo sys1rp, rpinfo sys2rp)
{
	int count;

	if (qno == 0)
		printf("Average measures over all the queries. \n");
	else
		printf("Querid : %d\n",qno);
	printf("Total number of documents over all queries\n");
	printf("    Retrieved:\t%ld\t%ld\n",sys1rp->retrieved,sys2rp->retrieved);
	printf("    Relevant:\t%ld\t%ld\n",no_relevant,no_relevant);
	printf("    Rel_ret:\t%ld\t%ld\n",sys1rp->retrieved_relevant,sys2rp->retrieved_relevant);
	printf("Interpolated Recall - Precision Averages - sys1 & sys2 :\n");
	for (count = 0; count < no_precision_levels; count++)
		printf("    at %1.2f\t%1.4f\t%1.4f\n", ((float) count / (no_precision_levels - 1)), sys1rp->ip[count], sys2rp->ip[count]);
	printf("Average precision (non-interpolated) over all rel docs\n");
	printf("\t\t%1.4f\t%1.4f\n", sys1rp->avg_precision, sys2rp->avg_precision);
	printf("Precision - sys1 & sys2 :\n");
	for (count = 0; count < no_doc_levels; count++)
		printf("  At%5d docs:\t%1.4f\t%1.4f\n", doc_levels[count], sys1rp->dp[count], sys2rp->dp[count]);
	printf("R-Precision (precision after R (= num_rel for a query) docs retrieved):\n");
   	printf("   Exact:\t%1.4f\t%1.4f\n\n", sys1rp->r_precision, sys2rp->r_precision);		
}



/* Delete the items in this list otherwise we may get loads of unwanted lists */
void deletedocs(recptr d1)
{
	recptr 	d2;
	
	d2 = d1;/* Traverse the list deleting all doc ids */
	while (d2 != NULL)
	{	
		d1 = d2->dnext;
		free(d2);
		d2 = d1;
	}
}



/* Is adoc id in the list return 1 if yes else 0 */
int member(char *docid, recptr d1)
{
      	int 	n; /* 0 = not found, > 0 = found (also the index of item in list) */
      	recptr 	d2;

	
	/* Need to use this to see if our docid is in a list */
      	d2 = d1;
      	n = 1;
      	/* Traverse through the list to find it */
      	while ((d2 != NULL) && (strcmp(d2->id,docid) != 0))
      	{	
      		n++;
        	d2 = d2->dnext;
      	}
      	/* have we found it ?? */
      	if (d2 == NULL)
         	n = 0;
      	return n;
}



/* Find the length of a list */
int length(recptr d1)
{
	recptr 	d2;
	int	n;
	
	d2 = d1;
	n = 0;
	/* Count all elements in the list */
	while (d2 != NULL)
	{
		d2 = d2->dnext;
		n++;
	}
	return n;
}


/* generic module to initialise any rpinfo module */
void init_rpinfo(rpinfo sysrp)
{
	int loop;
	
	for (loop = 0; loop < no_precision_levels; loop++)
		sysrp->ip[loop] = 0.0;
	for (loop = 0; loop < no_doc_levels; loop++)
		sysrp->dp[loop] = 0.0;
	sysrp->avg_precision = 0.0;
	sysrp->r_precision = 0.0;	
	sysrp->retrieved = 0;
	sysrp->retrieved_relevant = 0;
}
