/******************************************************************************

  rtgsubs.c  SW June 95

  Routing term selection routines

  startup_bss()
    Initialises BSS, opens database (in environment) etc
    Returns -1 on failure

  setup_termrecs(input_file)
    Reads term records from a file, makes sets for each, tentatively
    term records just consisting of <term>:<weight>. Puts them in global
    arrays of terms, weights, sets, np
    Returns number of termrecs read and processed, or -1 on failure

  double brookes(a, rels, N, R)
    Function to compute the Brookes measure for set "a" wrt set "rels" of
    relevant docs (see SER working paper 3 of Jun3 3 95).
    N & R have the usual meanings.
    It is assumed there's a database open, etc.
    Returns the Brookes measure, or a very large number on failure

  make_relset(relsfile, &R) 

    Make set from open file of DOCNOs, one per line
    Returns relset and writes R to the supplied address, or returns - on error
    Assumes attribute is called "dn"

  Etc, etc

******************************************************************************/

#include "rtg_select.h"

char *terms[MAXTERMS] ;
double weights[MAXTERMS] ;
double minweights[MAXTERMS] ;	/* Weight of last reldoc */
int sets[MAXTERMS] ;
int np[MAXTERMS] ;
int r_[MAXTERMS] ;
double br[MAXTERMS] ;
int incl[MAXTERMS] ;
int examined[MAXTERMS] ;
int successive_failures[MAXTERMS] ;
int dropped[MAXTERMS] ;
int skipped[MAXTERMS] ;
double assoc[MAXTERMS][MAXTERMS] ;

struct term Terms[MAXTERMS] ;
struct state States[MAXSTATES], *C_state, *Best_state, *Realbest_state ;
int Numstates ;			/* Number used */
struct state *Stage_final_state[MAXSTATES - 3] ;

int Hold = 0 ;
int Debug = DEBUG ;
int do_minwts = DO_MINWTS ;
int res ;
int silent = 0 ;
int Mark_early = 0 ;
int Use_cutoff = 0 ;
double Cutoff_factor = 0 ;
double Tar_factor = TAR_FACTOR ;
char *Op = OP ;
double k1 = 1.2, b = .75 ;

/******************************************************************************

  startup_bss() SW June 95
  
  Initialises BSS, opens database (in environment) etc

******************************************************************************/

startup_bss()
{
  char *envp ;
  char command[1024], *response ;
  iinit() ;
  if ( (envp = getenv("BSS_DB")) && *envp ) {
    sprintf(command, "ch %s", envp) ;
    I0 ;
  }
  else {
    fprintf(stderr, "No database in environment\n") ;
    goto error ;
  }
  if ( (envp = getenv("BSS_ATTRIBUTE")) && *envp ) {
    sprintf(command, "set a=%s", envp) ;
    I0 ;
  }
  if ( (envp = getenv("BSS_SETARGS")) && *envp ) {
    sprintf(command, "set %s", envp) ;
    I0;
  }
  return 0 ;
 error:
  return -1 ;
}


/******************************************************************************

  setup_termrecs() SW June 95

  Reads term records from a file, makes sets for each, tentatively
  term records just consisting of <term>:<weight>. Puts them in global
  arrays of terms, weights, sets, np

******************************************************************************/

setup_termrecs(termfile, relset, R)
     FILE *termfile ;
     int relset ;
     int R ;
{
  char buf[1024] ;
  char command[1024], *response ;
  int result ;
  int termnum = 0 ;
  int tempset ;
  double wt ;

  while ( (result = fscanf(termfile, " %[^:]:%lf", buf, &wt)) == 2 ) {
    if ( *buf == '#' ) continue ;
    if ( ! silent )
      fprintf(stderr, "Setting up term #%d \"%s\", wt %.3f\n", termnum, buf, wt) ;
    Terms[termnum].t = bss_strdup(buf) ;
    Terms[termnum].ow = wt ;
    /*sprintf(command, "f notf=1 t=%s", buf) ;*/
    sprintf(command, "f t=%s", buf) ;
    I0q ;
    tempset = atoi(response + 1) ;
    Terms[termnum].np = atoi(strstr(response, "np=") + 3) ;
    if ( Terms[termnum].np == 0 ) {
      fprintf(stderr, "No postings, term=%s, omitting it\n",
	      Terms[termnum].t) ;
      delete_set(tempset) ;
      bss_free(Terms[termnum].t) ;
      continue ;
    }
    sprintf(command, "f s=%d w=%f op=%s k1=%f bm25b=%f notf=1 hold=%d",
	    tempset, Terms[termnum].ow, Op, k1, b, Hold) ;
    I0 ;
    if ( ! silent && ! Mark_early ) fprintf(stderr, "%s", response) ;
    Terms[termnum].set = atoi(response + 1) ;
    sprintf(command, "del s=%d", tempset) ;
    I0q ;
    if ( Mark_early ) {
      sprintf(command, "f s=%d s=%d op=mark hold=%d",
	      Terms[termnum].set, relset, Hold) ;
      I0 ;
      if ( ! silent ) fprintf(stderr, "%s", response) ;
      tempset = atoi(response + 1) ;
      delete_set(Terms[termnum].set) ;
      Terms[termnum].set = tempset ;

      if ( ! silent ) {
	/* sprintf(command, "ds_trec s=%d bigr=%d n=%d",
	   Terms[termnum].set, R, 1000) ; */
	sprintf(command, "ds_trec s=%d bigr=%d n=%d",
		Terms[termnum].set, R, 0) ;
	I0 ;
	fprintf(stderr, "%s", response) ;
      }

      /* minwts doesn't appear to be useful */
      if ( do_minwts ) {
	int rr ;
	double w = wt_of_last_common(Terms[termnum].set, relset, &rr) ;
	if ( ! silent )
	  fprintf(stderr, "Recall %.3f, weight of lowest rel is %.3f\n",
		  (double)rr/(double)R, w) ;
	sprintf(command, "f s=%d aw=%f hold=%d",
		Terms[termnum].set, w, Hold) ;
	I0 ;
	fprintf(stderr, "%s\n", response) ;
	tempset = atoi(response + 1) ;
	delete_set(Terms[termnum].set) ;
	Terms[termnum].set = tempset ;
      }
    }
    termnum++ ;
  }
  return termnum ;
 error:
  return -1 ;
}

/******************************************************************************

  remake_current_set(target)

  Makes set from terms with incl[] true
  Warning: June97: can't be used if weight has changed
******************************************************************************/
remake_current_set(target, aw)
     int target ;
     double aw ;
{
  int j ;
  int set ;
  char command[16384], *response ;
  sprintf(command, "f") ;
  for ( j = 0 ; j < MAXTERMS ; j++ ) {
    if ( incl[j] )
      sprintf(command + strlen(command), " s=%d", Terms[j].set) ;
  }
  sprintf(command + strlen(command), " target=%d aw=%f", target, aw) ;
  I0 ;
  if ( ! silent ) fprintf(stderr,
			  "Remade current set, %s", response) ;
  set = atoi(response + 1) ;
  return set ;
error:
  return -1 ;
}

/******************************************************************************

  reduce_set()

******************************************************************************/
reduce_set(set, aw)
     int set ;
     double aw ;
{
  int new_set ;
  char command[1024], *response ;
  sprintf(command, "f s=%d aw=%f hold=%d", set, aw, Hold) ;
  I0 ;
  if ( ! silent ) fprintf(stderr, "%s\n", response) ;
  new_set = atoi(response + 1) ;
  return new_set ;

error:
  return -1 ;
}


/******************************************************************************

  brookes()  SW June 95

  Function to compute the Brookes measure for set a wrt set rels of relevant
  docs (see SER working paper 3 of Jun3 3 95).

  If R = 0 it uses the size of the set intersection as R

  It is assumed there's a database open, etc.

  Returns Brookes measure and writes size of the set intersection to
  a supplied int pointer.

******************************************************************************/
double
brookes(a, rels, N, R, relret)
     int a, rels ;		/*  Set numbers */
     int R ;			/* Size of r */
     int N ;			/* Size of collection */
     int *relret ;			/* To "return" size of intersection */
{
  double sum_an, sum_ar, sumsq_an, sumsq_ar ;
  double sum_a, sumsq_a ;
  double m_ar, m_an, v_ar, v_an ;
  double result = 1e34 ;
  char command[1024], *response ;

  if ( N - R <= 1 ) {
    fprintf(stderr, "can't do brookes(., ., %d, %d)\n", N, R) ;
    return result ;
  }
  sprintf(command, "f s=%d s=%d op=and2 save=0", a, rels) ;
  I0q ;
  *relret = atoi(strstr(response, "np=") + 3) ;
  if ( R == 0 ) R = *relret ;	/* Case where stats to be done on top_num only,
				   N should be top_num */
  sum_ar = atof(strstr(response, "sum=") + 4) ;
  sumsq_ar = atof(strstr(response, "sumsq=") + 6) ;
  sprintf(command, "ds s=%d", a) ;
  I0q ;
  sum_a = atof(strstr(response, "sum=") + 4) ;
  sumsq_a = atof(strstr(response, "sumsq=") + 6) ;

  sum_an = sum_a - sum_ar ;
  sumsq_an = sumsq_a - sumsq_ar ;

  if ( sumsq_ar < 0.0001 || sumsq_an < 0.0001 ) {
    fprintf(stderr,
	    "brookes() returning zero, zero recall or maybe no stats in set(s)\n") ;
    fprintf(stderr, "sum_a=%f sumsq_a=%f sum_ar=%f sumsq_ar=%f\n",
	    sum_a, sumsq_a, sum_ar, sumsq_ar) ;
    result = 0.0 ;
    goto error ;
  }
  if ( R == 0 ) R = *relret ;
  m_ar = sum_ar / R ; m_an = sum_an / (N - R) ;
  v_ar = (sumsq_ar - sum_ar * sum_ar / R) / (R - 1) ;
  v_an = (sumsq_an - sum_an * sum_an / (N - R)) / (N - R - 1) ;
  if ( v_ar < 0 || v_an < 0 ) {
    fprintf(stderr,
  "brookes(set %d, set %d, %d, %d): negative variance (v_ar=%.3f, v_an=%.3f)\n",
	    a, rels, N, R, v_ar, v_an) ;
   fprintf(stderr, "(sumsq_ar=%.3f, sum_ar=%.3f, sumsq_an=%.3f, sum_an=%.3f\n",
	    sumsq_ar, sum_ar, sumsq_an, sum_an) ;
    goto error ;
  }
  result = (m_ar - m_an) / sqrt(v_ar + v_an) ;
  
 error:
  return result ;
}


/******************************************************************************

  diff_means()  SW June 95

  Function to compute the difference in means for set a wrt set rels of
  relevant docs.

  If R = 0 it uses the size of the set intersection as R

  It is assumed there's a database open, etc.

  Returns the difference in means and writes size of the set intersection to
  a supplied int pointer.

******************************************************************************/
double
diff_means(a, rels, N, R, relret)
     int a, rels ;		/*  Set numbers */
     int R ;			/* Size of rels */
     int N ;			/* Size of collection */
     int *relret ;			/* To "return" size of intersection */
{
  double sum_an, sum_ar ;
  double sum_a ;
  double m_ar, m_an, m_a ;
  double result = 0.0 ;
  char command[1024], *response ;

  sprintf(command, "f s=%d s=%d op=and2 save=0", a, rels) ;
  I0q ;
  *relret = atoi(strstr(response, "np=") + 3) ;
  sum_ar = atof(strstr(response, "sum=") + 4) ;
  sprintf(command, "ds s=%d", a) ;
  I0q ;
  sum_a = atof(strstr(response, "sum=") + 4) ;
  m_a = sum_a / N ;

  sum_an = sum_a - sum_ar ;

  if ( R == 0 ) R = *relret ;
  if ( N - R <= 0 || R <= 0 ) {
    fprintf(stderr, "can't do difference in means(., ., %d, %d)\n", N, R) ;
    return result ;
  }
  m_ar = sum_ar / R ; m_an = sum_an / (N - R) ;
  result = (m_ar - m_an) / m_a ;
 error:  
  return result ;
}


/******************************************************************************

  recall()  SW July 95

  Function to compute the recall for set a wrt set rels of
  relevant docs.

  Returns the recall proportion and writes size of the set intersection to
  a supplied int pointer.

******************************************************************************/
double
recall(a, rels, R, relret)
     int a, rels ;		/*  Set numbers */
     int R ;			/* Size of rels */
     int *relret ;			/* To "return" size of intersection */
{
  double result = 0.0 ;
  char command[1024], *response ;

  if ( R == 0 ) {		/* Determine it */
    sprintf(command, "f s=%d s=%d save=0 op=and", rels, rels) ;
    I0q ;
    R = atoi(strstr(response, "np=") + 3) ;
  }
  sprintf(command, "f s=%d s=%d op=and2 save=0", a, rels) ;
  I0q ;
  *relret = atoi(strstr(response, "np=") + 3) ;
  result = (double)*relret / R ;
 error:  
  return result ;
}


/******************************************************************************

  combined_measure()  SW July 95

  Computes precision at .5R, R and 2R and returns the average.

  Assumes size of set a is exactly 2R (see score())

******************************************************************************/
double
combined_measure(a, rels, R, relret)
     int a, rels ;		/*  Set numbers */
     int R ;			/* Size of rels */
     int *relret ;			/* To "return" size of intersection */
{
  double score = 0.0 ;
  double part_score ;
  int temp_set ;
  int n ;
  char command[1024], *response ;

  if ( R == 0 ) {		/* Determine it */
    sprintf(command, "f s=%d s=%d save=0 op=and", rels, rels) ;
    I0q ;
    R = atoi(strstr(response, "np=") + 3) ;
  }
  sprintf(command, "f s=%d s=%d op=and2 save=0", a, rels) ;
  I0q ;
  n = *relret = atoi(strstr(response, "np=") + 3) ;
  part_score = (double)*relret/(double)(2 * R) ;
  /* fprintf(stderr, "P(2R)=%.3f ", part_score) ; */
  score += part_score ;
  sprintf(command, "f s=%d top=%d", a, R) ;
  I0q ;
  temp_set = atoi(response + 1) ;
  sprintf(command, "f s=%d s=%d op=and2 save=0", temp_set, rels) ;
  I0q ;
  n = atoi(strstr(response, "np=") + 3) ;
  part_score = (double)n / (double)R ;
  /* fprintf(stderr, "P(R)=%.3f ", part_score) ;*/
  score += part_score ;
  delete_set(temp_set) ;
  sprintf(command, "f s=%d top=%d", a, (int)rint(0.5 * (double)R)) ;
  I0q ;
  temp_set = atoi(response + 1) ;
  sprintf(command, "f s=%d s=%d op=and2 save=0", temp_set, rels) ;
  I0q ;
  n = atoi(strstr(response, "np=") + 3) ;
  part_score = (double)n / (0.5 * (double)R) ;
  /* fprintf(stderr, "P(.5R)=%.3f\n", part_score) ;*/
  score += part_score ;
  delete_set(temp_set) ;
  return score / 3 ;
 error:
  return -1 ;
}


double
prec(n, a, rels, relret)	/* COmpute precision on given sets
				   (assumed setsize n) */
     int n, a, rels, *relret ;
{
  char command[1024], *response ;
  int i, temp_set ;
  double score = 0.0 ;

  sprintf(command, "f s=%d s=%d save=0 op=and2", a, rels) ;
  I0q ;
  *relret = atoi(strstr(response, "np=") + 3) ;
  score = (double)*relret / (double)n ;
  return score ;
 error:
  return score ;
}

double
avep(scoreset, rels, relret, cutoff, gran)
     int scoreset, rels, *relret, cutoff, gran ;
     /* Assumes size of scoreset is exactly cutoff */

{
  char command[1024], *response ;
  int top_set ;
  int j, step, count, r ;
  double result = 0.0 ;

  step = cutoff / gran ;
  top_set = scoreset ;
  for ( j = cutoff, count = 0 ; j >= 1 ; j -= step, count++ ) {
    if ( count > 0 ) {
      sprintf(command, "f s=%d top=%d", top_set, j) ;
      I0q ;
      delete_set(top_set) ;
      top_set = atoi(response + 1) ;
    }
    sprintf(command, "f s=%d s=%d op=and2 save=0", top_set, rels) ;
    I0q ;
    r = atoi(strstr(response, "np=") + 3) ;
    if ( count == 0 ) *relret = r ;
    result += (double)r/(double)j ;
  }
  if ( count > 0 ) {
    delete_set(top_set) ;
    result /= count ;
  }
  return result ;
 error:
  return -1.0 ;
}
  
double
trec_avep(scoreset, rels, relret, R, topnum, code)
     int scoreset, rels, *relret, R ;
     int topnum ;
     int code ;
     /* Assumes size of scoreset is exactly cutoff */
     /* Above no longer true, don't have to do topn (May 96) */
{
  char command[1024], *response ;
  int marked_set ;
  double Avep, P30, P100, P200, Rprec, Rcl ;
  double result = 0.0 ;

  /* If scoreset is already marked (and right size) don't need to mark. */

  if ( ! Mark_early ) {
    sprintf(command, "f s=%d s=%d op=mark", scoreset, rels) ;
    I0 ;
    marked_set = atoi(response + 1) ;
  }
  else marked_set = scoreset ;
  sprintf(command, "ds_trec s=%d bigr=%d n=%d", marked_set, R, topnum) ;
  I0 ;
  Avep = atof(strstr(response, "Avep=") + 5) ;
  P30 = atof(strstr(response, "P30=") + 4) ;
  P100 = atof(strstr(response, "P100=") + 5) ;
  P200 = atof(strstr(response, "P200=") + 5) ;
  Rprec = atof(strstr(response, "Rprec=") + 6) ;
  Rcl = atof(strstr(response, "Rcl=") + 4) ;
  *relret = rint(Rcl * (double)R) ;
  switch (code) {
  case T_AVEP_RCL:
    result = Avep * Tar_factor + Rcl * (1 - Tar_factor) ;
    break ;
  default:
    result = Avep ;
  }
  if ( ! Mark_early ) delete_set(marked_set) ;
  return result ;
 error:
  return -1.0 ;
}

double
ftrfuncs(scoreset, rels, relret, R, topnum, code)
     int scoreset, rels, *relret, R ;
     int topnum ;
     int code ;
     /* Assumes size of scoreset is exactly cutoff */
     /* Above no longer true, don't have to do topn (May 96) */
{
  char command[1024], *response ;
  int marked_set ;
  double Avep, Rcl ;
  double funcval, maxval, maxw ;
  int maxn ;
  double result = 0.0 ;

  /* If scoreset is already marked (and right size) don't need to mark. */

  if ( ! Mark_early ) {
    sprintf(command, "f s=%d s=%d op=mark", scoreset, rels) ;
    I0 ;
    marked_set = atoi(response + 1) ;
  }
  else marked_set = scoreset ;
  sprintf(command, "ds_trec s=%d bigr=%d n=%d code=%d", marked_set, R, topnum, code) ;
  I0 ;
  Avep = atof(strstr(response, "Avep=") + 5) ;
  Rcl = atof(strstr(response, "Rcl=") + 4) ;
  funcval = atof(strstr(response, "Val=") + 4) ;
  maxval = atof(strstr(response, "Maxval=") + 7) ;
  maxw = atof(strstr(response, "Wt=") + 3) ;
  maxn = atoi(strstr(response, "Maxn=") + 5) ;
  *relret = rint(Rcl * (double)R) ;
  result = maxval ;

  if ( ! Mark_early ) delete_set(marked_set) ;
  return result ;
 error:
  return -1.0 ;
}
  

double
cbm1(scoreset, rels, relret, R, N)
     int scoreset, rels, *relret, R, N ;
{
  /* (BROOKES/10 + TREC_AVEP) / 2 */
  return (brookes(scoreset, rels, N, R, relret) / 10 +
	  trec_avep(scoreset, rels, relret, R)) / 2 ;
}


double
score(a, rels, N, R, relret, top_num, code, gran)
     int a, rels ;
     int N, R ;
     int *relret ;
     int top_num ;
     int code ;
     int gran ;			/* Only used if AVEP (TEST) */
{
  int score_set ;
  int make_scoreset ;
  double score ;
  char command[1024], *response ;
  switch (code) {
  case COMBINED_MEASURE:
    top_num = 2 * R ;
    make_scoreset = 1 ;
    break ;
  case RPREC:
    top_num = R ;
    make_scoreset = 1 ;
    break ;
  case TREC_AVEP:
  case T_AVEP_RCL:
  case TS_LF1:
  case TS_LF2:
  case TS_LF3:
  case TS_NF1:
  case TS_NF2:
    make_scoreset = 0 ;
    break ;
  default:
    if ( top_num > 0 ) make_scoreset = 1 ;
    else make_scoreset = 0 ;
  }
  if ( make_scoreset ) {
    sprintf(command, "f s=%d top=%d", a, top_num) ;
    I0 ;
    score_set = atoi(response + 1) ;
  }
  else score_set = a ; 
  switch (code) {
  case BROOKES:
    score = brookes(score_set, rels, N, R, relret) ;
    break ;
  case LBROOKES:
    if ( top_num == 0 ) score = brookes(score_set, rels, N, R, relret) ;
    else score = brookes(score_set, rels, top_num, 0, relret) ;
    break ;
  case DIFFMEANS:
    score = diff_means(score_set, rels, N, R, relret) ;
    break ;
  case RECALL:
    score = recall(score_set, rels, R, relret) ;
    break ;
  case RPREC:
    score = prec(R, score_set, rels, relret) ;
    break ;
  case COMBINED_MEASURE:
    score = combined_measure(score_set, rels, R, relret) ;
    break ;
  case AVEP:
    score = avep(score_set, rels, relret, top_num, gran) ;
    break ;
  case TREC_AVEP:
    score = trec_avep(score_set, rels, relret, R, top_num, TREC_AVEP) ;
    break ;
  case T_AVEP_RCL:
    score = trec_avep(score_set, rels, relret, R, top_num, T_AVEP_RCL) ;
    break ;
  case TS_LF1:
  case TS_LF2:
  case TS_LF3:
  case TS_NF1:
  case TS_NF2:
    score = ftrfuncs(score_set, rels, relret, R, top_num, code) ;
    break ;
  case CBM1:
    score = cbm1(score_set, rels, relret, R, N) ;
  }
  if ( make_scoreset ) delete_set(score_set) ;
  return score ;
 error:
  return -1.0 ;
}


/******************************************************************************

  wt_of_nth(set, n)   SW May 96

  Returns weight of nth record in set, or (lazy) 0 if n > setsize or 0


  setsize(set)  SW June 97

  Returns number of docs in set

******************************************************************************/

double
wt_of_nth(set, n)
     int set, n ;
{
  char command[1024], *response ;

  if ( n <= 0 ) return 0 ;
  if ( n > Setrecs[set]->naw ) return 0 ; /* Sorry (Apr 99) */
  sprintf(command, "s s=%d r=%d f=197", set, n - 1) ;
  I0 ;
  return atof(strstr(response, " ") + 1) ;
  
 error:
  if ( res < 0 ) return 0 ;	/* Not that many records in set */
}

setsize(set)
     int set ;
{
  char command[1024], *response ;
  sprintf(command, "dump np") ;
  I0 ;
  return atoi(strstr(response, " ") + 1) ;
error:
  return -1 ;

}
/******************************************************************************

  wt_of_last_common(set1, set2)   SW May 96

  Returns weight of last record common to set 1 and set 2

******************************************************************************/

double
wt_of_last_common(s1, s2, rr)
     int s1, s2 ;
     int *rr ;			/* Number of rel retrieved */
{
  char command[1024], *response ;
  double weight ;
  int set, np ;

  sprintf(command, "f s=%d s=%d op=and2", s1, s2) ;
  I0 ;
  set = atoi(response + 1) ;
  np = atoi(strstr(response, "np=") + 3) ;
  if ( np > 0 ) {
    sprintf(command, "s s=%d r=%d f=197", set, np - 1) ;
    I0 ;
    weight = atof(strstr(response, " ") + 1) ;
  }
  else weight = 0 ;
  delete_set(set) ;
  *rr = np ;
  return weight ;

 error:
  return -1 ;
}

/******************************************************************************

  make_relset(relsfile, &R)  SW June 95

  Make set from file of DOCNOs
  Returns relset and writes R to the supplied address, or - on error
  Uses BSS_DN_ATTRIBUTE if defined

******************************************************************************/

make_relset(relsfile, Rp)
     FILE *relsfile ;
     int *Rp ;			/* Number of rels */
{
  int j ;
  int numrels = 0 ;
  int relset = -1 ;
  int tempset ;
  int newset ;
  char buf[256] ;
  char command[1024], *response ;
  char *envp, *dn_attr ;
  int result ;

  if ( (envp = getenv("BSS_DN_ATTRIBUTE")) && *envp ) dn_attr = envp ;
  else dn_attr = "dn" ;
  sprintf(command, "f t=ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ") ;
  I0q ;
  relset = atoi(response + 1) ;

  while ( ! feof(relsfile) ) {
    if ( (result = fscanf(relsfile, "%s", buf)) == EOF ) break ;
    sprintf(command, "f a=%s t=%s", dn_attr, buf) ;
    I0q ;
    tempset = atoi(response + 1) ;
    sprintf(command, "f s=%d s=%d op=or", relset, tempset) ;
    I0q ;
    newset = atoi(response + 1) ;
    sprintf(command, "delete s=%d s=%d", relset, tempset) ;
    I0q ;
    relset = newset ;
    numrels++ ;
  }
  *Rp = numrels ;
  return relset ;

 error:
  return -1 ;
}

make_topn_set(n)
     int n ;
{
  char command[4096], *response ;
  int j, result ;
  sprintf(command, "f") ;
  for ( j = 0 ; j < n ; j++ ) {
    sprintf(command + strlen(command), " s=%d", Terms[j].set) ;
    incl[j] = 1 ;
  }
  if ( Hold ) sprintf(command + strlen(command), " hold=%d", Hold) ;
  I0 ; 
  result = atoi(response + 1) ;
  return result ;
 error:
  return -1 ;
}


/******************************************************************************

  remove_set(s1, s2) SW July 95

  Removes the effect of s2 (assumed included in s1) from s1

  add_set()

  NB for add_set, if top_num > 0 the output set is "targeted" and must be
  remade with zero top_num if to be retained. If s1 < 0 (i.e. doesn't exist)
  it just copies (maybe targeted) s2.

  copy_set()

******************************************************************************/

remove_set(set, termnum, cutoff)
     int set, termnum ;
     double cutoff ;
{
  char command[1024], *response ;
  int result ;
  double frig_factor = -C_state->w[termnum]/Terms[termnum].ow ;
  if ( Debug ) fprintf(stderr, "remove_set(): ff=%f\n", frig_factor) ;
  sprintf(command, "f s=%d s=%d frig=%f aw=%f hold=%d",
	  set, Terms[termnum].set, frig_factor, cutoff, Hold) ;
  I0 ;
  result = atoi(response + 1) ;
  return result ;
 error:
  return -1 ;
}

add_set(s1, s2, top_num, cutoff, frig_factor)
     int s1, s2 ;
     int top_num ;
     double cutoff ;		/* Minimum acceptable weight */
     double frig_factor ;
{
  char command[1024], *response ;
  int result ;

  if ( Debug ) fprintf(stderr, "add_set(): ff=%f\n", frig_factor) ;
  if ( s1 < 0 ) sprintf(command, "f s=%d target=%d aw=%f hold=%d",
			s2, top_num, cutoff, Hold) ;
  else sprintf(command, "f s=%d s=%d frig=%f target=%d aw=%f hold=%d",
	       s1, s2, frig_factor, top_num, cutoff, Hold) ;
  I0 ;
  result = atoi(response + 1) ;
  return result ;
error:
  return -1 ;
}

add_sets(s1, s, num_sets, top_num, cutoff, frig_factors)
     int s1, s[] ;
     int num_sets ;
     int top_num ;
     double cutoff ;
     double frig_factors[] ;
{
  char command[4096], *response ;
  int result ;
  int i ;

  sprintf(command, "f s=%d", s1) ;
  for ( i = 0 ; i < num_sets ; i++ )
    sprintf(command + strlen(command), " s=%d frig=%f", s[i], frig_factors[i]) ;
  sprintf(command + strlen(command),
	  " target=%d aw=%f hold=%d", top_num, cutoff, Hold) ;
  I0 ;
  result = atoi(response + 1) ;
  return result ;
error:
  return -1 ;
}

copy_set(s)
     int s ;
{
  char command[1024], *response ;
  int result ;

  sprintf(command, "f s=%d", s) ;
  I0 ;
  result = atoi(response + 1) ;
  return result ;
error:
  return -1 ;
}

double
do_assoc(i, j)			/* Does P(t|s) */
     int i, j ;	
{
  char command[1024], *response ;
  int intersect ;
  sprintf(command, "f s=%d s=%d op=and save=0", Terms[i].set, Terms[j].set) ;
  I0 ;
  intersect = atoi(strstr(response, "np=") + 3) ;
  return (double)intersect/(double)np[i] ;
 error:
  return -1.0 ;
}

setup_assocs(n)
     int n ;
{
  int j, k ;
  double result ;
  for ( j = 0 ; j < n ; j++ ) {
    for ( k = 0 ; k < n ; k++ ) {
      if ( k == j ) {
	assoc[j][k] = 1.0 ;
	continue ;
      }
      result = do_assoc(j, k) ;
      if ( result < 0 ) return -1 ;
      assoc[j][k] = result ;
      if ( ! silent )
	fprintf(stderr,
		"a[%s][%s]=%.2f ", Terms[j].t, Terms[k].t, assoc[j][k]) ;
    }
    if ( ! silent ) fprintf(stderr, "\n") ;
  }
  return 0 ;
}

delete_set(setnum)
     int setnum ;
{
  char command[1024], *response ;
  if ( setnum < 0 ) return 0 ;
  sprintf(command, "del s=%d", setnum) ;
  I0 ;
  return 0 ;
 error:
  return -1 ;
}

